// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace Smdn.Formats.Mime {
 internal static class Parser {
    public static MimeMessage Parse(TextReader reader)
    {
      Dictionary<string, string[]> headers;
      string body;
      MimeMessage[] subParts;

      ParseHeader(reader, out headers);

      ParseBody(reader, headers, out body, out subParts);

      return new MimeMessage(headers, body, subParts);
    }

    private static void ParseHeader(TextReader reader, out Dictionary<string, string[]> headers)
    {
      headers = new Dictionary<string, string[]>(StringComparer.InvariantCultureIgnoreCase);

      var current = string.Empty;

      for (;;) {
        var line = reader.ReadLine();

        // 空行(=ヘッダ部の終わり)
        if (string.IsNullOrEmpty(line))
          break;

        if (line[0] == '\t' || line[0] == ' ') {
          // ヘッダの折り返し(連結する)
          headers[current][headers[current].Length - 1] += " ";
          headers[current][headers[current].Length - 1] += line.Trim();
        }
        else {
          var div = line.IndexOf(":");

          if (div < 0) {
            // 無効な形式のヘッダ(無視する)
            current = string.Empty;
            continue;
          }

          var name    = line.Substring(0, div).Trim();
          var @value = line.Substring(div + 1).Trim();

          // MIME-Version
          if (name.Equals("MIME-Version", StringComparison.InvariantCultureIgnoreCase) &&
              !@value.Contains("1.0"))
            throw new NotSupportedException("unsupported MIME version");

          if (headers.ContainsKey(name)) {
            // 既に同名のヘッダがある場合は、配列を伸長して格納する
            var header = headers[name];

            Array.Resize<string>(ref header, header.Length + 1);

            header[header.Length - 1] = @value;

            headers[name] = header;
          }
          else {
            headers.Add(name, new string[] {@value});
          }

          current = name;
        }
      }
    }

    private static void ParseBody(TextReader reader, Dictionary<string, string[]> headers, out string body, out MimeMessage[] subParts)
    {
      body = string.Empty;
      subParts = new MimeMessage[] {};

      var parts = new List<MimeMessage>();

      // Content-Type
      var contentType = string.Empty;

      if (headers.ContainsKey("Content-Type"))
        contentType = headers["Content-Type"][0];

      if (contentType.ToLowerInvariant().StartsWith("multipart/")) {
        // multipart形式の場合

        // boundary
        var boundary = "--" + GetHeaderParameter(contentType, "boundary", true);
        var lastBoundary = boundary + "--";

        // パートに分割して解析する
        var isPart = false;
        var writer = new StringWriter(new StringBuilder());

        foreach (var line in ReadEachLineKeepNewlineChar(reader)) {
          if (line.StartsWith(boundary)) {
            // 新しいパート
            if (isPart) {
              // 2つめ以降はパートとして解析する
              parts.Add(Parse(new StringReader(writer.ToString())));
            }
            else {
              // 1つめはボディとして扱う
              body = writer.ToString();
            }

            if (line.StartsWith(lastBoundary)) {
              // 最後のパート
              break;
            }
            else {
              isPart = true;

              writer = new StringWriter(new StringBuilder());
            }
          }
          else {
            writer.Write(line);
          }
        }
      }
      else {
        // multipart以外の場合
        body = reader.ReadToEnd();
      }

      subParts = parts.ToArray();
    }

    private static IEnumerable<string> ReadEachLineKeepNewlineChar(TextReader reader)
    {
      var line = new StringBuilder();
      var cr = false;
      var newline = false;

      for (;;) {
        if (-1 == reader.Peek()) {
          yield return line.ToString();
          break;
        }

        var c = reader.Read();

        if (newline) {
          yield return line.ToString();
          line = new StringBuilder();
          newline = false;
        }

        if (c == 0x0d) { // CR
          cr = true;
        }
        else if (c == 0x0a || cr) { // LF, CRLF
          cr = false;
          newline = true;
        }

        line.Append((char)c);
      }
    }

    public static string GetHeaderParameter(string header, string parameter, bool dequote)
    {
      var paramStarts = header.ToLowerInvariant().IndexOf(parameter.ToLower() + "=");

      if (paramStarts == -1)
        return string.Empty;

      paramStarts += parameter.Length + 1;

      var paramEnds = header.IndexOf(";", paramStarts + 1);
      var param = string.Empty;

      if (paramEnds == -1)
        param = header.Substring(paramStarts);
      else
        param = header.Substring(paramStarts, paramEnds - paramStarts).Trim();

      // if quoted
      if (dequote && param.StartsWith("\"") && param.EndsWith("\""))
        param = param.Substring(1, param.Length - 2);

      return param;
    }
  }
}