// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Text;

namespace Smdn.Formats.Mime {
  public static class QuotedPrintable {
    public static string ToQuotedPrintableString(byte[] quotingStream)
    {
      var quoted = new MemoryStream(quotingStream.Length * 4);
      var charcount = 0;

      foreach (var octet in quotingStream) {
        if (73 < charcount) {
          // 次のエスケープで76文字を越える可能性がある場合
          var escaped = false;

          quoted.WriteByte(0x3d); // '=' 0x3d

          if (octet == 0x09 || octet == 0x20) {
            // '\t' 0x09 or ' ' 0x20
            quoted.WriteByte(hexChar[(octet & 0xf0) >> 4]);
            quoted.WriteByte(hexChar[octet & 0x0f]);

            escaped = true;
          }

          quoted.WriteByte(0x0d); // '\r' 0x0d
          quoted.WriteByte(0x0a); // '\n' 0x0a

          charcount = 0;

          if (escaped)
            continue;
        }

        if ((0x21 <= octet && octet <= 0x3c) ||
            (0x3e <= octet && octet <= 0x7e) ||
            octet == 0x09 ||
            octet == 0x20) {
          // printable char (except '=' 0x3d)
          quoted.WriteByte(octet);

          charcount++;
        }
        else {
          // '=' 0x3d or control char
          quoted.WriteByte(0x3d); // '=' 0x3d
          quoted.WriteByte(hexChar[(octet & 0xf0) >> 4]);
          quoted.WriteByte(hexChar[octet & 0x0f]);

          charcount += 3;
        }
      }

      return Encoding.ASCII.GetString(quoted.ToArray());
    }

    public static byte[] FromQuotedPrintableString(string quotedString)
    {
      var quoted = Encoding.ASCII.GetBytes(quotedString);
      var decoded = new MemoryStream(quotedString.Length);
      var prevQuoted = false;

      for (var index = 0; index < quoted.Length;) {
        if (quoted[index] == 0x3d) { // '=' 0x3d
          index++;

          if (quoted[index] == 0x0d && quoted[index + 1] == 0x0a) {
            // '\r' 0x0d, '\n' 0x0a
            index += 2;

            prevQuoted = false;
          }
          else {
            byte d = 0x00;

            for (var i = 0; i < 2; i++) {
              d <<= 4;

              if (0x30 <= quoted[index] && quoted[index] <= 0x39)
                // '0' 0x30 to '9' 0x39
                d |= (byte)(quoted[index++] - 0x30);
              else if (0x41 <= quoted[index] && quoted[index] <= 0x46)
                // 'A' 0x41 to 'F' 0x46
                d |= (byte)(quoted[index++] - 0x37);
              else if (0x61 <= quoted[index] && quoted[index] <= 0x66)
                // 'a' 0x61 to 'f' 0x66
                d |= (byte)(quoted[index++] - 0x57);
              else
                throw new FormatException("invalid quoted stream");
            }

            decoded.WriteByte(d);

            prevQuoted = true;
          }
        }
        else if (quoted[index] == 0x0d ||
                 quoted[index] == 0x0a) { // '\r' 0x0d, '\n' 0x0a
          if (!prevQuoted)
            decoded.WriteByte(quoted[index]);

          index++;
        }
        else {
          decoded.WriteByte(quoted[index++]);

          prevQuoted = false;
        }
      }

      return decoded.ToArray();
    }

    private static readonly byte[] hexChar = new byte[] {
         0x30, 0x31, 0x32, 0x33,  // '0', '1', '2', '3'
         0x34, 0x35, 0x36, 0x37,  // '4', '5', '6', '7'
         0x38, 0x39, 0x41, 0x42,  // '8', '9', 'A', 'B'
         0x43, 0x44, 0x45, 0x46}; // 'C', 'D', 'E', 'F'
  }
}
