// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Security.Cryptography;

namespace Smdn.Protocols.Imap4.Client.Protocol.Authentication {
  /*
   * RFC 2831 - Using Digest Authentication as a SASL Mechanism
   * http://tools.ietf.org/html/rfc2831
   */
  internal class DigestMD5AuthenticationModule : StringAuthenticationModuleBase {
    public override ImapAuthenticationMechanism Mechanism {
      get { return ImapAuthenticationMechanism.DigestMD5; }
    }

    protected override string ExchangeChallengeResponse(string challenge, ImapConnection connection, System.Net.NetworkCredential credential)
    {
      var pairs = ParseChallenge(challenge);

      if (AuthenticationStep == 1) {
        /* 
         * 2.1.3 Step Three
         */
        if (pairs.ContainsKey("rspauth"))
          return string.Empty; // success
        else
          return null; // unexpected server challenge
      }

      /*
       * 2.1.1 Step One
       */

      //    algorithm
      //       This directive is required for backwards compatibility with HTTP
      //       Digest., which supports other algorithms. . This directive is
      //       required and MUST appear exactly once; if not present, or if
      //       multiple instances are present, the client should abort the
      //       authentication exchange.
      //
      //         algorithm         = "algorithm" "=" "md5-sess"
      if (!(pairs.ContainsKey("algorithm") && pairs["algorithm"] == "md5-sess"))
        return null;

      //    charset
      //       This directive, if present, specifies that the server supports
      //       UTF-8 encoding for the username and password. If not present, the
      //       username and password must be encoded in ISO 8859-1 (of which
      //       US-ASCII is a subset). The directive is needed for backwards
      //       compatibility with HTTP Digest, which only supports ISO 8859-1.
      //       This directive may appear at most once; if multiple instances are
      //       present, the client should abort the authentication exchange.
      //
      //         charset           = "charset" "=" "utf-8"
      Encoding charset;

      if (pairs.ContainsKey("charset") && pairs["charset"] == "utf-8")
        charset = Encoding.UTF8;
      else
        charset = Encoding.GetEncoding("iso-8859-1");

      string realm, nonce, qop;

      if (!pairs.TryGetValue("realm", out realm))
        return null;
      if (!pairs.TryGetValue("nonce", out nonce))
        return null;
      if (!pairs.TryGetValue("qop", out qop))
        return null;

      qop = "auth"; // ignore qop-options

      /*
       * 2.1.2 Step Two
       */
      AuthenticationStep = 1;

      //    cnonce
      //       It is RECOMMENDED that it contain at least 64 bits of entropy.
      string cnonce;

      {
        var cn = new char[64];
        var rand = new Random((int)DateTime.Now.Ticks);

        for (var i = 0; i < cn.Length; i++) {
          cn[i] = (char)rand.Next(0x61, 0x7b); // a-z
        }

        cnonce = new string(cn);
      }

      //    nonce-count
      //       The nc-value is the hexadecimal count of the number of requests
      //       (including the current request) that the client has sent with the
      //       nonce value in this request.  For example, in the first request
      //       sent in response to a given nonce value, the client sends
      //       "nc=00000001"
      var nc = 1.ToString("D8");

      //    digest-uri
      //       Indicates the principal name of the service with which the client
      //       wishes to connect, formed from the serv-type, host, and serv-name.
      //       For example, the FTP service on "ftp.example.com" would have a
      //       "digest-uri" value of "ftp/ftp.example.com"; the SMTP server from
      //       the example above would have a "digest-uri" value of
      //       "smtp/mail3.example.com/example.com".
      var digestUri = string.Format("imap/{0}", credential.Domain ?? connection.Host);

      //       response-value  =
      //          HEX( KD ( HEX(H(A1)),
      //                  { nonce-value, ":" nc-value, ":",
      //                    cnonce-value, ":", qop-value, ":", HEX(H(A2)) }))
      var h = MD5.Create();

      var responseValue = HexKD(h,
                                HexHA1(h, charset, credential.UserName, realm, credential.Password ?? string.Empty, nonce, cnonce, null),
                                nonce,
                                nc,
                                cnonce,
                                qop,
                                HexHA2(h, qop, digestUri));

      return string.Format("{0}username=\"{1}\",realm=\"{2}\"," +
                           "nonce=\"{3}\",nc={4},cnonce=\"{5}\",digest-uri=\"{6}\"," +
                           "response={7},qop={8}",
                           ((charset == Encoding.UTF8) ? "charset=utf-8," : string.Empty), credential.UserName, realm,
                           nonce, nc, cnonce, digestUri,
                           responseValue, qop);
    }

    private string HexHA1(MD5 h, Encoding charset, string username, string realm, string password, string nonce, string cnonce, string authzid)
    {
      //    If authzid is specified, then A1 is
      // 
      //       A1 = { H( { username-value, ":", realm-value, ":", passwd } ),
      //            ":", nonce-value, ":", cnonce-value, ":", authzid-value }
      // 
      //    If authzid is not specified, then A1 is
      // 
      //       A1 = { H( { username-value, ":", realm-value, ":", passwd } ),
      //            ":", nonce-value, ":", cnonce-value }
      var a1 = h.ComputeHash(ArrayExtension.Concat(charset.GetBytes(username),
                                                   ImapEncoding.Transfer7Bit.GetBytes(string.Format(":{0}:", realm)),
                                                   charset.GetBytes(password)));

      var tmp = (authzid == null)
        ? string.Format(":{0}:{1}", nonce, cnonce)
        : string.Format(":{0}:{1}:{2}", nonce, cnonce, authzid);

      return HexH(h, ArrayExtension.Concat(a1, ImapEncoding.Transfer7Bit.GetBytes(tmp)));
    }

    private string HexHA2(MD5 h, string qop, string digestUri)
    {
      //    If the "qop" directive's value is "auth", then A2 is:
      // 
      //       A2       = { "AUTHENTICATE:", digest-uri-value }
      // 
      //    If the "qop" value is "auth-int" or "auth-conf" then A2 is:
      // 
      //       A2       = { "AUTHENTICATE:", digest-uri-value,
      //                ":00000000000000000000000000000000" }
      string a2;

      if (string.Equals(qop, "auth-int", StringComparison.InvariantCultureIgnoreCase) ||
          string.Equals(qop, "auth-conf", StringComparison.InvariantCultureIgnoreCase))
        a2 = string.Format("AUTHENTICATE:{0}:00000000000000000000000000000000", digestUri);
      else
        a2 = string.Format("AUTHENTICATE:{0}", digestUri);

      return HexH(h, ImapEncoding.Transfer7Bit.GetBytes(a2));
    }

    private string HexKD(MD5 h, string hexHA1, string nonce, string nc, string cnonce, string qop, string hexHA2)
    {
      //       response-value  =
      //          HEX( KD ( HEX(H(A1)),
      //                  { nonce-value, ":" nc-value, ":",
      //                    cnonce-value, ":", qop-value, ":", HEX(H(A2)) }))

      //    Let KD(k, s) be H({k, ":", s}), i.e., the 16 octet hash of the string
      //    k, a colon and the string s.
      var tmp = string.Format("{0}:{1}:{2}:{3}:{4}:{5}", hexHA1, nonce, nc, cnonce, qop, hexHA2);

      return HexH(h, Encoding.ASCII.GetBytes(tmp));
    }

    private string HexH(MD5 h, byte[] s)
    {
      //    Let H(s) be the 16 octet MD5 hash [RFC 1321] of the octet string s.

      //    Let HEX(n) be the representation of the 16 octet MD5 hash n as a
      //    string of 32 hex digits (with alphabetic characters always in lower
      //    case, since MD5 is case sensitive).
      return ToLowerCaseHexString(h.ComputeHash(s));
    }

    private Dictionary<string, string> ParseChallenge(string str)
    {
      // quoted-string  = ( <"> qdstr-val <"> )
      // qdstr-val      = *( qdtext | quoted-pair )
      // qdtext         = <any TEXT except <">>
      // quoted-pair    = "\" CHAR
      var splitAt = new List<int>();
      var quoted = false;

      for (var i = 0; i < str.Length; i++) {
        if (str[i] == '\\') // escape
          i++;
        else if (str[i] == '\"')
          quoted = !quoted;
        else if (!quoted && str[i] == ',')
          splitAt.Add(i);
      }

      splitAt.Add(str.Length); // end of line

      var pairs = new string[splitAt.Count];

      for (var i = 0; i < pairs.Length; i++) {
        if (i == 0)
          pairs[i] = str.Substring(0, splitAt[i]).Trim();
        else
          pairs[i] = str.Substring(splitAt[i - 1] + 1, splitAt[i] - splitAt[i - 1] - 1).Trim();
      }

      var ret = new Dictionary<string, string>();

      foreach (var p in Array.ConvertAll(pairs, delegate(string pair) {
        var delim = pair.IndexOf('=');
        var dequote = (pair[delim + 1] == '\"') && pair.EndsWith("\"");

        if (dequote)
          return new KeyValuePair<string, string>(pair.Substring(0, delim), pair.Substring(delim + 2, pair.Length - (delim + 3)));
        else
          return new KeyValuePair<string, string>(pair.Substring(0, delim), pair.Substring(delim + 1));
      })) {
        // TODO: quoted-pair
        ret.Add(p.Key, p.Value);
      }

      return ret;
    }
  }
}