// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;
using System.Reflection;

namespace Smdn.Protocols.Imap4.Client.Protocol.Authentication {
  public static class ImapAuthenticationManager {
    public static IEnumerable<Type> RegisteredModuleTypes {
      get { return modules.Values; }
    }

    static ImapAuthenticationManager()
    {
      var itype = typeof(IImapAuthenticationModule);

      foreach (var type in Assembly.GetExecutingAssembly().GetTypes()) {
        if (type.IsAbstract || !itype.IsAssignableFrom(type))
          continue;

        try {
          Register(type);
        }
        catch (TargetInvocationException) {
          // ignore
          Trace.Info("authentication module not registered: {0}", type.FullName);
        }
      }
    }

    public static void Register(Type moduleType)
    {
      if (moduleType == null)
        throw new ArgumentNullException("moduleType");

      var module = Activator.CreateInstance(moduleType) as IImapAuthenticationModule;

      if (modules.ContainsKey(module.Mechanism))
        Unregister(modules[module.Mechanism]);

      lock ((modules as System.Collections.ICollection).SyncRoot) {
        modules.Add(module.Mechanism, moduleType);
      }

      Trace.Info("authentication module registered: {0} : {1}", module.Mechanism, moduleType.FullName);
    }

    public static void Unregister(ImapAuthenticationMechanism mechanism)
    {
      if (mechanism == null)
        throw new ArgumentNullException("module");

      if (modules.ContainsKey(mechanism))
        Unregister(modules[mechanism]);
    }

    public static void Unregister(Type moduleType)
    {
      if (moduleType == null)
        throw new ArgumentNullException("moduleType");

      ImapAuthenticationMechanism mechanism = null;

      lock ((modules as System.Collections.ICollection).SyncRoot) {
        foreach (var module in modules) {
          if (module.Value == moduleType) {
            mechanism = module.Key;
            break;
          }
        }

        if (mechanism == null)
          return;

        modules.Remove(mechanism);
      }

      Trace.Info("authentication module unregistered: {0} : {1}", mechanism, moduleType.FullName);
    }

    public static IImapAuthenticationModule CreateModule(ImapAuthenticationMechanism mechanism)
    {
      if (!modules.ContainsKey(mechanism))
        return null;

      try {
        return Activator.CreateInstance(modules[mechanism]) as IImapAuthenticationModule;
      }
      catch (TargetInvocationException) {
        // ignore exception
        return null;
      }
    }

    private static Dictionary<ImapAuthenticationMechanism, Type> modules = new Dictionary<ImapAuthenticationMechanism, Type>();
  }
}
