// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Net;
using System.Reflection;

#if MONO_SECURITY_DLL
using Mono.Security.Protocol.Ntlm;
#endif

namespace Smdn.Protocols.Imap4.Client.Protocol.Authentication {
  /*
   * [MS-NLMP]: NT LAN Manager (NTLM) Authentication Protocol Specification
   * http://msdn.microsoft.com/en-us/library/cc236621(PROT.13).aspx
   * 
   * Microsoft Open Specification Support Team Blog : NTLM Overview
   * http://blogs.msdn.com/openspecification/archive/2009/05/01/ntlm-overview.aspx
   * 
   * The NTLM Authentication Protocol and Security Support Provider
   * http://davenport.sourceforge.net/ntlm.html
   */
  internal class NTLMAuthenticationModule : BinaryAuthenticationModuleBase {
    public override ImapAuthenticationMechanism Mechanism {
      get { return ImapAuthenticationMechanism.NTLM; }
    }

    public NTLMAuthenticationModule()
    {
#if !MONO_SECURITY_DLL
      try {
        var assm = Assembly.Load("Mono.Security, Version=2.0.0.0, Culture=neutral, PublicKeyToken=0738eb9f132ed756");

        typeOfType1Message = assm.GetType("Mono.Security.Protocol.Ntlm.Type1Message", true);
        typeOfType2Message = assm.GetType("Mono.Security.Protocol.Ntlm.Type2Message", true);
        typeOfType3Message = assm.GetType("Mono.Security.Protocol.Ntlm.Type3Message", true);
      }
      catch (System.IO.FileNotFoundException) {
        throw new NotSupportedException();
      }
      catch (System.IO.FileLoadException) {
        throw new NotSupportedException();
      }
      catch (BadImageFormatException) {
        throw new NotSupportedException();
      }
      catch (TypeLoadException) {
        throw new NotSupportedException();
      }
#endif
    }

    protected override byte[] ExchangeChallengeResponse(byte[] challenge, ImapConnection connection, System.Net.NetworkCredential credential)
    {
      switch (AuthenticationStep) {
        case 0: { // send NTLM negotiate message (Type 1)
#if MONO_SECURITY_DLL
          var type1 = new Type1Message();

          type1.Host = connection.Host;
          type1.Domain = credential.Domain ?? connection.Host;

          AuthenticationStep++;

          return type1.GetBytes();
#else
          var type1 = Activator.CreateInstance(typeOfType1Message);

          typeOfType1Message.GetProperty("Host").SetValue(type1, connection.Host, null);
          typeOfType1Message.GetProperty("Domain").SetValue(type1, credential.Domain ?? connection.Host, null);

          AuthenticationStep++;

          return typeOfType1Message.GetMethod("GetBytes").Invoke(type1, null) as byte[];
#endif
        }

        case 1: { // receive NTLM challenge message (Type 2) and send NTLM authenticate message (Type 3)
#if MONO_SECURITY_DLL
          var type2 = new Type2Message(challenge);
          var type3 = new Type3Message();

          type3.Host = connection.Host;
          type3.Domain = credential.Domain ?? connection.Host;
          type3.Challenge = type2.Nonce;
          type3.Password = credential.Password ?? string.Empty;
          type3.Username = credential.UserName;

          AuthenticationStep++;

          return type3.GetBytes();
#else
          var type2 = Activator.CreateInstance(typeOfType2Message, challenge);
          var type3 = Activator.CreateInstance(typeOfType3Message);

          var nonce = typeOfType2Message.GetProperty("Nonce").GetValue(type2, null) as byte[];

          typeOfType3Message.GetProperty("Host").SetValue(type3, connection.Host, null);
          typeOfType3Message.GetProperty("Domain").SetValue(type3, credential.Domain ?? connection.Host, null);
          typeOfType3Message.GetProperty("Challenge").SetValue(type3, nonce, null);
          typeOfType3Message.GetProperty("Password").SetValue(type3, credential.Password ?? string.Empty, null);
          typeOfType3Message.GetProperty("Username").SetValue(type3, credential.UserName, null);

          AuthenticationStep++;

          return typeOfType3Message.GetMethod("GetBytes").Invoke(type3, null) as byte[];
#endif
        }

        default:
          return null; // unexpected server challenge
      }
    }

#if !MONO_SECURITY_DLL
    private /*readonly*/ Type typeOfType1Message;
    private /*readonly*/ Type typeOfType2Message;
    private /*readonly*/ Type typeOfType3Message;
#endif
  }
}