// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Text;

namespace Smdn.Protocols.Imap4.Client.Protocol {
  public class ImapBufferedStream : System.IO.Stream {
    private const byte CR = 0x0d;
    private const byte LF = 0x0a;

    public Stream BaseStream {
      get { return baseStream; }
    }

    public override bool CanSeek {
      get { return baseStream.CanSeek; }
    }

    public override bool CanRead {
      get { return baseStream.CanWrite; }
    }

    public override bool CanWrite {
      get { return baseStream.CanRead; }
    }

    public override bool CanTimeout {
      get { return baseStream.CanTimeout; }
    }

    public override long Position {
      get { throw new NotSupportedException(); }
      set { throw new NotSupportedException(); }
    }

    public override long Length {
      get { throw new NotSupportedException(); }
    }

    public int BufferSize {
      get { return buffer.Length; }
    }

    public ImapBufferedStream(Stream baseStream)
    {
      if (baseStream == null)
        throw new ArgumentNullException("baseStream");

      this.baseStream = baseStream;
      this.buffer = new byte[8192];
    }

    public override void Close()
    {
      Dispose(true);
    }

    protected override void Dispose(bool disposing)
    {
      if (disposed)
        return;

      baseStream = null;
      buffer = null;

      disposed = true;
    }

    public override void SetLength(long value)
    {
      throw new NotSupportedException();
    }

    public override long Seek(long offset, SeekOrigin origin)
    {
      throw new NotSupportedException();
    }

    public override void Flush()
    {
      throw new NotSupportedException();
    }

    public override int ReadByte()
    {
      RejectDisposed();

      if (bufRemain == 0)
        if (FillBuffer() <= 0)
          return -1;

      bufRemain--;

      return buffer[bufOffset++];
    }

    public int ReadToCrLf(int readAtLeast, out string line)
    {
      RejectDisposed();

      if (readAtLeast < 0)
        throw new ArgumentOutOfRangeException("readAtLeast", "must be zero or positive number");

      if (!bufferRead) {
        // discard
        bufRemain = 0;
        bufferRead = true;
      }

      line = null;

      var buf = new StringBuilder(Math.Max(readAtLeast, buffer.Length));
      var cr = false;
      var lf = false;

      for (;;) {
        if (bufRemain == 0)
          if (FillBuffer() <= 0)
            return 0;

        var b = buffer[bufOffset];

        buf.Append((char)b);

        bufOffset++;
        bufRemain--;

        lf = (cr && b == LF);

        if (lf && readAtLeast <= buf.Length)
          break;

        cr = (b == CR);
      }

      line = buf.ToString();

      return buf.Length;
    }

    public override int Read(byte[] dest, int offset, int count)
    {
      RejectDisposed();

      if (dest == null)
        throw new ArgumentNullException("dest");
      if (offset < 0)
        throw new ArgumentOutOfRangeException("offset", "must be greater than or equals to 0");
      if (count < 0)
        throw new ArgumentOutOfRangeException("count", "must be greater than or equals to 0");
      if (dest.Length < offset + count)
        throw new ArgumentException("invalid range");

      if (!bufferRead) {
        // discard
        FillBuffer();
        bufferRead = true;
      }

      if (count <= bufRemain) {
        Buffer.BlockCopy(buffer, bufOffset, dest, offset, count);
        bufOffset += count;
        bufRemain -= count;

        return count;
      }

      var read = 0;

      if (bufRemain != 0) {
        Buffer.BlockCopy(buffer, bufOffset, dest, offset, bufRemain);

        read    = bufRemain;
        offset += bufRemain;
        count  -= bufRemain;

        bufRemain = 0;
      }

      if (buffer.Length <= count) {
        // read from base stream
        read += baseStream.Read(dest, offset, count);
      }
      else {
        FillBuffer();

        if (count <= bufRemain) {
          Buffer.BlockCopy(buffer, 0, dest, offset, count);

          read += count;

          bufOffset += count;
          bufRemain -= count;
        }
        else {
          Buffer.BlockCopy(buffer, 0, dest, offset, bufRemain);

          read += bufRemain;

          bufRemain = 0;
        }
      }

      return read;
    }

    private int FillBuffer()
    {
      bufOffset = 0;
      bufRemain = baseStream.Read(buffer, 0, buffer.Length);

      return bufRemain;
    }

    public void WriteToCrLf(byte[] src)
    {
      RejectDisposed();

      if (bufferRead) {
        // discard
        bufOffset = 0;
        bufRemain = buffer.Length;
        bufferRead = false;
      }

      var offset = 0;
      var count = src.Length;

      if (bufRemain < count) {
        // flush buffered
        baseStream.Write(buffer, 0, bufOffset);

        bufOffset = 0;
        bufRemain = buffer.Length;

        for (;;) {
          if (count < buffer.Length)
            break;

          baseStream.Write(src, offset, buffer.Length);

          offset  += buffer.Length;
          count   -= buffer.Length;
        }
      }

      Buffer.BlockCopy(src, offset, buffer, bufOffset, count);

      bufRemain -= count;
      bufOffset += count;

      if (2 <= bufOffset && buffer[bufOffset - 2] == CR && buffer[bufOffset - 1] == LF) {
        baseStream.Write(buffer, 0, bufOffset);

        bufOffset = 0;
        bufRemain = buffer.Length;
      }
    }

    public void Write(Stream stream)
    {
      RejectDisposed();

      if (bufferRead) {
        // discard
        bufOffset = 0;
        bufRemain = buffer.Length;
        bufferRead = false;
      }
      else if (0 < bufOffset) {
        // flush buffered
        baseStream.Write(buffer, 0, bufOffset);

        bufOffset = 0;
        bufRemain = buffer.Length;
      }

      var read = 0;

      for (;;) {
        var r = stream.Read(buffer, 0, buffer.Length);

        if (r <= 0)
          continue;

        baseStream.Write(buffer, 0, r);

        read += r;

        if (read == stream.Length)
          return;
      }
    }

    public override void Write(byte[] src, int offset, int count)
    {
      RejectDisposed();

      throw new NotSupportedException();
    }

    internal void ClearBuffer()
    {
      bufOffset = 0;
      bufRemain = buffer.Length;
    }

    private void RejectDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().Name);
    }

    private Stream baseStream;
    private byte[] buffer;
    private bool bufferRead = true;
    private bool disposed = false;
    private int bufOffset = 0;
    private int bufRemain  = 0;
  }
}