// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#if DUMP
#define TRACING
#endif

#if TRACE
#define TRACING
#endif

using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Net.Security;
using System.Threading;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;

namespace Smdn.Protocols.Imap4.Client.Protocol {
  // Layers:
  //   Session.ImapSession
  //   Transaction.IImapTransaction/Transaction.ImapTransactionRequest/Transaction.ImapTransactionResult
  // * Protocol.ImapConnection/Protocol.ImapCommand/Protocol.ImapResponse
  //   (IMAP4)

  public class ImapConnection : IDisposable {
    /*
     * class members
     */
    public const int ImapDefaultPort = 143;
    public const int ImapsDefaultPort = 993;
    public const int DefaultTimeoutMilliseconds = 16000;

    public static X509Certificate2Collection CertCollection {
      get { return certCollection; }
    }

    public static RemoteCertificateValidationCallback CertificateValidationCallback {
      get { return certificateValidationCallback; }
      set { certificateValidationCallback = value; }
    }

    public static LocalCertificateSelectionCallback CertificateSelectionCallback {
      get { return certificateSelectionCallback; }
      set { certificateSelectionCallback = value; }
    }

    public static Stream CreateSslStream(ImapConnection connection, Stream baseStream)
    {
      var sslStream = new SslStream(baseStream, false, certificateValidationCallback, certificateSelectionCallback);

      sslStream.AuthenticateAsClient(connection.Host, certCollection, SslProtocols.Default, true);

      Trace.Log(connection, "created SSL/TLS stream for '{0}'", connection.Host);
      Trace.Log(connection, "  protocol: {0}", sslStream.SslProtocol);
      Trace.Log(connection, "  cipher: {0} strength {1}", sslStream.CipherAlgorithm, sslStream.CipherStrength);
      Trace.Log(connection, "  hash: {0} strength {1}", sslStream.HashAlgorithm, sslStream.HashStrength);
      Trace.Log(connection, "  key exchange: {0} strength {1}", sslStream.KeyExchangeAlgorithm, sslStream.KeyExchangeStrength);
      Trace.Log(connection, "  authenticated: {0}", sslStream.IsAuthenticated);
      Trace.Log(connection, "  authenticated as server: {0}", sslStream.IsServer);
      Trace.Log(connection, "  mutually authenticated: {0}", sslStream.IsMutuallyAuthenticated);
      Trace.Log(connection, "  signed: {0}", sslStream.IsSigned);
      Trace.Log(connection, "  encrypted: {0}", sslStream.IsEncrypted);

      return sslStream;
    }

    private static int connectionIdCount = 0;
    private static readonly X509Certificate2Collection certCollection = new X509Certificate2Collection();
    private static RemoteCertificateValidationCallback certificateValidationCallback = null;
    private static LocalCertificateSelectionCallback certificateSelectionCallback = null;

    /*
     * properties
     */
    internal int Id {
      get { return id; }
    }

    public string Host {
      get
      {
        CheckDisposed();
        CheckConnected();
        return host;
      }
    }

    public int Port {
      get
      {
        CheckDisposed();
        CheckConnected();
        return remoteEndPoint.Port;
      }
    }

    internal IPEndPoint RemoteEndPoint {
      get
      {
        CheckDisposed();
        CheckConnected();
        return remoteEndPoint;
      }
    }

    internal IPEndPoint LocalEndPoint {
      get
      {
        CheckDisposed();
        CheckConnected();
        return localEndPoint;
      }
    }

    internal bool IsSslConnection {
      get
      {
        CheckDisposed();
        CheckConnected();
#if TRACING
        return ((stream.BaseStream as ImapTracingStream).BaseStream is SslStream);
#else
        return (stream.BaseStream is SslStream);
#endif
      }
    }

    public bool IsIdling {
      get
      {
        CheckDisposed();
        CheckConnected();
        return isIdling;
      }
      internal set { isIdling = value; }
    }

    public Socket Client {
      get
      {
        CheckDisposed();
        CheckConnected();
        return client.Client;
      }
    }

    // for '5.4. Autologout Timer'
    public DateTime LastSentTime {
      get
      {
        CheckDisposed();
        return lastSentTime;
      }
    }

    // IMAP4 non-synchronizing literals
    // http://tools.ietf.org/html/rfc2088
    public bool SynchronizeLiterals
    {
      get
      {
        CheckDisposed();
        return sender.SynchronizeLiterals;
      }
      set
      {
        CheckDisposed();
        sender.SynchronizeLiterals = value;
      }
    }

    public ImapConnection(string host)
      : this(host, -1, null)
    {
    }

    public ImapConnection(string host, int port)
      : this(host, port, null)
    {
    }

    public ImapConnection(string host, ImapUpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
      : this(host, -1, createAuthenticatedStreamCallback)
    {
    }

    public ImapConnection(string host, int port, ImapUpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      this.id = Interlocked.Increment(ref connectionIdCount);

      if (connectionIdCount == int.MaxValue)
        Interlocked.Exchange(ref connectionIdCount, 0);

      Connect(host, port, createAuthenticatedStreamCallback);
    }

    public void Connect(string host)
    {
      Connect(host, -1, null);
    }

    public void Connect(string host, int port)
    {
      Connect(host, port, null);
    }

    public void Connect(string host, ImapUpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      Connect(host, -1, createAuthenticatedStreamCallback);
    }

    public void Connect(string host, int port, ImapUpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      if (port == -1) {
        if (createAuthenticatedStreamCallback == null)
          port = ImapDefaultPort;
        else
          port = ImapsDefaultPort;
      }

      if (host == null)
        throw new ArgumentNullException("host");
      if (port < IPEndPoint.MinPort || IPEndPoint.MaxPort < port)
        throw new ArgumentOutOfRangeException("port");

      this.host = host;

      Trace.Info("CID:{0} connecting to {1}", id, host);

      try {
        var client = new TcpClient();

        client.ReceiveTimeout = DefaultTimeoutMilliseconds;
        client.SendTimeout    = DefaultTimeoutMilliseconds;

        client.Connect(host, port);

        this.client = client;
        this.remoteEndPoint = client.Client.RemoteEndPoint as IPEndPoint;
        this.localEndPoint = client.Client.LocalEndPoint as IPEndPoint;

        SetStream(client.GetStream());

        if (createAuthenticatedStreamCallback != null)
          UpgradeStream(createAuthenticatedStreamCallback);
      }
      catch (SocketException ex) {
        if (this.client != null) {
          this.client.Close();
          this.client = null;
        }
        throw new ImapConnectionException(ex.Message, ex);
      }
      catch {
        if (this.client != null) {
          this.client.Close();
          this.client = null;
        }
        throw;
      }

      Trace.Info("CID:{0} connected to {1} ({2})", id, host, remoteEndPoint);
    }

    void IDisposable.Dispose()
    {
      if (disposed)
        return;

      try {
        Trace.Info("CID:{0} disconnecting from {1} ({2})", id, host, remoteEndPoint);

        stream = null;

        if (client != null) {
          client.Close();
          client = null;
        }

        Trace.Info("CID:{0} disconnected", id);
      }
      finally {
        disposed = true;
      }
    }

    public void Close()
    {
      CheckDisposed();

      (this as IDisposable).Dispose();
    }

    private void SetStream(Stream newStream)
    {
#if TRACING
      stream = new ImapBufferedStream(new ImapTracingStream(this, newStream));
#else
      stream = new ImapBufferedStream(newStream);
#endif

      receiver = new ImapResponseReceiver(stream);
      sender = new ImapCommandSender(stream, (sender == null ? true : sender.SynchronizeLiterals));
    }

    internal void UpgradeStream(ImapUpgradeConnectionStreamCallback upgradeStreamCallback)
    {
      CheckDisposed();
      CheckConnected();

      try {
#if TRACING
        SetStream(upgradeStreamCallback(this, (stream.BaseStream as ImapTracingStream).BaseStream));
#else
        SetStream(upgradeStreamCallback(this, stream.BaseStream));
#endif
      }
      catch (AuthenticationException ex) {
        throw new ImapConnectionException(string.Format("upgrading stream failed (callback: {0})", upgradeStreamCallback.Method), ex);
      }
      catch (IOException ex) {
        throw new ImapConnectionException(string.Format("upgrading stream failed (callback: {0})", upgradeStreamCallback.Method), ex);
      }
    }

    public ImapCommand CreateCommand(string command, params ImapString[] arguments)
    {
      CheckDisposed();
      CheckConnected();

      var comm = new ImapCommand(command, commandTag.ToString("D4"), arguments);

      if (Interlocked.Increment(ref commandTag) == 0x10000)
        Interlocked.Exchange(ref commandTag, 0);

      return comm;
    }

    public ImapCommand CreateContinuingCommand(params ImapString[] arguments)
    {
      CheckDisposed();
      CheckConnected();

      return new ImapCommand(null, null, arguments);
    }

#region "seinding / receiving"
    public interface ICommandContinuationContext {
    }

    private class SendCommandContext : ICommandContinuationContext {
      public ImapCommand Command;
      public bool Sent;

      public SendCommandContext(ImapCommand command)
      {
        this.Command = command;
        this.Sent = false;
      }
    }

    public ICommandContinuationContext SendCommand(ImapCommand command)
    {
      return SendCommand(new SendCommandContext(command));
    }

    public ICommandContinuationContext SendCommand(ICommandContinuationContext context)
    {
      CheckDisposed();
      CheckConnected();

      try {
        var c = context as SendCommandContext;

        if (c.Sent) {
          sender.SendFragments();
        }
        else {
          sender.Send(c.Command);
          c.Sent = true;
        }

        lastSentTime = DateTime.Now;

        if (sender.CommandContinuing)
          return c;
        else
          return null;
      }
      catch (IOException ex) {
        if (ex.InnerException is SocketException)
          throw new ImapConnectionException((ex.InnerException as SocketException).Message, ex.InnerException);
        else
          throw ex;
      }
    }

    public ImapResponse TryReceiveResponse()
    {
      CheckDisposed();
      CheckConnected();

      // return queued response
      if (0 < responseQueue.Count)
        return responseQueue.Dequeue();

      // receive and parse response
      for (;;) {
        try {
          var response = receiver.Receive();

          if (response == null)
            break;

          responseQueue.Enqueue(response);

          if (!receiver.ResponseContinuing)
            break;
        }
        catch (IOException ex) {
          var socketException = ex.InnerException as SocketException;

          if (socketException == null)
            // unexpected IO error
            throw ex;

          if (isIdling && socketException.SocketErrorCode == SocketError.WouldBlock) {
            // expected socket exception
            Trace.Log(this, "idling");
            continue;
          }
          else {
            // unexpected socket exception
            throw new ImapConnectionException(socketException.Message, socketException.InnerException);
          }
        }
      }

      // return received response
      if (0 < responseQueue.Count)
        return responseQueue.Dequeue();

      return null;
    }

    internal void ClearBuffer()
    {
      stream.ClearBuffer();
    }
#endregion

    public override string ToString()
    {
      CheckDisposed();
      CheckConnected();

      return string.Format("{{Host={0}, RemoteEndPoint={1}, LocalEndPoint={2}, Connected={3}}}",
                           host,
                           remoteEndPoint,
                           localEndPoint,
                           client.Client.Connected);
    }

    private void CheckDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().FullName);
    }

    private void CheckConnected()
    {
      if (client == null)
        throw new InvalidOperationException("not connected");
    }

    private readonly int id = 0;
    private string host = null;
    private IPEndPoint remoteEndPoint;
    private IPEndPoint localEndPoint;
    private bool disposed = false;
    private bool isIdling = false;
    private DateTime lastSentTime = DateTime.MinValue;
    private TcpClient client = null;
    private ImapBufferedStream stream = null;
    private int commandTag = 0;
    private ImapResponseReceiver receiver = null;
    private ImapCommandSender sender = null;
    private /*readonly*/ Queue<ImapResponse> responseQueue = new Queue<ImapResponse>();
  }
}
