// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Collections.Generic;
using System.Text;

namespace Smdn.Protocols.Imap4.Client.Protocol {
  public class ImapResponseReceiver {
    public bool ResponseContinuing {
      get { return dataResponseContinuing; }
    }

    public ImapResponseReceiver(ImapBufferedStream stream)
    {
      if (stream == null)
        throw new ArgumentNullException("stream");

      this.stream = stream;
      this.fragmentedResponseBuffer = new StringBuilder(stream.BufferSize);
    }

    private const string CRLF = "\x0d\x0a";
    private const char CR = '\x0d';
    private const char LF = '\x0a';
    private const char SP = ' ';
    private const char DQUOTE = '"';

    private ImapBufferedStream stream;
    private StringBuilder fragmentedResponseBuffer;

    private int requiredDataLengthAtLeast = 0;
    private bool dataResponseContinuing = false;

    public ImapResponse Receive()
    {
      string line;

      if (0 == stream.ReadToCrLf(requiredDataLengthAtLeast, out line))
        throw new ImapConnectionException("read failed");

      fragmentedResponseBuffer.Append(line);

      if (fragmentedResponseBuffer.Length < requiredDataLengthAtLeast)
        return null; // need to read more

      try {
        string notParsedLines = null;

        var response = (dataResponseContinuing)
          ? ParseDataResponse(fragmentedResponseBuffer.ToString(), out notParsedLines, out requiredDataLengthAtLeast)
          : ParseResponce(fragmentedResponseBuffer.ToString(), out notParsedLines, out requiredDataLengthAtLeast);

        fragmentedResponseBuffer.Length = 0; // discard buffered

        if (notParsedLines != null)
          fragmentedResponseBuffer.Append(notParsedLines);

        return response;
      }
      catch {
        fragmentedResponseBuffer.Length = 0; // discard buffered
        throw;
      }
    }

    private ImapResponse ParseResponce(string line, out string notParsedLines, out int requiredBytes)
    {
      notParsedLines = null;
      requiredBytes = 0;

      // response        = *(continue-req / response-data) response-done

      // continue-req    = "+" SP (resp-text / base64) CRLF
      if (0 == line.IndexOf("+ "))
        return new ImapCommandContinuationRequest(line.Substring(2, line.Length - 4), // remove leading "+" SP and trailing CRLF
                                                  line);

      // response-done   = response-tagged / response-fatal
      // response-data   = "*" SP (resp-cond-state / resp-cond-bye /
      //                   mailbox-data / message-data / capability-data) CRLF
      // response-fatal  = "*" SP resp-cond-bye CRLF
      //                     ; Server closes connection immediately
      // response-tagged = tag SP resp-cond-state CRLF

      // ("*" / tag) SP
      var tagSep = line.IndexOf(SP);

      if (tagSep == -1)
        // response-done and response-data must contain SP
        throw new ImapMalformedResponseException("malformed response-done/response-data", line);

      var tag = line.Substring(0, tagSep);

      // ("OK" / "BAD" / "NO" / "BYE" / "PREAUTH" / text) SP
      var respCondSep = line.IndexOf(' ', tagSep + 1);
      ImapResponseCondition? cond = null;

      if (respCondSep == -1) {
        if (tag == "*")
          // '* SEARCH\r\n' (mailbox-data which contains no SP)
          cond = null;
        else
          throw new ImapMalformedResponseException("malformed response-data", line);
      }
      else {
        cond = ParseCondition(line.Substring(tagSep + 1, respCondSep - tagSep - 1));
      }

      if (cond != null || 0 == line.IndexOf("* [")) {
        // resp-cond-auth / resp-cond-state / resp-cond-bye
        string respText;

        if (cond == null)
          respText = line.Substring(tagSep + 1);
        else
          respText = line.Substring(respCondSep + 1);

        var responseText = ParseRespText(respText);

        if (tag == "*")
          return new ImapUntaggedStatusResponse((ImapResponseCondition)cond,
                                            responseText,
                                            line);
        else
          return new ImapTaggedStatusResponse(tag,
                                          (ImapResponseCondition)cond,
                                          responseText,
                                          line);
      }

      // mailbox-data / message-data / capability-data etc.
      return ParseDataResponse(line, out notParsedLines, out requiredBytes);
    }

    private ImapResponseCondition? ParseCondition(string cond)
    {
      switch (cond.ToUpperInvariant()) {
        case "OK":      return ImapResponseCondition.Ok;
        case "BAD":     return ImapResponseCondition.Bad;
        case "NO":      return ImapResponseCondition.No;
        case "BYE":     return ImapResponseCondition.Bye;
        case "PREAUTH": return ImapResponseCondition.PreAuth;
        default: return null;
      }
    }

    private ImapResponseText ParseRespText(string respText)
    {
      // resp-text       = ["[" resp-text-code "]" SP] text
      if (respText[0] != '[')
        // no resp-text-code
        return new ImapResponseText(respText.Substring(0, respText.Length - 2)); // remove trailing CRLF

      // exists resp-text-code
      var respTextCodeEnd = respText.IndexOf(']');
      var respTextCode    = respText.Substring(1, respTextCodeEnd - 1);
      var codeSep         = respTextCode.IndexOf(SP);
      var text            = string.Empty;

      if (respTextCodeEnd < respText.Length) {
        text = respText.Substring(respTextCodeEnd + 1);
        text = text.Trim(); // remove SP, CRLF
      }

      if (codeSep == -1)
        // no arguments; "READ-ONLY" / "READ-WRITE" / etc...
        return new ImapResponseText(ImapResponseCode.GetKnownOrCreate(respTextCode), new ImapData[] {}, text);

      // exist arguments; "UIDNEXT" SP nz-number / "UIDVALIDITY" SP nz-number / etc...
      var codeString  = respTextCode.Substring(0, codeSep);
      var argsString  = respTextCode.Substring(codeSep + 1);

      ImapData[] arguments;
      int discard = 0;

      using (var reader = new DataReader(argsString)) {
        if (!ParseData(reader, out arguments, ref discard))
          throw new ImapMalformedResponseException("malformed resp-text-code", argsString);
      }

      return new ImapResponseText(ImapResponseCode.GetKnownOrCreate(codeString), arguments, text);
    }

    private ImapDataResponse ParseDataResponse(string lines, out string notParsedLines, out int requiredBytes)
    {
      notParsedLines = null;
      requiredBytes = 0;

      // mailbox-data / message-data / capability-data etc.

      // ("*" / tag) SP
      var tagSep = lines.IndexOf(SP);

      using (var dataReader = new DataReader(lines.Substring(tagSep + 1))) {
        ImapData[] parsed = null;

        if (ParseData(dataReader, out parsed, ref requiredBytes)) {
          dataResponseContinuing = false;
          requiredBytes = 0;
          notParsedLines = dataReader.ReadToEnd();
          return ImapDataResponse.Create(parsed, lines.Substring(0, lines.Length - notParsedLines.Length));
        }
        else {
          dataResponseContinuing = true;
          requiredBytes += tagSep + 1;
          notParsedLines = lines;
          return null;
        }
      }
    }

    private bool ParseData(DataReader dataReader, out ImapData[] data, ref int requiredBytes)
    {
      data = null;

      var parsed = new List<ImapData>();

      for (;;) {
        if (dataReader.IsEndOfData)
          break;

        var ch = dataReader.PeekChar();

        if (ch == SP || ch == CR) {
          dataReader.Read();
          continue;
        }

        if (ch == LF) {
          dataReader.Read();
          break;
        }

        ImapData d;
        var builder = new StringBuilder();

        if (ch == DQUOTE) {
          if (ParseQuotedString(dataReader, builder, false))
            d = ImapData.CreateTextData(builder.ToString());
          else
            return false;
        }
        else if (ch == '(') {
          if (ParseParenthesizedData(dataReader, builder, ref requiredBytes)) {
            using (var r = new DataReader(builder.ToString())) {
              ImapData[] list = null;

              if (!ParseData(r, out list, ref requiredBytes))
                return false;

              d = ImapData.CreateListData(list);
            }
          }
          else {
            return false;
          }
        }
        else if (ch == '{') {
          if (ParseLiteralString(dataReader, builder, false, ref requiredBytes))
            d = ImapData.CreateTextData(builder.ToString());
          else
            return false;
        }
        else {
          if (ParseNonQuotedText(dataReader, builder)) {
            var text = builder.ToString();

            if (string.Equals(text, "NIL", StringComparison.InvariantCultureIgnoreCase))
              d = ImapData.CreateNilData();
            else
              d = ImapData.CreateTextData(text);
          }
          else {
            return false;
          }
        }

        parsed.Add(d);
      }

      data = parsed.ToArray();

      return true;
    }

    private bool ParseQuotedString(DataReader reader, StringBuilder builder, bool keepDelimiter)
    {
      // quoted          = DQUOTE *QUOTED-CHAR DQUOTE
      // QUOTED-CHAR     = <any TEXT-CHAR except quoted-specials> / 
      //                   "\" quoted-specials
      // TEXT-CHAR       = <any CHAR except CR and LF>
      // CHAR            = <any 7-bit US-ASCII character except NUL,
      //                   0x01 - 0x7f>
      // quoted-specials = DQUOTE / "\"
      // date-time       = DQUOTE date-day-fixed "-" date-month "-" date-year

      var c = reader.ReadChar(); // DQUOTE

      if (keepDelimiter)
        builder.Append(c);

      var escape = false;

      for (;;) {
        if (reader.IsEndOfData)
          return false; // abort

        c = reader.ReadChar();

        escape = (c == '\\');

        if (escape) {
          if (keepDelimiter)
            builder.Append(c);
          c = reader.ReadChar(); // quoted-specials
        }

        if (c == CR || c == LF || (!escape && c == DQUOTE)) {
          if (keepDelimiter)
            builder.Append(c);

          return true; // success
        }
        else { // QUOTED-CHAR
          builder.Append(c);
        }
      }
    }

    private bool ParseLiteralString(DataReader reader, StringBuilder builder, bool keepDelimiter, ref int requiredBytes)
    {
      // literal         = "{" number "}" CRLF *CHAR8
      //                ; Number represents the number of CHAR8s
      // CHAR8           = %x01-ff ; any OCTET except NUL, %x00

      reader.Read(); // "{"

      var octetString = string.Empty;

      for (;;) {
        if (reader.IsEndOfData)
          return false; // abort

        var c = reader.ReadChar();

        if ('0' <= c && c <= '9')
          octetString += c; // number
        else if (c == '}')
          break; // "}"
        else
          throw new ImapMalformedResponseException("malformed literal (invalid number)");
      }

      if (keepDelimiter) {
        builder.Append('{');
        builder.Append(octetString);
        builder.Append('}');
      }

      var octet = int.Parse(octetString);

      // CRLF
      if (reader.IsEndOfData)
        return false; // abort
      else if (CR != reader.ReadChar())
        throw new ImapMalformedResponseException("malformed literal (missing CR)");

      if (keepDelimiter)
        builder.Append(CR);

      if (reader.IsEndOfData)
        return false; // abort
      else if (LF != reader.ReadChar())
        throw new ImapMalformedResponseException("malformed literal (missing LF)");

      if (keepDelimiter)
        builder.Append(LF);

      // *CHAR8
      if (octet <= reader.Length - reader.Position) {
        var buffer = new char[octet];

        reader.Read(buffer, 0, octet);

        builder.Append(buffer);

        return true; // success
      }
      else {
        var remain = octet - (reader.Length - reader.Position);

        if (requiredBytes < remain)
          requiredBytes = remain;

        return false; // abort
      }
    }

    private bool ParseParenthesizedData(DataReader reader, StringBuilder builder, ref int requiredBytes)
    {
      var nested = 0;

      reader.Read(); // "("

      for (;;) {
        if (reader.IsEndOfData)
          return false; // abort

        var c = reader.PeekChar();

        if (c == DQUOTE) {
          // quoted
          if (ParseQuotedString(reader, builder, true))
            continue;
          else
            return false;
        }
        else if (c == '{') {
          // literal
          if (ParseLiteralString(reader, builder, true, ref requiredBytes))
            continue;
          else
            return false;
        }

        c = reader.ReadChar();

        if (c == '(') {
          nested++;
        }
        else if (c == ')') {
          nested--;
          if (nested == -1) { // not nested
            return true; // success
          }
        }

        builder.Append(c);
      }
    }

    private bool ParseNonQuotedText(DataReader reader, StringBuilder builder)
    {
      // atom            = 1*ATOM-CHAR
      // ATOM-CHAR       = <any CHAR except atom-specials>
      // atom-specials   = "(" / ")" / "{" / SP / CTL / list-wildcards /
      //                   quoted-specials / resp-specials
      // number          = 1*DIGIT
      // list-wildcards  = "%" / "*"
      // quoted-specials = DQUOTE / "\"
      // resp-specials   = "]"

      // section         = "[" [section-spec] "]"
      // section-msgtext = "HEADER" / "HEADER.FIELDS" [".NOT"] SP header-list /
      //                   "TEXT"
      //                     ; top-level or MESSAGE/RFC822 part
      // section-part    = nz-number *("." nz-number)
      //                     ; body part nesting
      // section-spec    = section-msgtext / (section-part ["." section-text])
      // section-text    = section-msgtext / "MIME"
      //                     ; text other than actual body part (headers, etc.)
      var section = false;

      for (;;) {
        if (reader.IsEndOfData || reader.IsNextChar(CR) || reader.IsNextChar(LF))
          break;

        var ch = reader.ReadChar();

        if (!section && ch == SP) {
          break;
        }
        else {
          builder.Append(ch);

          if (ch == '[')
            section = true;
          else if (ch == ']')
            section = false;
        }
      }

      return true; // success
    }

#region "class DataReader"
    private class DataReader : TextReader {
      public int Position {
        get { return pos; }
      }

      public int Length {
        get { return len; }
      }

      public bool IsEndOfData {
        get { return len <= pos; }
      }

      public string Source {
        get { return source; }
      }

      public DataReader(string s)
      {
        if (s == null)
          throw new ArgumentNullException("s");

        source = s;
        len = s.Length;
        pos = 0;
      }

      public override void Close()
      {
        Dispose(true);
      }

      protected override void Dispose(bool disposing)
      {
        source = null;
        base.Dispose(disposing);
      }

      public override int Peek()
      {
        if (len <= pos)
          return -1;
        else
          return (int)source[pos];
      }

      public char PeekChar()
      {
        if (len <= pos)
          throw new EndOfStreamException();
        else
          return source[pos];
      }

      public override int Read()
      {
        if (len <= pos)
          return -1;
        else
          return (int)source[pos++];
      }

      public char ReadChar()
      {
        if (len <= pos)
          throw new EndOfStreamException();
        else
          return source[pos++];
      }

      public bool IsNextChar(char ch)
      {
        if (len <= pos)
          return false;
        else
          return source[pos] == ch;
      }

      public override int Read(char[] buffer, int index, int count)
      {
        if (source == null)
          throw new ObjectDisposedException(GetType().Name);
        if (buffer == null)
          throw new ArgumentNullException("buffer");
        if (index < 0)
          throw new ArgumentOutOfRangeException("index");
        if (count < 0)
          throw new ArgumentOutOfRangeException("count");
        if (buffer.Length - index < count)
          throw new ArgumentException();

        var lengthToRead = (len <= pos + count) ? len - pos : count;

        source.CopyTo(pos, buffer, index, count);

        pos += lengthToRead;

        return lengthToRead;
      }

      public override string ReadLine()
      {
        throw new NotImplementedException();
      }

      public override string ReadToEnd()
      {
        if (source == null)
          throw new ObjectDisposedException(GetType().Name);

        var substr = source.Substring(pos, len - pos);

        pos = len;

        return substr;
      }

      private string source;
      private readonly int len;
      private int pos;
    }
#endregion
  }
}