// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;
using System.Net;

using Threading = System.Threading;

using Smdn.Protocols.Imap4.Client.Protocol;
using Smdn.Protocols.Imap4.Client.Transaction;
using Smdn.Protocols.Imap4.Client.Transaction.BuiltIn;

namespace Smdn.Protocols.Imap4.Client.Session {
  // Layers:
  // * Session.ImapSession
  //   Transaction.IImapTransaction/Transaction.ImapTransactionRequest/Transaction.ImapTransactionResult
  //   Protocol.ImapConnection/Protocol.ImapCommand/Protocol.ImapResponse
  //   (IMAP4)

  public class ImapSession : IDisposable {
    /*
     * properties
     */
    internal int Id {
      get { return connection.Id; }
    }

    public ICredentialsByHost Credentials {
      get { return credentials; }
      set { credentials = value; }
    }

    public bool HandlesReferralAsException {
      get { return handlesReferralAsException; }
      set { handlesReferralAsException = value; }
    }

    public bool HandlesIncapableAsException {
      get { return handlesIncapableAsException; }
      set { handlesIncapableAsException = value; }
    }

    public int SendTimeout {
      get
      {
        CheckDisposed();
        return connection.Client.SendTimeout;
      }
      set
      {
        CheckDisposed();
        connection.Client.SendTimeout = value;
      }
    }

    public int ReceiveTimeout {
      get
      {
        CheckDisposed();
        return connection.Client.SendTimeout;
      }
      set
      {
        CheckDisposed();
        connection.Client.SendTimeout = value;
      }
    }

    public int TransactionTimeout {
      get
      {
        CheckDisposed();
        return transactionTimeout;
      }
      set
      {
        CheckDisposed();
        if (value < -1)
          throw new ArgumentOutOfRangeException("Timeout", "must be greater than or equals to -1");
        transactionTimeout = value;
      }
    }

    public bool IsTransactionProceeding {
      get
      {
        if (!transactionProceedingSemaphore.WaitOne(0)) {
          return true;
        }
        else {
          transactionProceedingSemaphore.Release();
          return false;
        }
      }
    }

    /*
     * construction / destruction
     */
    public ImapSession(string host)
      : this(host, ImapConnection.ImapDefaultPort, false, DefaultTransactionTimeout, null)
    {
    }

    public ImapSession(string host, int port)
      : this(host, port, false, DefaultTransactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, int transactionTimeout)
      : this(host, port, false, transactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, ImapUpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
      : this(host, port, false, DefaultTransactionTimeout, createAuthenticatedStreamCallback)
    {
    }

    public ImapSession(string host, int port, int transactionTimeout, ImapUpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
      : this(host, port, false, transactionTimeout, createAuthenticatedStreamCallback)
    {
    }

    public ImapSession(string host, bool handlesReferralAsException)
      : this(host, ImapConnection.ImapDefaultPort, handlesReferralAsException, DefaultTransactionTimeout, null)
    {
    }

    public ImapSession(string host, bool handlesReferralAsException, int transactionTimeout)
      : this(host, ImapConnection.ImapDefaultPort, handlesReferralAsException, transactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, bool handlesReferralAsException)
      : this(host, port, handlesReferralAsException, DefaultTransactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, bool handlesReferralAsException, int transactionTimeout)
      : this(host, port, handlesReferralAsException, transactionTimeout, null)
    {
    }

    public ImapSession(string host, int port, bool handlesReferralAsException, ImapUpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
      : this(host, port, handlesReferralAsException, DefaultTransactionTimeout, createAuthenticatedStreamCallback)
    {
    }

    public ImapSession(string host, int port, bool handlesReferralAsException, int transactionTimeout, ImapUpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      if (transactionTimeout < -1)
        throw new ArgumentOutOfRangeException("transactionTimeout", "must be greater than or equals to -1");

      this.handlesReferralAsException = handlesReferralAsException;
      this.transactionTimeout = transactionTimeout;
      this.mailboxManager = new ImapMailboxManager(this);

      InitializeResponseProcessors();

      Connect(host, port, createAuthenticatedStreamCallback);
    }

    public override string ToString()
    {
      CheckDisposed();

      return string.Format("{{Connection={0}, State={0}}}",
                           connection,
                           state);
    }

    private const int DefaultTransactionTimeout = Timeout.Infinite;

    private bool disposed = false;
    private ICredentialsByHost credentials = null;
    private bool handlesReferralAsException;
    private bool handlesIncapableAsException = false;
    private int transactionTimeout = DefaultTransactionTimeout;
    private Semaphore transactionProceedingSemaphore = new Semaphore(1, 1);

    /*
     * session state properties
     */
    internal ImapConnection Connection {
      get { return connection; }
    }

    internal ImapMailboxManager MailboxManager {
      get { return mailboxManager; }
    }

    public bool IsIdling {
      get
      {
        CheckDisposed();
        return connection.IsIdling;
      }
    }

    public ImapSessionState State {
      get
      {
        CheckDisposed();
        return state;
      }
    }

    public ImapUpdatedStatus LastUpdatedStatus {
      get
      {
        CheckDisposed();
        return lastUpdatedStatus;
      }
      internal set { lastUpdatedStatus = value; }
    }

    public ImapMailbox SelectedMailbox {
      get
      {
        CheckDisposed();
        return selectedMailbox;
      }
      internal set { selectedMailbox = value; }
    }

    public ImapUri ServerUrl {
      get
      {
        CheckDisposed();
        return serverUrl;
      }
    }

    public ImapCapabilityList ServerCapabilities {
      get
      {
        CheckDisposed();
        return serverCapabilities;
      }
      internal set { serverCapabilities = value; }
    }

    public ImapNamespace Namespaces {
      get
      {
        CheckDisposed();
        return namespaces;
      }
    }

    public Dictionary<string, string> ServerID {
      get
      {
        CheckDisposed();
        return serverID;
      }
    }

    private ImapConnection connection = null;
    private /*readonly*/ ImapMailboxManager mailboxManager;
    private ImapSessionState state = ImapSessionState.NotConnected;
    private ImapUpdatedStatus lastUpdatedStatus = new ImapUpdatedStatus();
    private ImapTransactionResult lastTransactionResult = null;
    private ImapMailbox selectedMailbox = null;
    private ImapUri serverUrl = null;
    private ImapCapabilityList serverCapabilities = new ImapCapabilityList(new[] {ImapCapability.Imap4Rev1 /* required by GreetingTransaction */});
    private ImapNamespace namespaces = new ImapNamespace();
    private Dictionary<string, string> serverID = new Dictionary<string, string>();

    /*
     * events
     */
    public event EventHandler<ImapStatusUpdatedEventArgs> StatusUpdated;

    private void StatusUpdatedHandler(object sender, ImapStatusUpdatedEventArgs e)
    {
      OnStatusUpdated(e);
    }

    private void OnStatusUpdated(ImapStatusUpdatedEventArgs e)
    {
      var statusUpdated = this.StatusUpdated;

      if (statusUpdated != null)
        statusUpdated(this, e);
    }

    /*
     * transaction methods : connect/disconnect
     */
    private void Connect(string host, int port, ImapUpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      this.connection = new ImapConnection(host, port, createAuthenticatedStreamCallback);

      using (var t = new GreetingTransaction(connection)) {
        if (ProcessTransaction(t, handlesIncapableAsException).Failed)
          throw new ImapException("initial greeting failed"); // XXX

        if (t.Result.Code == ImapTransactionResultCode.PreAuth)
          TransitStateTo(ImapSessionState.Authenticated);
        else
          TransitStateTo(ImapSessionState.NotAuthenticated);
      }

      TraceVerbose("connected");
      Trace.Info("CID:{0} session started", connection.Id);
    }

    /// <summary>disconnects session</summary>
    /// <remarks>this method sends CLOSE and LOGOUT before disconnect the session.</remarks>
    public void Disconnect()
    {
      Disconnect(true);
    }

    /// <summary>disconnects session</summary>
    public void Disconnect(bool logout)
    {
      CheckDisposed();

      if (logout) {
        if (state == ImapSessionState.Selected)
          Close();
        if (state == ImapSessionState.Authenticated)
          Logout();
      }

      (this as IDisposable).Dispose();
    }

    private void CloseConnection()
    {
      if (connection == null)
        return;

      try {
        Trace.Info("CID:{0} session closing", this.Id);

        connection.Close();

        Trace.Info("CID:{0} session closed", this.Id);
      }
      finally {
        TransitStateTo(ImapSessionState.NotConnected);

        connection = null;
      }
    }

    void IDisposable.Dispose()
    {
      if (disposed)
        return;

      CloseConnection();

      transactionProceedingSemaphore.Close();
      transactionProceedingSemaphore = null;

      disposed = true;
    }

    /*
     * transaction methods : any state
     */

    /// <summary>sends CAPABILITY command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult Capability()
    {
      ImapCapabilityList discard;

      return Capability(out discard);
    }

    /// <summary>sends CAPABILITY command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult Capability(out ImapCapabilityList capabilities)
    {
      RejectNonConnectedState();

      capabilities = null;

      using (var t = new CapabilityTransaction(connection)) {
        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          SetServerCapabilities(t.Result.Value);
          capabilities = t.Result.Value;
        }

        return t.Result;
      }
    }

    /// <summary>sends NOOP command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult NoOp()
    {
      RejectNonConnectedState();

      using (var t = new NoOpTransaction(connection)) {
        return ProcessTransaction(t, handlesIncapableAsException);
      }
    }

    /// <summary>sends LOGOUT command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult Logout()
    {
      CheckDisposed();

      if (state == ImapSessionState.NotConnected)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestDone, "already logged out or disconnected");

      using (var t = new LogoutTransaction(connection)) {
        if (ProcessTransaction(t, handlesIncapableAsException).Code == ImapTransactionResultCode.Ok) {
          TransitStateTo(ImapSessionState.NotAuthenticated);
          CloseConnection();
        }

        return t.Result;
      }
    }

    /// <summary>sends ID command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult ID(Dictionary<string, string> clientParameterList)
    {
      Dictionary<string, string> discard;

      return ID(clientParameterList, out discard);
    }

    /// <summary>sends ID command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult ID(Dictionary<string, string> clientParameterList, out Dictionary<string, string> serverParameterList)
    {
      RejectNonConnectedState();

      serverParameterList = null;

      using (var t = new IdTransaction(connection)) {
        // client parameter list
        if (clientParameterList != null && 0 < clientParameterList.Count) {
          // RFC 2971 IMAP4 ID extension
          // http://tools.ietf.org/html/rfc2971
          // 3.3. Defined Field Values
          //    Field strings MUST NOT be longer than 30 octets.  Value strings MUST
          //    NOT be longer than 1024 octets.  Implementations MUST NOT send more
          //    than 30 field-value pairs.
          if (30 < clientParameterList.Count)
            throw new ArgumentException("number of client parameter list pairs must be less than or equals to 30");

          var idParamsList = new List<ImapString>();

          try {
            foreach (var param in clientParameterList) {
              if (30 < ImapEncoding.Transfer7Bit.GetByteCount(param.Key))
                throw new ArgumentException("field string of client parameter list must be shorter than or equals to 30 octets");

              idParamsList.Add(new ImapQuotedString(param.Key));

              if (param.Value == null) {
                idParamsList.Add(new ImapNilString());
              }
              else {
                if (1024 < ImapEncoding.Transfer7Bit.GetByteCount(param.Value))
                  throw new ArgumentException("value string of client parameter list must be shorter than or equals to 1024 octets");
                idParamsList.Add(new ImapQuotedString(param.Value));
              }
            }
          }
          catch (EncoderFallbackException) {
            throw new ArgumentException("both field and value string must not contain non-ASCII chars");
          }

          t.Request.Arguments["client parameter list"] = new ImapParenthesizedString(idParamsList.ToArray());
        }

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          this.serverID = t.Result.Value;
          serverParameterList = t.Result.Value;
        }

        return t.Result;
      }
    }

    /// <summary>sends COMPRESS command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapTransactionResult Compress(ImapCompressionMechanism compressionMechanism)
    {
      RejectNonConnectedState();

      using (var t = new CompressTransaction(connection)) {
        t.Request.Arguments["compression mechanism"] = compressionMechanism;

        return ProcessTransaction(t, handlesIncapableAsException);
      }
    }

    /*
     * transaction methods : non-authenticated state
     */

    /// <summary>sends STARTTLS command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult StartTls(ImapUpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      return StartTls(createAuthenticatedStreamCallback, false);
    }

    /// <summary>sends STARTTLS command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult StartTls(ImapUpgradeConnectionStreamCallback createAuthenticatedStreamCallback, bool reissueCapability)
    {
      RejectNonConnectedState();

      if (state != ImapSessionState.NotAuthenticated)
        throw new InvalidOperationException("already authenticated");

      if (createAuthenticatedStreamCallback == null)
        throw new ArgumentNullException("createAuthenticatedStreamCallback");

      ImapTransactionResult result;

      using (var t = new StartTlsTransaction(connection, createAuthenticatedStreamCallback)) {
        if ((result = ProcessTransaction(t, handlesIncapableAsException)).Succeeded)
          // 6.2.1. STARTTLS Command
          //       Once [TLS] has been started, the client MUST discard cached
          //       information about server capabilities and SHOULD re-issue the
          //       CAPABILITY command.
          SetServerCapabilities(new ImapCapabilityList());
        else
          return result;
      }

      if (reissueCapability)
        Capability();

      return result;
    }

    /// <summary>sends AUTHENTICATE command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Authenticate(ImapAuthenticationMechanism authenticationMechanism)
    {
      return Authenticate(null, authenticationMechanism);
    }

    /// <summary>sends AUTHENTICATE command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Authenticate(string username, ImapAuthenticationMechanism authenticationMechanism)
    {
      RejectNonConnectedState();

      if (state != ImapSessionState.NotAuthenticated)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestDone, "already authenticated");

      if (authenticationMechanism == null)
        throw new ArgumentNullException("authenticationMechanism");

      var credential = GetCredential(username, authenticationMechanism);

      if (credential == null)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestError, string.Format("credential not found for {0};AUTH={1}@{2}:{3}", username, authenticationMechanism, connection.Host, connection.Port));

      using (var t = new AuthenticateTransaction(connection, credential)) {
        t.Request.Arguments["authentication mechanism name"] = authenticationMechanism;

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          UpdateServerUrl(credential.UserName, authenticationMechanism);
          TransitStateTo(ImapSessionState.Authenticated);
        }

        return t.Result;
      }
    }

    /// <summary>sends LOGIN command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Login()
    {
      return Login(null);
    }

    /// <summary>sends LOGIN command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapTransactionResult Login(string username)
    {
      RejectNonConnectedState();

      if (state != ImapSessionState.NotAuthenticated)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestDone, "already logged in");

      // RFC 2595 Using TLS with IMAP, POP3 and ACAP
      // http://tools.ietf.org/html/rfc2595
      // 3.2. IMAP LOGINDISABLED capability
      //    An IMAP client which complies with this specification MUST NOT issue
      //    the LOGIN command if this capability is present.
      if (serverCapabilities.Has(Imap4.ImapCapability.LoginDisabled))
        throw new ImapIncapableException("LOGIN is disabled");

      var credential = GetCredential(username, null);

      if (credential == null)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestError, string.Format("credential not found for {0}@{1}:{2}", username, connection.Host, connection.Port));

      using (var t = new LoginTransaction(connection, credential)) {
        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          UpdateServerUrl(credential.UserName, null);
          TransitStateTo(ImapSessionState.Authenticated);
        }

        return t.Result;
      }
    }

    private NetworkCredential GetCredential(string username, ImapAuthenticationMechanism mechanism)
    {
      if (credentials == null)
        return null;

      var credential = credentials.GetCredential(connection.Host, connection.Port, (mechanism == null) ? string.Empty : (string)mechanism);

      if (credential == null || (!string.IsNullOrEmpty(username) && credential.UserName != username))
        return null;
      else
        return credential;
    }

    internal void UpdateServerUrl(string username, ImapAuthenticationMechanism authType)
    {
      // TODO: use ImapUriBuilder
      var scheme = connection.IsSslConnection ? ImapUri.UriSchemeImaps : ImapUri.UriSchemeImap;
      var isDefaultPort = connection.IsSslConnection
        ? (connection.Port == ImapConnection.ImapsDefaultPort)
        : (connection.Port == ImapConnection.ImapDefaultPort);
      var hostPort = isDefaultPort ? connection.Host : string.Format("{0}:{1}", connection.Host, connection.Port);

      serverUrl = (username == null)
        ? new ImapUri(string.Format("{0}://{1}/", scheme, hostPort))
        : (authType == null)
          ? new ImapUri(string.Format("{0}://{1}@{2}/", scheme, username, hostPort))
          : new ImapUri(string.Format("{0}://{1};AUTH={2}@{3}/", scheme, username, authType, hostPort));
    }

    /*
     * transaction methods : authenticated state
     */

    /// <summary>sends SELECT command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Select(ImapMailbox mailbox)
    {
      return Select(mailbox.Name);
    }

    /// <summary>sends SELECT command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Select(string mailboxName)
    {
      using (var t = new SelectTransaction(connection, null)) {
        return SelectExamineInternal(t, mailboxName, null);
      }
    }

    /// <summary>sends SELECT command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    public ImapTransactionResult SelectCondStore(ImapMailbox mailbox)
    {
      return SelectCondStore(mailbox.Name);
    }

    /// <summary>sends SELECT command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    public ImapTransactionResult SelectCondStore(string mailboxName)
    {
      using (var t = new SelectTransaction(connection, ImapCapability.CondStore)) {
        return SelectExamineInternal(t, mailboxName, new ImapParenthesizedString("CONDSTORE"));
      }
    }

    /// <summary>sends EXAMINE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Examine(ImapMailbox mailbox)
    {
      return Examine(mailbox.Name);
    }

    /// <summary>sends EXAMINE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Examine(string mailboxName)
    {
      using (var t = new ExamineTransaction(connection, null)) {
        return SelectExamineInternal(t, mailboxName, null);
      }
    }


    /// <summary>sends EXAMINE command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    public ImapTransactionResult ExamineCondStore(ImapMailbox mailbox)
    {
      return ExamineCondStore(mailbox.Name);
    }

    /// <summary>sends EXAMINE command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    public ImapTransactionResult ExamineCondStore(string mailboxName)
    {
      using (var t = new ExamineTransaction(connection, ImapCapability.CondStore)) {
        return SelectExamineInternal(t, mailboxName, new ImapParenthesizedString("CONDSTORE"));
      }
    }

    private ImapTransactionResult SelectExamineInternal(SelectTransactionBase t, string mailboxName, ImapParenthesizedString selectParameters)
    {
      RejectNonAuthenticatedState();

      var selectingMailbox = mailboxManager.GetExist(mailboxName);
      var selectingMailboxExists = (selectingMailbox != null);

      if (selectingMailboxExists) {
        if (selectedMailbox == selectingMailbox) {
          return new ImapTransactionResult(ImapTransactionResultCode.RequestDone, "mailbox has already been selected");
        }
        else if (selectingMailbox.Flags != null &&
            selectingMailbox.Flags.Has(ImapMailboxFlag.NoSelect)) {
          throw new InvalidOperationException("mailbox is not selectable");
        }
      }
      else {
        selectingMailbox = new ImapMailbox(mailboxName);
      }

      /*
       * RFC 4466 - Collected Extensions to IMAP4 ABNF
       * http://tools.ietf.org/html/rfc4466
       *
       *    examine         = "EXAMINE" SP mailbox [select-params]
       *                      ;; modifies the original IMAP EXAMINE command
       *                      ;; to accept optional parameters
       *    select          = "SELECT" SP mailbox [select-params]
       *                      ;; modifies the original IMAP SELECT command to
       *                      ;; accept optional parameters
       *    select-params   = SP "(" select-param *(SP select-param) ")"
       *    select-param    = select-param-name [SP select-param-value]
       *                      ;; a parameter to SELECT may contain one or
       *                      ;; more atoms and/or strings and/or lists.
       *    select-param-name= tagged-ext-label
       *    select-param-value= tagged-ext-val
       *                      ;; This non-terminal shows recommended syntax
       *                      ;; for future extensions
       */
      t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

      // select-params
      if (selectParameters != null)
        t.Request.Arguments["select parameters"] = selectParameters;

      selectedMailbox = selectingMailbox; // set temporarily (to update size and status)

      if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
        if (selectingMailboxExists)
          selectedMailbox = selectingMailbox;
        else
          selectedMailbox = mailboxManager.Add(selectingMailbox);

        TransitStateTo(ImapSessionState.Selected);
      }
      else {
        selectedMailbox = null;
        if (state != ImapSessionState.NotConnected)
          TransitStateTo(ImapSessionState.Authenticated);
      }

      return t.Result;
    }

    /// <summary>sends CREATE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Create(string mailboxName, out ImapMailbox createdMailbox)
    {
      RejectNonAuthenticatedState();

      createdMailbox = null;

      using (var t = new CreateTransaction(connection)) {
        /*
         * RFC 4466 - Collected Extensions to IMAP4 ABNF
         * http://tools.ietf.org/html/rfc4466
         *
         *    create          = "CREATE" SP mailbox
         *                      [create-params]
         *                      ;; Use of INBOX gives a NO error.
         *    create-params   = SP "(" create-param *( SP create-param) ")"
         *    create-param-name = tagged-ext-label
         *    create-param      = create-param-name [SP create-param-value]
         *    create-param-value= tagged-ext-val
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions.
         */
        t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

        // create-params
        //t.Request.Arguments["create parameters"] = new ImapParenthesizedString();

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
          createdMailbox = mailboxManager.Add(new ImapMailbox(mailboxName));

        return t.Result;
      }
    }

    /// <summary>sends DELETE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Delete(ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      return Delete(mailbox.Name);
    }

    /// <summary>sends DELETE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Delete(string mailboxName)
    {
      RejectNonAuthenticatedState();

      // disallow deleting selected mailbox
      if (selectedMailbox != null && selectedMailbox == mailboxManager.GetExist(mailboxName))
        throw new InvalidOperationException("close or unselect mailbox first");

      using (var t = new DeleteTransaction(connection)) {
        t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
          mailboxManager.Delete(mailboxName);

        return t.Result;
      }
    }

    /// <summary>sends RENAME command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Rename(ImapMailbox existingMailbox, string newMailboxName)
    {
      ValidateMailboxRelationship(existingMailbox);

      ImapMailbox discard;

      return Rename(existingMailbox.Name, newMailboxName, out discard);
    }

    /// <summary>sends RENAME command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Rename(ImapMailbox existingMailbox, string newMailboxName, out ImapMailbox renamedMailbox)
    {
      ValidateMailboxRelationship(existingMailbox);

      return Rename(existingMailbox.Name, newMailboxName, out renamedMailbox);
    }

    /// <summary>sends RENAME command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Rename(string existingMailboxName, string newMailboxName)
    {
      ImapMailbox discard;

      return Rename(existingMailboxName, newMailboxName, out discard);
    }

    /// <summary>sends RENAME command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Rename(string existingMailboxName, string newMailboxName, out ImapMailbox renamedMailbox)
    {
      RejectNonAuthenticatedState();

      // disallow renaming selected mailbox
      if (selectedMailbox != null && selectedMailbox == mailboxManager.GetExist(existingMailboxName))
        throw new InvalidOperationException("close or unselect mailbox first");

      renamedMailbox = null;

      using (var t = new RenameTransaction(connection)) {
        /*
         * RFC 4466 - Collected Extensions to IMAP4 ABNF
         * http://tools.ietf.org/html/rfc4466
         *
         *    rename          = "RENAME" SP mailbox SP mailbox
         *                      [rename-params]
         *                      ;; Use of INBOX as a destination gives
         *                      ;; a NO error, unless rename-params
         *                      ;; is not empty.
         *    rename-params     = SP "(" rename-param *( SP rename-param) ")"
         *    rename-param      = rename-param-name [SP rename-param-value]
         *    rename-param-name = tagged-ext-label
         *    rename-param-value= tagged-ext-val
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions.
         */
        t.Request.Arguments["existing mailbox name"] = new ImapMailboxNameString(existingMailboxName);
        t.Request.Arguments["new mailbox name"] = new ImapMailboxNameString(newMailboxName);

        // rename-params
        //t.Request.Arguments["rename parameters"] = new ImapParenthesizedString();

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
          renamedMailbox = mailboxManager.Rename(existingMailboxName, newMailboxName);

        return t.Result;
      }
    }

    /// <summary>sends SUBSCRIBE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Subscribe(ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      return Subscribe(mailbox.Name);
    }

    /// <summary>sends SUBSCRIBE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Subscribe(string mailboxName)
    {
      RejectNonAuthenticatedState();

      using (var t = new SubscribeTransaction(connection)) {
        t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

        return ProcessTransaction(t, handlesIncapableAsException);
      }
    }

    /// <summary>sends UNSUBSCRIBE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Unsubscribe(ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      return Unsubscribe(mailbox.Name);
    }

    /// <summary>sends UNSUBSCRIBE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Unsubscribe(string mailboxName)
    {
      RejectNonAuthenticatedState();

      using (var t = new UnsubscribeTransaction(connection)) {
        t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

        return ProcessTransaction(t, handlesIncapableAsException);
      }
    }

    /// <summary>sends LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult List(out ImapMailbox[] mailboxes)
    {
      return List(null, null, out mailboxes);
    }

    /// <summary>sends LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult List(string mailboxName, out ImapMailbox[] mailboxes)
    {
      return List(null, mailboxName, out mailboxes);
    }

    /// <summary>sends LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult List(string referenceName, string mailboxName, out ImapMailbox[] mailboxes)
    {
      using (var t = new ListTransaction(connection)) {
        return ListLsubInternal(t, referenceName, mailboxName, out mailboxes);
      }
    }

    /// <summary>sends LSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Lsub(out ImapMailbox[] mailboxes)
    {
      return Lsub(null, null, out mailboxes);
    }

    /// <summary>sends LSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Lsub(string mailboxName, out ImapMailbox[] mailboxes)
    {
      return Lsub(null, mailboxName, out mailboxes);
    }

    /// <summary>sends LSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Lsub(string referenceName, string mailboxName, out ImapMailbox[] mailboxes)
    {
      using (var t = new LsubTransaction(connection)) {
        return ListLsubInternal(t, referenceName, mailboxName, out mailboxes);
      }
    }

    /// <summary>sends RLIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult RList(out ImapMailbox[] mailboxes)
    {
      return RList(null, null, out mailboxes);
    }

    /// <summary>sends RLIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult RList(string mailboxName, out ImapMailbox[] mailboxes)
    {
      return RList(null, mailboxName, out mailboxes);
    }

    /// <summary>sends RLIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult RList(string referenceName, string mailboxName, out ImapMailbox[] mailboxes)
    {
      using (var t = new RListTransaction(connection)) {
        return ListLsubInternal(t, referenceName, mailboxName, out mailboxes);
      }
    }

    /// <summary>sends RLSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult RLsub(out ImapMailbox[] mailboxes)
    {
      return RLsub(null, null, out mailboxes);
    }

    /// <summary>sends RLSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult RLsub(string mailboxName, out ImapMailbox[] mailboxes)
    {
      return RLsub(null, mailboxName, out mailboxes);
    }

    /// <summary>sends RLSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult RLsub(string referenceName, string mailboxName, out ImapMailbox[] mailboxes)
    {
      using (var t = new RLsubTransaction(connection)) {
        return ListLsubInternal(t, referenceName, mailboxName, out mailboxes);
      }
    }

    private ImapTransactionResult ListLsubInternal(ListTransactionBase t, string referenceName, string mailboxName, out ImapMailbox[] mailboxes)
    {
      RejectNonAuthenticatedState();

      mailboxes = null;

      if (referenceName == null)
        t.Request.Arguments["reference name"] = new ImapQuotedString(string.Empty);
      else
        t.Request.Arguments["reference name"] = new ImapQuotedString(referenceName);

      if (mailboxName == null)
        t.Request.Arguments["mailbox name"] = "*";
      else
        t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

      if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
        mailboxes = Array.ConvertAll(t.Result.Value, delegate(ImapMailboxList list) {
          return mailboxManager.GetExistOrCreate(list);
        });

      return t.Result;
    }

    /// <summary>sends STATUS command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Status(ImapMailbox mailbox, ImapStatusDataItem statusDataItem)
    {
      ValidateMailboxRelationship(mailbox);

      return Status(mailbox.Name, statusDataItem);
    }

    /// <summary>sends STATUS command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Status(string mailboxName, ImapStatusDataItem statusDataItem)
    {
      ImapStatusAttributeList discard;

      return Status(mailboxName, statusDataItem, out discard);
    }

    /// <summary>sends STATUS command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Status(string mailboxName, ImapStatusDataItem statusDataItem, out ImapStatusAttributeList statusAttributes)
    {
      RejectNonAuthenticatedState();

      statusAttributes = null;

      var statusMailbox = mailboxManager.GetExistOrCreate(mailboxName);

      using (var t = new StatusTransaction(connection)) {
        // mailbox name
        t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

        // status data item names
        t.Request.Arguments["status data item names"] = statusDataItem;

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          statusAttributes = t.Result.Value;
          statusMailbox.UpdateStatus(statusAttributes);
        }

        return t.Result;
      }
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Append(ImapMessage message, ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      ImapAppendedUidSet discard;
      ImapMailbox discard2;

      return AppendInternal(new[] {message}, mailbox.Name, false, out discard, out discard2);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// </remarks>
    public ImapTransactionResult AppendMultiple(IEnumerable<ImapMessage> messages, ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      ImapAppendedUidSet discard;
      ImapMailbox discard2;

      return AppendInternal(messages, mailbox.Name, false, out discard, out discard2);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult Append(ImapMessage message, ImapMailbox mailbox, out ImapAppendedUidSet appendedUids)
    {
      ValidateMailboxRelationship(mailbox);

      ImapMailbox discard;

      return AppendInternal(new[] {message}, mailbox.Name, false, out appendedUids, out discard);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult AppendMultiple(IEnumerable<ImapMessage> messages, ImapMailbox mailbox, out ImapAppendedUidSet appendedUids)
    {
      ValidateMailboxRelationship(mailbox);

      ImapMailbox discard;

      return AppendInternal(messages, mailbox.Name, false, out appendedUids, out discard);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Append(ImapMessage message, string mailboxName)
    {
      ImapAppendedUidSet discard;
      ImapMailbox discard2;

      return AppendInternal(new[] {message}, mailboxName, false, out discard, out discard2);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// </remarks>
    public ImapTransactionResult AppendMultiple(IEnumerable<ImapMessage> messages, string mailboxName)
    {
      ImapAppendedUidSet discard;
      ImapMailbox discard2;

      return AppendInternal(messages, mailboxName, false, out discard, out discard2);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult Append(ImapMessage message, string mailboxName, out ImapAppendedUidSet appendedUids)
    {
      ImapMailbox discard;

      return AppendInternal(new[] {message}, mailboxName, false, out appendedUids, out discard);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult AppendMultiple(IEnumerable<ImapMessage> messages, string mailboxName, out ImapAppendedUidSet appendedUids)
    {
      ImapMailbox discard;

      return AppendInternal(messages, mailboxName, false, out appendedUids, out discard);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// </remarks>
    public ImapTransactionResult Append(ImapMessage message, string mailboxName, out ImapMailbox createdMailbox)
    {
      ImapAppendedUidSet discard;

      return AppendInternal(new[] {message}, mailboxName, true, out discard, out createdMailbox);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// </remarks>
    public ImapTransactionResult AppendMultiple(IEnumerable<ImapMessage> messages, string mailboxName, out ImapMailbox createdMailbox)
    {
      ImapAppendedUidSet discard;

      return AppendInternal(messages, mailboxName, true, out discard, out createdMailbox);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult Append(ImapMessage message, string mailboxName, out ImapAppendedUidSet appendedUids, out ImapMailbox createdMailbox)
    {
      return AppendInternal(new[] {message}, mailboxName, true, out appendedUids, out createdMailbox);
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult AppendMultiple(IEnumerable<ImapMessage> messages, string mailboxName, out ImapAppendedUidSet appendedUids, out ImapMailbox createdMailbox)
    {
      return AppendInternal(messages, mailboxName, true, out appendedUids, out createdMailbox);
    }

    private ImapTransactionResult AppendInternal(IEnumerable<ImapMessage> messages, string mailboxName, bool tryCreate, out ImapAppendedUidSet appendedUids, out ImapMailbox createdMailbox)
    {
      RejectNonAuthenticatedState();

      appendedUids = null;
      createdMailbox = null;

      // append message
      var messagesToUpload = new List<ImapString>();
      var messageCount = 0;

      foreach (var message in messages) {
        /*
         * RFC 4466 - Collected Extensions to IMAP4 ABNF
         * http://tools.ietf.org/html/rfc4466
         * 
         *    append          = "APPEND" SP mailbox 1*append-message
         *                      ;; only a single append-message may appear
         *                      ;; if MULTIAPPEND [MULTIAPPEND] capability
         *                      ;; is not present
         *    append-message  = append-opts SP append-data
         *    append-ext      = append-ext-name SP append-ext-value
         *                      ;; This non-terminal define extensions to
         *                      ;; to message metadata.
         *    append-ext-name = tagged-ext-label
         *    append-ext-value= tagged-ext-val
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions.
         *    append-data     = literal / literal8 / append-data-ext
         *    append-data-ext = tagged-ext
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions,
         *                      ;; i.e., a mandatory label followed
         *                      ;; by parameters.
         *    append-opts     = [SP flag-list] [SP date-time] *(SP append-ext)
         *                      ;; message metadata
         */

        // flag-list
        if (message.Flags != null)
          messagesToUpload.Add(message.Flags.GetNonApplicableFlagsRemoved().ToParenthesizedString());

        // date-time
        if (message.InternalDate.HasValue)
          messagesToUpload.Add(ImapDateTimeFormat.ToDateTimeString(message.InternalDate.Value));

        // append-data
        messagesToUpload.Add(new ImapLiteralString(message.Body));

        messageCount++;
      }

      ImapTransactionResult failedResult = null;

      for (var i = 0; i < 2; i++) {
        var respTryCreate = false;

        using (var t = new AppendTransaction(connection, 1 < messageCount)) {
          // mailbox name
          t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

          // messages to upload
          t.Request.Arguments["messages to upload"] = new ImapStringList(messagesToUpload.ToArray());

          if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
            appendedUids = t.Result.Value;
            return t.Result;
          }
          else if (!tryCreate) {
            return t.Result;
          }

          failedResult = t.Result;

          // 6.3.11. APPEND Command
          //       If the destination mailbox does not exist, a server MUST return an
          //       error, and MUST NOT automatically create the mailbox.  Unless it
          //       is certain that the destination mailbox can not be created, the
          //       server MUST send the response code "[TRYCREATE]" as the prefix of
          //       the text of the tagged NO response.  This gives a hint to the
          //       client that it can attempt a CREATE command and retry the APPEND
          //       if the CREATE is successful.
          var tryCreateRespCode = t.Result.GetResponseCode(ImapResponseCode.TryCreate);

          respTryCreate = (tryCreateRespCode != null && (tryCreateRespCode is ImapTaggedStatusResponse));
        }

        // try create
        if (i == 0 && respTryCreate)
          if (Create(mailboxName, out createdMailbox).Failed)
            return failedResult;
      }

      return failedResult;
    }

    /// <summary>begins to send APPEND command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public IAsyncResult BeginAppend(Stream messageBodyStream, DateTimeOffset? internalDate, ImapMessageFlagList flags, ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      return BeginAppend(messageBodyStream, internalDate, flags, mailbox.Name);
    }

    /// <summary>begins to send APPEND command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public IAsyncResult BeginAppend(Stream messageBodyStream, DateTimeOffset? internalDate, ImapMessageFlagList flags, string mailboxName)
    {
      RejectNonAuthenticatedState();

      // append message
      var messagesToUpload = new List<ImapString>();

      // flag-list
      if (flags != null)
        messagesToUpload.Add(flags.GetNonApplicableFlagsRemoved().ToParenthesizedString());

      // date-time
      if (internalDate.HasValue)
        messagesToUpload.Add(ImapDateTimeFormat.ToDateTimeString(internalDate.Value));

      // append-data
      messagesToUpload.Add(new ImapLiteralStream(messageBodyStream));

      AppendTransaction t = null;

      try {
        t = new AppendTransaction(connection, false);

        // mailbox name
        t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

        // messages to upload
        t.Request.Arguments["messages to upload"] = new ImapStringList(messagesToUpload.ToArray());

        return BeginProcessTransaction(t, handlesIncapableAsException);
      }
      catch {
        if (t != null) {
          t.Dispose();
          t = null;
        }

        throw;
      }
    }

    /// <summary>ends to send APPEND command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult EndAppend(IAsyncResult asyncResult, out ImapAppendedUidSet appendedUid)
    {
      appendedUid = null;

      var appendAsyncResult = asyncResult as TransactionAsyncResult;

      if (appendAsyncResult == null)
        throw new ArgumentException("invalid IAsyncResult", "asyncResult");

      using (var t = appendAsyncResult.Transaction as AppendTransaction) {
        if (EndProcessTransaction(appendAsyncResult).Succeeded)
          appendedUid = t.Result.Value;

        return t.Result;
      }
    }

    /// <summary>sends ENABLE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Enable(params string[] capabilityNames)
    {
      ImapCapabilityList discard;

      return Enable(out discard, capabilityNames);
    }

    /// <summary>sends ENABLE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapTransactionResult Enable(out ImapCapabilityList enabledCapabilities, params string[] capabilityNames)
    {
      RejectNonAuthenticatedState();

      enabledCapabilities = null;

      using (var t = new EnableTransaction(connection)) {
        // capability names
        t.Request.Arguments["capability names"] = new ImapStringList(capabilityNames);

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
          enabledCapabilities = t.Result.Value;

        return t.Result;
      }
    }

    /*
     * transaction methods : selected state
     */

    /// <summary>sends CHECK command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Check()
    {
      RejectNonSelectedState();

      using (var t = new CheckTransaction(connection)) {
        return ProcessTransaction(t, handlesIncapableAsException);
      }
    }

    /// <summary>sends CLOSE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Close()
    {
      RejectNonAuthenticatedState();

      if (state != ImapSessionState.Selected)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestDone, "mailbox has already closed or has not selected yet");

      using (var t = new CloseTransaction(connection)) {
        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          selectedMailbox = null;
          TransitStateTo(ImapSessionState.Authenticated);
        }

        return t.Result;
      }
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Expunge()
    {
      long[] discard;

      return Expunge(out discard);
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Expunge(out long[] expungedMessages)
    {
      using (var t = new ExpungeTransaction(connection, false)) {
        return ExpungeInternal(t, null, out expungedMessages);
      }
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidExpunge(ImapSequenceSet uidSet)
    {
      long[] discard;

      return UidExpunge(uidSet, out discard);
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidExpunge(ImapSequenceSet uidSet, out long[] expungedMessages)
    {
      if (uidSet == null)
        throw new ArgumentNullException("uidSet");
      if (!uidSet.IsUidSet)
        throw new ArgumentException("sequence set is not UID");

      using (var t = new ExpungeTransaction(connection, true)) {
        return ExpungeInternal(t, uidSet, out expungedMessages);
      }
    }

    private ImapTransactionResult ExpungeInternal(ExpungeTransaction t, ImapSequenceSet uidSet, out long[] expungedMessages)
    {
      RejectNonSelectedState();

      expungedMessages = null;

      // sequence set for UID EXPUNGE
      if (uidSet != null)
        t.Request.Arguments["sequence set"] = uidSet.ToString();

      if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
        expungedMessages = t.Result.Value;

      return t.Result;
    }

    /// <summary>sends SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Search(ImapSearchCriteria searchingCriteria, out ImapMatchedSequenceSet matched)
    {
      return Search(searchingCriteria, Encoding.UTF8, out matched);
    }

    /// <summary>sends SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Search(ImapSearchCriteria searchingCriteria, Encoding charset, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SearchTransaction(connection, false, null)) {
        return SearchSortInternal(t, null, searchingCriteria, charset, null, out matched);
      }
    }

    /// <summary>sends SEARCH command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESEARCH extension.
    /// </remarks>
    public ImapTransactionResult ESearch(ImapSearchCriteria searchingCriteria, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      return ESearch(searchingCriteria, Encoding.UTF8, resultOptions, out matched);
    }

    /// <summary>sends SEARCH command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESEARCH extension.
    /// </remarks>
    public ImapTransactionResult ESearch(ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SearchTransaction(connection, false, ImapCapability.ESearch)) {
        return ESearchESortInternal(t, false, null, searchingCriteria, charset, resultOptions, out matched);
      }
    }

    /// <summary>sends UID SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidSearch(ImapSearchCriteria searchingCriteria, out ImapMatchedSequenceSet matched)
    {
      return UidSearch(searchingCriteria, Encoding.UTF8, out matched);
    }

    /// <summary>sends UID SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidSearch(ImapSearchCriteria searchingCriteria, Encoding charset, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SearchTransaction(connection, true, null)) {
        return SearchSortInternal(t, null, searchingCriteria, charset, null, out matched);
      }
    }

    /// <summary>sends SEARCH command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESEARCH extension.
    /// </remarks>
    public ImapTransactionResult UidESearch(ImapSearchCriteria searchingCriteria, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      return UidESearch(searchingCriteria, Encoding.UTF8, resultOptions, out matched);
    }

    /// <summary>sends SEARCH command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESEARCH extension.
    /// </remarks>
    public ImapTransactionResult UidESearch(ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SearchTransaction(connection, true, ImapCapability.ESearch)) {
        return ESearchESortInternal(t, true, null, searchingCriteria, charset, resultOptions, out matched);
      }
    }

    /// <summary>sends SORT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Sort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, out ImapMatchedSequenceSet matched)
    {
      return Sort(sortingCriteria, searchingCriteria, Encoding.UTF8, out matched);
    }

    /// <summary>sends SORT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Sort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SortTransaction(connection, false, null)) {
        return SearchSortInternal(t, sortingCriteria, searchingCriteria, charset, null, out matched);
      }
    }

    /// <summary>sends SORT command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESORT extension.
    /// </remarks>
    public ImapTransactionResult ESort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      return ESort(sortingCriteria, searchingCriteria, Encoding.UTF8, resultOptions, out matched);
    }

    /// <summary>sends SORT command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESORT extension.
    /// </remarks>
    public ImapTransactionResult ESort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SortTransaction(connection, false, ImapCapability.ESort)) {
        return ESearchESortInternal(t, false, sortingCriteria, searchingCriteria, charset, resultOptions, out matched);
      }
    }

    /// <summary>sends UID SORT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidSort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, out ImapMatchedSequenceSet matched)
    {
      return UidSort(sortingCriteria, searchingCriteria, Encoding.UTF8, out matched);
    }

    /// <summary>sends UID SORT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidSort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SortTransaction(connection, true, null)) {
        return SearchSortInternal(t, sortingCriteria, searchingCriteria, charset, null, out matched);
      }
    }

    /// <summary>sends UID SORT command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESORT extension.
    /// </remarks>
    public ImapTransactionResult UidESort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      return UidESort(sortingCriteria, searchingCriteria, Encoding.UTF8, resultOptions, out matched);
    }

    /// <summary>sends UID SORT command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESORT extension.
    /// </remarks>
    public ImapTransactionResult UidESort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SortTransaction(connection, true, ImapCapability.ESort)) {
        return ESearchESortInternal(t, true, sortingCriteria, searchingCriteria, charset, resultOptions, out matched);
      }
    }

    private ImapTransactionResult ESearchESortInternal(SearchTransactionBase t, bool uid, ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      /*
       * RFC 5182 - IMAP Extension for Referencing the Last SEARCH Result
       * http://tools.ietf.org/html/rfc5182
       */
      var saveResult = resultOptions.ContainsOneOf(ImapSearchResultOptions.Save);

      // ESEARCH or ESORT capability
      if (t is SearchTransaction) {
        if (!resultOptions.RequiredCapabilities.Contains(ImapCapability.ESearch))
          resultOptions.RequiredCapabilities.Add(ImapCapability.ESearch);
        if (resultOptions.RequiredCapabilities.Contains(ImapCapability.ESort))
          resultOptions.RequiredCapabilities.Remove(ImapCapability.ESort);
      }
      else if (t is SortTransaction) {
        if (!resultOptions.RequiredCapabilities.Contains(ImapCapability.ESort))
          resultOptions.RequiredCapabilities.Add(ImapCapability.ESort);
        if (resultOptions.RequiredCapabilities.Contains(ImapCapability.ESearch))
          resultOptions.RequiredCapabilities.Remove(ImapCapability.ESearch);
      }

      SearchSortInternal(t, sortingCriteria, searchingCriteria, charset, resultOptions, out matched);

      if (saveResult) {
        var saved = t.Result.Succeeded && (t.Result.GetResponseCode(ImapResponseCode.NotSaved) == null);

        if (saved) {
          if (matched == null)
            matched = new ImapMatchedSequenceSet(ImapSequenceSet.CreateSet(uid, new long[] {}));
          matched.IsSavedResult = true;
        }
      }

      return t.Result;
    }

    private ImapTransactionResult SearchSortInternal(SearchTransactionBase t, ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      RejectNonSelectedState();

      matched = null;

      // charset specification
      var containsLiteral = false;

      // TODO: charset
      // http://tools.ietf.org/html/rfc5256
      // BASE.6.4.SORT. SORT Command
      //   The US-ASCII and [UTF-8] charsets MUST be implemented.
      searchingCriteria.Traverse(delegate(ImapString criterion) {
        if (criterion is ImapLiteralString) {
          containsLiteral = true;
          (criterion as ImapLiteralString).Charset = charset;
        }
      });

      /*
       * RFC 4466 - Collected Extensions to IMAP4 ABNF
       * http://tools.ietf.org/html/rfc4466
       *
       *    search          = "SEARCH" [search-return-opts]
       *                      SP search-program
       *    search-correlator  = SP "(" "TAG" SP tag-string ")"
       *    search-program     = ["CHARSET" SP charset SP]
       *                         search-key *(SP search-key)
       *                         ;; CHARSET argument to SEARCH MUST be
       *                         ;; registered with IANA.
       *    search-return-data = search-modifier-name SP search-return-value
       *                         ;; Note that not every SEARCH return option
       *                         ;; is required to have the corresponding
       *                         ;; ESEARCH return data.
       *    search-return-opts = SP "RETURN" SP "(" [search-return-opt
       *                         *(SP search-return-opt)] ")"
       *    search-return-opt = search-modifier-name [SP search-mod-params]
       *    search-return-value = tagged-ext-val
       *                         ;; Data for the returned search option.
       *                         ;; A single "nz-number"/"number" value
       *                         ;; can be returned as an atom (i.e., without
       *                         ;; quoting).  A sequence-set can be returned
       *                         ;; as an atom as well.
       *    search-modifier-name = tagged-ext-label
       *    search-mod-params = tagged-ext-val
       *                      ;; This non-terminal shows recommended syntax
       *                      ;; for future extensions.
       */
      if (t is SortTransaction || containsLiteral)
        t.Request.Arguments["charset specification"] = charset.WebName;

      // search-return-opt
      if (resultOptions != null)
        t.Request.Arguments["result specifier"] = resultOptions;

      // searching criteria
      t.Request.Arguments["searching criteria"] = searchingCriteria;

        // sort criteria
      if (sortingCriteria != null)
        t.Request.Arguments["sort criteria"] = sortingCriteria;

      if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
        matched = t.Result.Value;

      // TODO: BADCHARSET
      // 6.4.4. SEARCH Command
      //       If the server does not support the specified [CHARSET], it MUST
      //       return a tagged NO response (not a BAD).  This response SHOULD
      //       contain the BADCHARSET response code, which MAY list the
      //       [CHARSET]s supported by the server.

      return t.Result;
    }

    /// <summary>sends THREAD command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Thread(ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, out ImapThreadList threadList)
    {
      using (var t = new ThreadTransaction(connection, false)) {
        return ThreadInternal(t, threadingAlgorithm, searchingCriteria, Encoding.UTF8, out threadList);
      }
    }

    /// <summary>sends THREAD command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Thread(ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, Encoding charset, out ImapThreadList threadList)
    {
      using (var t = new ThreadTransaction(connection, false)) {
        return ThreadInternal(t, threadingAlgorithm, searchingCriteria, charset, out threadList);
      }
    }

    /// <summary>sends UID THREAD command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidThread(ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, out ImapThreadList threadList)
    {
      using (var t = new ThreadTransaction(connection, true)) {
        return ThreadInternal(t, threadingAlgorithm, searchingCriteria, Encoding.UTF8, out threadList);
      }
    }

    /// <summary>sends UID THREAD command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidThread(ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, Encoding charset, out ImapThreadList threadList)
    {
      using (var t = new ThreadTransaction(connection, true)) {
        return ThreadInternal(t, threadingAlgorithm, searchingCriteria, charset, out threadList);
      }
    }

    private ImapTransactionResult ThreadInternal(ThreadTransaction t, ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, Encoding charset, out ImapThreadList threadList)
    {
      RejectNonSelectedState();

      threadList = null;

      // TODO: charset
      // http://tools.ietf.org/html/rfc5256
      // BASE.6.4.THREAD. THREAD Command
      //   The US-ASCII and [UTF-8] charsets MUST be implemented.
      searchingCriteria.Traverse(delegate(ImapString criterion) {
        if (criterion is ImapLiteralString)
          (criterion as ImapLiteralString).Charset = charset;
      });

      // threading algorithm
      t.Request.Arguments["threading algorithm"] = threadingAlgorithm;

      // charset specification
      t.Request.Arguments["charset specification"] = charset.WebName;

      // searching criteria
      t.Request.Arguments["searching criteria"] = searchingCriteria;

      if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
        threadList = t.Result.Value;

      return t.Result;
    }

    /// <summary>sends FETCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Fetch(ImapSequenceSet sequenceOrUidSet, ImapFetchDataItem messageDataItems, out ImapMessage[] messages)
    {
      return FetchInternal(sequenceOrUidSet, messageDataItems, null, null, out messages);
    }

    /// <summary>sends FETCH command with CHANGEDSINCE FETCH modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    public ImapTransactionResult FetchChangedSince(ImapSequenceSet sequenceOrUidSet, ImapFetchDataItem messageDataItems, ulong modSeq, out ImapMessage[] messages)
    {
      var fetchModifiers = new ImapParenthesizedString("CHANGEDSINCE", modSeq.ToString());

      return FetchInternal(sequenceOrUidSet, messageDataItems, fetchModifiers, ImapCapability.CondStore, out messages);
    }

    private ImapTransactionResult FetchInternal(ImapSequenceSet sequenceOrUidSet, ImapFetchDataItem messageDataItems, ImapParenthesizedString fetchModifiers, ImapCapability fetchModifiersCapabilityRequirement, out ImapMessage[] messages)
    {
      RejectNonSelectedState();

      messages = null;

      using (var t = new FetchTransaction(connection, sequenceOrUidSet.IsUidSet, fetchModifiersCapabilityRequirement)) {
        /*
         * RFC 4466 - Collected Extensions to IMAP4 ABNF
         * http://tools.ietf.org/html/rfc4466
         *
         *    fetch           = "FETCH" SP sequence-set SP ("ALL" / "FULL" /
         *                      "FAST" / fetch-att /
         *                      "(" fetch-att *(SP fetch-att) ")")
         *                      [fetch-modifiers]
         *                      ;; modifies the original IMAP4 FETCH command to
         *                      ;; accept optional modifiers
         *    fetch-modifiers = SP "(" fetch-modifier *(SP fetch-modifier) ")"
         *    fetch-modifier  = fetch-modifier-name [ SP fetch-modif-params ]
         *    fetch-modif-params  = tagged-ext-val
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions.
         *    fetch-modifier-name = tagged-ext-label
         */
        // sequence set
        t.Request.Arguments["sequence set"] = sequenceOrUidSet.ToString();

        // message data item names or macro
        t.Request.Arguments["message data item names or macro"] = messageDataItems;

        // fetch-modifiers
        if (fetchModifiers != null)
          t.Request.Arguments["fetch modifiers"] = fetchModifiers;

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          messages = t.Result.Value;

          foreach (var message in messages) {
            message.Mailbox = selectedMailbox;
          }
        }

        return t.Result;
      }
    }

    /// <summary>sends STORE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Store(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem)
    {
      ImapMessage[] discard;

      return Store(sequenceOrUidSet, storeDataItem, out discard);
    }

    /// <summary>sends STORE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Store(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, out ImapMessage[] messages)
    {
      return StoreInternal(sequenceOrUidSet, storeDataItem, null, null, out messages);
    }

    /// <summary>sends STORE command with UNCHANGEDSINCE STORE modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    public ImapTransactionResult StoreUnchangedSince(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, ulong modSeq)
    {
      ImapMessage[] discard;
      ImapSequenceSet discard2;

      return StoreUnchangedSince(sequenceOrUidSet, storeDataItem, modSeq, out discard, out discard2);
    }

    /// <summary>sends STORE command with UNCHANGEDSINCE STORE modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    public ImapTransactionResult StoreUnchangedSince(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, ulong modSeq, out ImapMessage[] messages)
    {
      ImapSequenceSet discard;

      return StoreUnchangedSince(sequenceOrUidSet, storeDataItem, modSeq, out messages, out discard);
    }

    /// <summary>sends STORE command with UNCHANGEDSINCE STORE modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    public ImapTransactionResult StoreUnchangedSince(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, ulong modSeq, out ImapSequenceSet failedMessageSet)
    {
      ImapMessage[] discard;

      return StoreUnchangedSince(sequenceOrUidSet, storeDataItem, modSeq, out discard, out failedMessageSet);
    }

    /// <summary>sends STORE command with UNCHANGEDSINCE STORE modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    public ImapTransactionResult StoreUnchangedSince(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, ulong modSeq, out ImapMessage[] messages, out ImapSequenceSet failedMessageSet)
    {
      failedMessageSet = null;
      messages = null;

      var storeModifiers = new ImapParenthesizedString("UNCHANGEDSINCE", modSeq.ToString());

      var ret = StoreInternal(sequenceOrUidSet, storeDataItem, storeModifiers, ImapCapability.CondStore, out messages);

      var modified = lastTransactionResult.GetResponseCode(ImapResponseCode.Modified);

      if (modified is ImapTaggedStatusResponse)
        failedMessageSet = ImapResponseTextConverter.FromModified(modified.ResponseText, sequenceOrUidSet.IsUidSet);

      return ret;
    }

    private ImapTransactionResult StoreInternal(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, ImapParenthesizedString storeModifiers, ImapCapability storeModifiersCapabilityRequirement, out ImapMessage[] messages)
    {
      RejectNonSelectedState();

      messages = null;

      using (var t = new StoreTransaction(connection, sequenceOrUidSet.IsUidSet, storeModifiersCapabilityRequirement)) {
        /*
         * RFC 4466 - Collected Extensions to IMAP4 ABNF
         * http://tools.ietf.org/html/rfc4466
         *
         *    store           = "STORE" SP sequence-set [store-modifiers]
         *                      SP store-att-flags
         *                      ;; extend [IMAP4] STORE command syntax
         *                      ;; to allow for optional store-modifiers
         *    store-modifiers =  SP "(" store-modifier *(SP store-modifier)
         *                        ")"
         *    store-modifier  = store-modifier-name [SP store-modif-params]
         *    store-modif-params = tagged-ext-val
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions.
         *    store-modifier-name = tagged-ext-label
         */
        // message set
        t.Request.Arguments["message set"] = sequenceOrUidSet.ToString();

        // store-modifiers
        if (storeModifiers != null)
          t.Request.Arguments["store modifiers"] = storeModifiers;

        // message data item name
        t.Request.Arguments["message data item name"] = storeDataItem.ItemName;

        // value for message data item
        t.Request.Arguments["value for message data item"] = storeDataItem;

        var ret = ProcessTransaction(t, handlesIncapableAsException);

        messages = t.Result.Value;

        if (messages != null) {
          foreach (var message in messages) {
            message.Mailbox = selectedMailbox;
          }
        }

        return ret;
      }
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      ImapCopiedUidSet discard;
      ImapMailbox discard2;

      return CopyInternal(sequenceOrUidSet, mailbox.Name, false, out discard, out discard2);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, ImapMailbox mailbox, out ImapCopiedUidSet copiedUids)
    {
      ValidateMailboxRelationship(mailbox);

      ImapMailbox discard;

      return CopyInternal(sequenceOrUidSet, mailbox.Name, false, out copiedUids, out discard);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, string mailboxName)
    {
      ImapCopiedUidSet discard;
      ImapMailbox discard2;

      return CopyInternal(sequenceOrUidSet, mailboxName, false, out discard, out discard2);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, string mailboxName, out ImapCopiedUidSet copiedUids)
    {
      ImapMailbox discard;

      return CopyInternal(sequenceOrUidSet, mailboxName, false, out copiedUids, out discard);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code
    /// </remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, string mailboxName, out ImapMailbox createdMailbox)
    {
      ImapCopiedUidSet discard;

      return CopyInternal(sequenceOrUidSet, mailboxName, true, out discard, out createdMailbox);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, string mailboxName, out ImapCopiedUidSet copiedUids, out ImapMailbox createdMailbox)
    {
      return CopyInternal(sequenceOrUidSet, mailboxName, true, out copiedUids, out createdMailbox);
    }

    private ImapTransactionResult CopyInternal(ImapSequenceSet sequenceOrUidSet, string mailboxName, bool tryCreate, out ImapCopiedUidSet copiedUids, out ImapMailbox createdMailbox)
    {
      RejectNonSelectedState();

      copiedUids = null;
      createdMailbox = null;

      ImapTransactionResult failedResult = null;

      for (var i = 0; i < 2; i++) {
        var respTryCreate = false;

        using (var t = new CopyTransaction(connection, sequenceOrUidSet.IsUidSet)) {
          // sequence set
          t.Request.Arguments["sequence set"] = sequenceOrUidSet.ToString();

          // mailbox name
          t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

          if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
            copiedUids = t.Result.Value;
            return t.Result;
          }
          else if (!tryCreate) {
            return t.Result;
          }

          failedResult = t.Result;

          // 6.4.7. COPY Command
          //       If the destination mailbox does not exist, a server SHOULD return
          //       an error.  It SHOULD NOT automatically create the mailbox.  Unless
          //       it is certain that the destination mailbox can not be created, the
          //       server MUST send the response code "[TRYCREATE]" as the prefix of
          //       the text of the tagged NO response.  This gives a hint to the
          //       client that it can attempt a CREATE command and retry the COPY if
          //       the CREATE is successful.
          var tryCreateRespCode = t.Result.GetResponseCode(ImapResponseCode.TryCreate);

          respTryCreate = (tryCreateRespCode != null && (tryCreateRespCode is ImapTaggedStatusResponse));
        }

        // try create
        if (i == 0 && respTryCreate)
          if (Create(mailboxName, out createdMailbox).Failed)
            return failedResult;
      }

      return failedResult;
    }

    /// <summary>sends UNSELECT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Unselect()
    {
      RejectNonSelectedState();

      using (var t = new UnselectTransaction(connection)) {
        var r = ProcessTransaction(t, handlesIncapableAsException);

        if (r.Succeeded) {
          selectedMailbox = null;
          TransitStateTo(ImapSessionState.Authenticated);
        }

        return r;
      }
    }

    /*
     * transaction methods : authenticated/selected state
     */


    public ImapTransactionResult Idle(TimeSpan timeout)
    {
      return Idle((int)timeout.TotalMilliseconds);
    }

    public ImapTransactionResult Idle(int millisecondsTimeout)
    {
      if (millisecondsTimeout < -1)
        throw new ArgumentOutOfRangeException("millisecondsTimeout", "must be greater than or equals to -1");

      var async = BeginIdle();

      Threading.Thread.Sleep(millisecondsTimeout);

      return EndIdle(async);
    }

    public ImapTransactionResult Idle(Predicate<ImapUpdatedStatus> endIdlePredicate)
    {
      return Idle(Timeout.Infinite, endIdlePredicate);
    }

    public ImapTransactionResult Idle(TimeSpan idleTimeout, Predicate<ImapUpdatedStatus> endIdlePredicate)
    {
      return Idle((int)idleTimeout.TotalMilliseconds, endIdlePredicate);
    }

    public ImapTransactionResult Idle(int idleMillisecondsTimeout, Predicate<ImapUpdatedStatus> endIdlePredicate)
    {
      if (idleMillisecondsTimeout < -1)
        throw new ArgumentOutOfRangeException("idleMillisecondsTimeout", "must be greater than or equals to -1");
      if (endIdlePredicate == null)
        throw new ArgumentNullException("endIdlePredicate");

      var async = BeginIdle();
      var endIdleEvent = new AutoResetEvent(false);

      StatusUpdated +=  delegate(object sender, ImapStatusUpdatedEventArgs e) {
        if (endIdlePredicate(e.UpdatedStatus))
          endIdleEvent.Set();
      };

      endIdleEvent.WaitOne(idleMillisecondsTimeout);

      return EndIdle(async);
    }

    /// <summary>sends IDLE command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public IAsyncResult BeginIdle()
    {
      IdleTransaction t = null;

      try {
        RejectNonAuthenticatedState();
        RejectIdling();

        t = new IdleTransaction(connection, StatusUpdatedHandler);

        return BeginProcessTransaction(t, handlesIncapableAsException);
      }
      catch {
        if (t != null) {
          t.Dispose();
          t = null;
        }

        throw;
      }
    }

    /// <summary>sends DONE for IDLE command continuation request responses</summary>
    /// <remarks>valid in idling</remarks>
    public ImapTransactionResult EndIdle(IAsyncResult asyncResult)
    {
      var idleAsyncResult = asyncResult as TransactionAsyncResult;

      if (idleAsyncResult == null)
        throw new ArgumentException("invalid IAsyncResult", "asyncResult");

      using (var t = idleAsyncResult.Transaction as IdleTransaction) {
        // send 'DONE'
        t.SendDone();

        return EndProcessTransaction(idleAsyncResult);
      }
    }

    /// <summary>sends NAMESPACE command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapTransactionResult Namespace()
    {
      ImapNamespace discard;

      return Namespace(out discard);
    }

    /// <summary>sends NAMESPACE command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapTransactionResult Namespace(out ImapNamespace namespaces)
    {
      RejectNonAuthenticatedState();

      namespaces = null;

      using (var t = new NamespaceTransaction(connection)) {
        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          this.namespaces = t.Result.Value;
          namespaces = t.Result.Value;
        }

        return t.Result;
      }
    }

    /*
     * transaction methods : generic use, exntension, etc.
     */

    /// <summary>sends generic/undefined command</summary>
    public ImapTransactionResult GenericCommand(string command, params ImapString[] arguments)
    {
      ImapDataResponse[] discard;

      return GenericCommand(command, out discard, arguments);
    }

    /// <summary>sends generic/undefined command</summary>
    public ImapTransactionResult GenericCommand(string command, out Protocol.ImapDataResponse[] dataResponses, params ImapString[] arguments)
    {
      RejectNonConnectedState();

      dataResponses = null;

      using (var t = new GenericCommandTransaction(connection, command)) {
        if (arguments != null && 0 < arguments.Length)
          t.Request.Arguments["arguments"] = new ImapStringList(arguments);

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
          dataResponses = t.Result.Value;

        return t.Result;
      }
    }

    /*
     * methods for processing transaction
     */

    private class TransactionAsyncResult : IAsyncResult {
      object IAsyncResult.AsyncState {
        get { return processTransactionAsyncResult.AsyncState; }
      }

      Threading.WaitHandle IAsyncResult.AsyncWaitHandle {
        get { return processTransactionAsyncResult.AsyncWaitHandle; }
      }

      bool IAsyncResult.CompletedSynchronously {
        get { return processTransactionAsyncResult.CompletedSynchronously; }
      }

      bool IAsyncResult.IsCompleted {
        get { return processTransactionAsyncResult.IsCompleted; }
      }

      public IImapTransaction Transaction { get { return transaction; } }
      public IAsyncResult ProcessTransactionAsyncResult { get { return processTransactionAsyncResult; } }

      public TransactionAsyncResult(IImapTransaction transaction, IAsyncResult processTransactionAsyncResult)
      {
        this.transaction = transaction;
        this.processTransactionAsyncResult = processTransactionAsyncResult;
      }

      private /*readonly*/ IImapTransaction transaction;
      private /*readonly*/ IAsyncResult processTransactionAsyncResult;
    }

    public ImapTransactionResult ProcessTransaction(IImapTransaction t)
    {
      return ProcessTransaction(t, false);
    }

    public ImapTransactionResult ProcessTransaction(IImapTransaction t, bool exceptionIfIncapable)
    {
      if (transactionTimeout == Timeout.Infinite) {
        // no timeout
        PreProcessTransaction(t, exceptionIfIncapable);

        ProcessTransactionInternal(t);

        return PostProcessTransaction(t);
      }
      else {
        var async = BeginProcessTransaction(t, exceptionIfIncapable);

        if (async.AsyncWaitHandle.WaitOne(transactionTimeout)) {
          return EndProcessTransaction(async);
        }
        else {
          CloseConnection();
          throw new Exception("timed out");
        }
      }
    }

    public IAsyncResult BeginProcessTransaction(IImapTransaction t)
    {
      return BeginProcessTransaction(t, false);
    }

    public IAsyncResult BeginProcessTransaction(IImapTransaction t, bool exceptionIfIncapable)
    {
      return BeginProcessTransaction(t, exceptionIfIncapable, null, null);
    }

    public IAsyncResult BeginProcessTransaction(IImapTransaction t, bool exceptionIfIncapable, AsyncCallback callback, object state)
    {
      PreProcessTransaction(t, exceptionIfIncapable);

      var processTransaction = new ProcessTransactionDelegate(ProcessTransactionInternal);

      return new TransactionAsyncResult(t, processTransaction.BeginInvoke(t, callback, state));
    }

    public ImapTransactionResult EndProcessTransaction(IAsyncResult asyncResult)
    {
      var ar = asyncResult as TransactionAsyncResult;

      if (ar == null)
        throw new ArgumentException("invalid IAsyncResult", "asyncResult");

      var ptar = ar.ProcessTransactionAsyncResult as System.Runtime.Remoting.Messaging.AsyncResult;

      if (ptar.EndInvokeCalled)
        throw new InvalidOperationException("EndProcessTransaction already called");

      var processTransaction = ptar.AsyncDelegate as ProcessTransactionDelegate;

      return PostProcessTransaction(processTransaction.EndInvoke(ptar));
    }

    private delegate IImapTransaction ProcessTransactionDelegate(IImapTransaction t);

    private IImapTransaction ProcessTransactionInternal(IImapTransaction t)
    {
      try {
        transactionProceedingSemaphore.WaitOne();

        Trace.LogRequest(t);

        t.Start();

        for (;;) {
          if (t.State == ImapTransactionState.Finished) {
            // transaction is finished
            Trace.LogResponse(t);
            return t;
          }

          t.Process();
        }
      }
      finally {
        transactionProceedingSemaphore.Release();
      }
    }

    private void PreProcessTransaction(IImapTransaction t, bool exceptionIfIncapable)
    {
      // TODO: '5.5. Multiple Commands in Progress'
      //   The following are examples of valid non-waiting command sequences:
      //      FETCH + STORE + SEARCH + CHECK
      //      STORE + COPY + EXPUNGE
      if (IsTransactionProceeding)
        throw new InvalidOperationException("another transaction proceesing");

      RejectIdling();

      // check capability
      if (!exceptionIfIncapable)
        return;

      if (!serverCapabilities.Has(Imap4.ImapCapability.Imap4Rev1))
        throw new ImapIncapableException(Imap4.ImapCapability.Imap4Rev1);

      if (t is IImapExtension)
        CheckServerCapability(t as IImapExtension);
      else if (t is IImapMultipleExtension)
        CheckServerCapability(t as IImapMultipleExtension);

      foreach (var arg in t.Request.Arguments.Values) {
        if (arg is IImapExtension)
          CheckServerCapability(arg as IImapExtension);
        else if (arg is IImapMultipleExtension)
          CheckServerCapability(arg as IImapMultipleExtension);
      }
    }

    private ImapResponseProcessorBase[] responseProcessors;

    private void InitializeResponseProcessors()
    {
      responseProcessors = new ImapResponseProcessorBase[] {
        new ImapCapabilityResponseProcessor(this),
        new ImapUpdatedStatusResponseProcessor(this),
        new ImapReferralResponseProcessor(this),
      };
    }

    private ImapTransactionResult PostProcessTransaction(IImapTransaction t)
    {
      lastTransactionResult = t.Result;

      if (t.Result.Code == ImapTransactionResultCode.InternalError) {
        // disconnect without processing responses
        TransitStateTo(ImapSessionState.NotAuthenticated);
        CloseConnection();
      }

      lastUpdatedStatus = new ImapUpdatedStatus();

      var unprocessedResponses = new List<ImapResponse>(t.Result.ReceivedResponses);

      foreach (var processor in responseProcessors) {
        foreach (var resp in unprocessedResponses.ToArray()) {
          if (processor.Process(resp, t))
            unprocessedResponses.Remove(resp);
        }
      }

      if (t.Result.Code == ImapTransactionResultCode.Bye) {
        TransitStateTo(ImapSessionState.NotAuthenticated);
        CloseConnection();
      }

      return t.Result;
    }

    /*
     * methods for internal state management
     */
    internal void SetServerCapabilities(ImapCapabilityList newCapabilities)
    {
      serverCapabilities = newCapabilities;

      connection.SynchronizeLiterals = !serverCapabilities.Has(ImapCapability.LiteralNonSync);
    }

    private void TransitStateTo(ImapSessionState newState)
    {
      if (newState == state)
        return;

      switch (newState) {
        case ImapSessionState.NotConnected:
          TraceVerbose("now in non-connected state");
          break;

        case ImapSessionState.NotAuthenticated:
          UpdateServerUrl(null, null);
          TraceVerbose("now in non-authenticated state");
          if (state == ImapSessionState.Selected || state == ImapSessionState.Authenticated)
            Trace.Info("CID:{0} logged out", Id);
          break;

        case ImapSessionState.Authenticated:
          TraceVerbose("now in authenticated state");
          if (state == ImapSessionState.NotAuthenticated || state == ImapSessionState.NotConnected)
            Trace.Info("CID:{0} logged in", Id);
          break;

        case ImapSessionState.Selected:
          TraceVerbose("now in selected state (selected '{0}')", selectedMailbox.Name);
          break;
      }

      state = newState;
    }

    /*
     * methods for checking and validating session status
     */
    private void CheckDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().FullName);
    }

    private void RejectNonConnectedState()
    {
      CheckDisposed();

      if (state == ImapSessionState.NotConnected)
        throw new InvalidOperationException("session has been disconnected");
    }

    private void RejectNonAuthenticatedState()
    {
      RejectNonConnectedState();

      if (state == ImapSessionState.NotAuthenticated)
        throw new InvalidOperationException("session is not in authenticated state");
    }

    private void RejectNonSelectedState()
    {
      RejectNonConnectedState();

      if (state != ImapSessionState.Selected)
        throw new InvalidOperationException("session is not in selected state");
    }

    private void RejectIdling()
    {
      if (IsIdling)
        throw new InvalidOperationException("session is idling now");
    }

    private void ValidateMailboxRelationship(ImapMailbox mailbox)
    {
      if (mailbox == null)
        throw new ArgumentNullException("mailbox");

      if (mailbox.Session != this)
        throw new InvalidOperationException("invalid mailbox");
    }

    private void CheckServerCapability(IImapExtension extension)
    {
      CheckServerCapability(extension.RequiredCapability);
    }

    private void CheckServerCapability(IImapMultipleExtension extensions)
    {
      foreach (var capability in extensions.RequiredCapabilities) {
        CheckServerCapability(capability);
      }
    }

    private void CheckServerCapability(ImapCapability capability)
    {
      if (capability != null && !serverCapabilities.Has(capability))
        throw new ImapIncapableException(capability);
    }

    /*
     * tracing
     */
    [System.Diagnostics.Conditional("TRACE")]
    private void TraceVerbose(string format, params object[] arguments)
    {
      Trace.Log(this, format, arguments);
    }
  }
}
