// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;
using System.Text;

using Smdn.Protocols.Imap4.Client.Protocol;
using Smdn.Protocols.Imap4.Client.Protocol.Authentication;

namespace Smdn.Protocols.Imap4.Client.Transaction.BuiltIn {
  internal class AuthenticateTransaction : ImapTransactionBase {
    public AuthenticateTransaction(ImapConnection connection, NetworkCredential credential)
      : base(connection)
    {
      if (credential == null)
        throw new ArgumentNullException("credential");

      this.credential = credential;
    }

    protected override ProcessTransactionDelegate Reset()
    {
      if (!Request.Arguments.ContainsKey("authentication mechanism name"))
        return ProcessArgumentNotSetted;

      authenticationModule = ImapAuthenticationManager.CreateModule(Request.Arguments["authentication mechanism name"] as ImapAuthenticationMechanism);

      if (authenticationModule == null)
        return ProcessUnsupported;

      return ProcessAuthenticate;
    }

    private void ProcessArgumentNotSetted()
    {
      FinishError(ImapTransactionResultCode.RequestError, "arguments 'authentication mechanism name' must be setted");
    }

    private void ProcessUnsupported()
    {
      FinishError(ImapTransactionResultCode.RequestError, "unsupported authentication mechanism");
    }

    // 6.2.2. AUTHENTICATE Command
    //    Arguments:  authentication mechanism name
    //    Responses:  continuation data can be requested
    //    Result:     OK - authenticate completed, now in authenticated state
    //                NO - authenticate failure: unsupported authentication
    //                     mechanism, credentials rejected
    //                BAD - command unknown or arguments invalid,
    //                     authentication exchange cancelled
    private void ProcessAuthenticate()
    {
      SendCommand("AUTHENTICATE", ProcessReceiveResponse, authenticationModule.Mechanism);
    }

    protected override void OnCommandContinuationRequestReceived(ImapCommandContinuationRequest continuationRequest)
    {
      var response = authenticationModule.ExchangeChallengeResponse(continuationRequest,
                                                                    Connection,
                                                                    credential);

      if (response == null)
        // If the client wishes to cancel an authentication exchange,
        // it issues a line consisting of a single "*".
        SendContinuation("*");
      else
        SendContinuation(response);
    }

    private /*readonly*/ NetworkCredential credential;
    private /*readonly*/ IImapAuthenticationModule authenticationModule;
  }
}
