// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Protocols.Imap4.Client.Protocol;

namespace Smdn.Protocols.Imap4.Client.Transaction.BuiltIn {
  internal class IdleTransaction : ImapTransactionBase, IImapExtension {
    ImapCapability IImapExtension.RequiredCapability {
      get { return ImapCapability.Idle; }
    }

    public IdleTransaction(ImapConnection connection, EventHandler<ImapStatusUpdatedEventArgs> statusUpdatedHandler)
      : base(connection)
    {
      this.statusUpdatedHandler = statusUpdatedHandler;
    }

    protected override ProcessTransactionDelegate Reset()
    {
      return ProcessIdle;
    }

    // 3. Specification
    //    IDLE Command
    //    Arguments:  none
    //    Responses:  continuation data will be requested; the client sends
    //                the continuation data "DONE" to end the command
    //    Result:     OK - IDLE completed after client sent "DONE"
    //                NO - failure: the server will not allow the IDLE
    //                     command at this time
    //               BAD - command unknown or arguments invalid
    private void ProcessIdle()
    {
      SendCommand("IDLE", ProcessReceiveResponse);
    }

    internal void SendDone()
    {
      SendContinuation("DONE");
    }

    protected override void OnCommandContinuationRequestReceived(ImapCommandContinuationRequest continuationRequest)
    {
      Connection.IsIdling = true;
    }

    protected override void OnDataResponseReceived(ImapDataResponse data)
    {
      // 3. Specification
      //   IDLE Command
      //   The IDLE command remains active until the client
      //   responds to the continuation, and as long as an IDLE command is
      //   active, the server is now free to send untagged EXISTS, EXPUNGE, and
      //   other messages at any time
      if (statusUpdatedHandler != null) {
        ImapUpdatedStatus updatedStatus = null;

        if (data.Type == ImapDataResponseType.Fetch)
          updatedStatus = ImapUpdatedStatus.CreateMessageUpdated(ImapDataResponseConverter.FromFetch(data));
        else if (data.Type == ImapDataResponseType.Expunge)
          updatedStatus = ImapUpdatedStatus.CreateExpungedUpdated(ImapDataResponseConverter.FromExpunge(data));
        else if (data.Type == ImapDataResponseType.Exists)
          updatedStatus = ImapUpdatedStatus.CreateExistsUpdated(ImapDataResponseConverter.FromExists(data));
        else if (data.Type == ImapDataResponseType.Recent)
          updatedStatus = ImapUpdatedStatus.CreateRecentUpdated(ImapDataResponseConverter.FromRecent(data));
        else if (data.Type == ImapDataResponseType.Flags)
          updatedStatus = ImapUpdatedStatus.CreateFlagsUpdated(ImapDataResponseConverter.FromFlags(data));

        if (updatedStatus != null && statusUpdatedHandler != null)
          statusUpdatedHandler(this, new ImapStatusUpdatedEventArgs(updatedStatus));
      }

      base.OnDataResponseReceived(data);
    }

    protected override void OnTaggedStatusResponseReceived(ImapTaggedStatusResponse tagged)
    {
      Connection.IsIdling = false;

      Connection.ClearBuffer(); // XXX

      base.OnTaggedStatusResponseReceived(tagged);
    }

    private /*readonly*/ EventHandler<ImapStatusUpdatedEventArgs> statusUpdatedHandler;
  }
}