// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Protocols.Imap4.Client.Protocol;

namespace Smdn.Protocols.Imap4.Client.Transaction.BuiltIn {
  internal abstract class ListTransactionBase : ImapTransactionBase<ImapTransactionResult<ImapMailboxList[]>> {
    protected ListTransactionBase(string command, ImapConnection connection)
      : base(connection)
    {
      commandString = command;
    }

    protected override ProcessTransactionDelegate Reset()
    {
      if (!Request.Arguments.ContainsKey("reference name"))
        return ProcessArgumentNotSetted;
      else if (!Request.Arguments.ContainsKey("mailbox name"))
        return ProcessArgumentNotSetted;
      else
        return ProcessList;
    }

    private void ProcessArgumentNotSetted()
    {
      FinishError(ImapTransactionResultCode.RequestError, "arguments 'reference name' and 'mailbox name' must be setted");
    }

    // 6.3.8. LIST Command
    //    Arguments:  reference name
    //                mailbox name with possible wildcards
    //    Responses:  untagged responses: LIST
    //    Result:     OK - list completed
    //                NO - list failure: can't list that reference or name
    //                BAD - command unknown or arguments invalid
    
    // 6.3.9. LSUB Command
    //    Arguments:  reference name
    //                mailbox name with possible wildcards
    //    Responses:  untagged responses: LSUB
    //    Result:     OK - lsub completed
    //                NO - lsub failure: can't list that reference or name
    //                BAD - command unknown or arguments invalid

    // http://tools.ietf.org/html/rfc2193
    // RFC 2193 IMAP4 Mailbox Referrals
    // 
    // 5.1 RLIST command
    //   Arguments:  reference name
    //               mailbox name with possible wildcards
    //   Responses:  untagged responses: LIST
    //   Result:     OK - RLIST Completed
    //               NO - RLIST Failure
    //               BAD - command unknown or arguments invalid
    //
    //   The RLIST command behaves identically to its LIST counterpart, except
    //   remote mailboxes are returned in addition to local mailboxes in the
    //   LIST responses.
    // 
    // 5.2 RLSUB Command
    //    Arguments:  reference name
    //                mailbox name with possible wildcards
    //    Responses:  untagged responses: LSUB
    //    Result:     OK - RLSUB Completed
    //                NO - RLSUB Failure
    //                BAD - command unknown or arguments invalid
    // 
    //    The RLSUB command behaves identically to its LSUB counterpart, except
    //    remote mailboxes are returned in addition to local mailboxes in the
    //    LSUB responses.
    private void ProcessList()
    {
      SendCommand(commandString,
                  ProcessReceiveResponse,
                  Request.Arguments["reference name"],
                  Request.Arguments["mailbox name"]);
    }

    protected override void OnDataResponseReceived(ImapDataResponse data)
    {
      if (this is ListTransaction) {
        if (data.Type == ImapDataResponseType.List)
          mailboxLists.Add(ImapDataResponseConverter.FromList(data));
      }
      else {
        if (data.Type == ImapDataResponseType.Lsub)
          mailboxLists.Add(ImapDataResponseConverter.FromLsub(data));
      }

      base.OnDataResponseReceived(data);
    }

    protected override void OnTaggedStatusResponseReceived(ImapTaggedStatusResponse tagged)
    {
      if (tagged.Condition == ImapResponseCondition.Ok) {
        Finish(new ImapTransactionResult<ImapMailboxList[]>(mailboxLists.ToArray(), tagged.ResponseText.Text));
        mailboxLists.Clear();
      }
      else {
        base.OnTaggedStatusResponseReceived(tagged);
      }
    }

    private string commandString = string.Empty;
    private /*readonly*/ List<ImapMailboxList> mailboxLists = new List<ImapMailboxList>();
  }
}
