// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Threading;

namespace Smdn.Protocols.Imap4.WebClient {
  internal class ImapAppendMessageBodyStream : Stream {
    public override bool CanSeek {
      get { return false; }
    }

    public override bool CanRead {
      get { return true; }
    }

    public override bool CanWrite {
      get { return true; }
    }

    public override bool CanTimeout {
      get { return true; }
    }

    public override long Position {
      get { return writeOffset; }
      set { writeOffset = value; }
    }

    public override long Length {
      get
      {
        lengthSettledEvent.WaitOne();
        return length;
      }
    }

    public ImapAppendMessageBodyStream()
    {
      buffer = new MemoryStream(4096);
    }

    public override void Close()
    {
      if (!lengthSettledEvent.WaitOne(0)) {
        this.length = buffer.Length;

        lengthSettledEvent.Set();
      }
    }

    protected override void Dispose(bool disposing)
    {
      if (disposed)
        return;

      base.Dispose(disposing);

      lengthSettledEvent.Close();
      lengthSettledEvent = null;

      buffer.Close();
      buffer = null;

      disposed = true;
    }

    public override void SetLength(long @value)
    {
      if (lengthSettledEvent.WaitOne(0))
        throw new InvalidOperationException("length has been settled");

      length = @value;
      lengthSettledEvent.Set();
    }

    public override long Seek(long offset, SeekOrigin origin)
    {
      throw new NotSupportedException();
    }

    public override void Flush()
    {
      throw new NotSupportedException();
    }

    public override void Write(byte[] src, int offset, int count)
    {
      RejectDisposed();

      lock (readWriteLock) {
        writeOffset = buffer.Seek(writeOffset, SeekOrigin.Begin);

        buffer.Write(src, offset, count);

        writeOffset += count;
      }

      if (lengthSettledEvent.WaitOne(0) && 40960 < writeOffset - readOffset)
        // wait for reading
        Thread.Sleep(50);
    }

    public override int Read(byte[] dest, int offset, int count)
    {
      RejectDisposed();

      lengthSettledEvent.WaitOne();

      lock (readWriteLock) {
        readOffset = buffer.Seek(readOffset, SeekOrigin.Begin);

        var read = buffer.Read(dest, offset, count);

        readOffset += read;

        return read;
      }
    }

    private void RejectDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().Name);
    }

    private bool disposed = false;
    private long writeOffset = 0;
    private long readOffset = 0;
    private long length = 0;

    private MemoryStream buffer;
    private ManualResetEvent lengthSettledEvent = new ManualResetEvent(false);
    private object readWriteLock = new object();
  }
}
