// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Net;

using Smdn.Protocols.Imap4.Client.Session;

namespace Smdn.Protocols.Imap4.WebClient {
  internal class ImapMailboxWebRequest : ImapWebRequest {
    public ImapMailboxWebRequest(ImapUri requestUri, ImapSessionManager sessionManager)
      : base(requestUri, sessionManager)
    {
    }

    private bool IsRequestAppend {
      get { return string.Equals(ImapWebRequestMethods.Append, Method, StringComparison.InvariantCultureIgnoreCase); }
    }

    protected override ImapWebResponse InternalGetResponse(ImapUri requestUri)
    {
      if (!string.Equals(Method, ImapWebRequestMethods.Append, StringComparison.InvariantCultureIgnoreCase)) {
        var response = UnselectMailbox();

        if (response != null)
          return response;
      }

      if (string.IsNullOrEmpty(Method) ||
          string.Equals(Method, ImapWebRequestMethods.Fetch, StringComparison.InvariantCultureIgnoreCase))
        return GetFetchResponse(requestUri);
      else if (string.Equals(Method, ImapWebRequestMethods.Append, StringComparison.InvariantCultureIgnoreCase))
        return GetAppendResponse(requestUri);
      else if (string.Equals(Method, ImapWebRequestMethods.Create, StringComparison.InvariantCultureIgnoreCase))
        return GetCreateResponse(requestUri);
      else if (string.Equals(Method, ImapWebRequestMethods.Delete, StringComparison.InvariantCultureIgnoreCase))
        return GetDeleteResponse(requestUri);
      else if (string.Equals(Method, ImapWebRequestMethods.Expunge, StringComparison.InvariantCultureIgnoreCase))
        return GetExpungeResponse(requestUri);
      else if (string.Equals(Method, ImapWebRequestMethods.Examine, StringComparison.InvariantCultureIgnoreCase))
        return GetExamineResponse(requestUri);
      else if (string.Equals(Method, ImapWebRequestMethods.Rename, StringComparison.InvariantCultureIgnoreCase))
        return GetRenameResponse(requestUri);
      else if (string.Equals(Method, ImapWebRequestMethods.Subscribe, StringComparison.InvariantCultureIgnoreCase))
        return GetSubscribeResponse(requestUri);
      else if (string.Equals(Method, ImapWebRequestMethods.Unsubscribe, StringComparison.InvariantCultureIgnoreCase))
        return GetUnsubscribeResponse(requestUri);
      else
        throw new NotSupportedException(string.Format("unsupported request method: {0}", Method));
    }

    private ImapWebResponse GetFetchResponse(ImapUri requestUri)
    {
      var response = SelectMailbox(requestUri);

      if (response != null)
        return response;

      ImapMessage[] messages;

      var result = Session.Fetch(ImapSequenceSet.CreateUidFromSet(1), ImapFetchDataItem.All, out messages);

      if (result.Failed)
        return new ImapWebResponse(result);

      response = new ImapWebResponse(result);

      response.Messages = messages;

      return response;
    }

#region "APPEND"
    public override Stream GetRequestStream()
    {
      return EndGetRequestStream(BeginGetRequestStream(null, null));
    }

    public override IAsyncResult BeginGetRequestStream(AsyncCallback callback, object state)
    {
      if (!IsRequestAppend)
        // throws NotSupportedException
        return base.BeginGetRequestStream(callback, state);

      if (appendMessageBodyStream != null)
        throw new InvalidOperationException("already called");

      appendMessageBodyStream = new ImapAppendMessageBodyStream();

      // Content-Length
      if (0 < ContentLength)
        appendMessageBodyStream.SetLength(ContentLength);

      var session = GetSession();

      appendAsyncResult = session.BeginAppend(appendMessageBodyStream, DateTimeOffset.Now, null, (RequestUri as ImapUri).Mailbox);

      return appendAsyncResult;
    }

    public override Stream EndGetRequestStream(IAsyncResult asyncResult)
    {
      if (!IsRequestAppend)
        // throws NotSupportedException
        return base.EndGetRequestStream(asyncResult);

      if (asyncResult != appendAsyncResult)
        throw new ArgumentException("invalid IAsyncResult", "asyncResult");

      return appendMessageBodyStream;
    }

    private ImapWebResponse GetAppendResponse(ImapUri requestUri)
    {
      if (appendAsyncResult == null)
        throw new InvalidOperationException("GetRequestStream not called");

      ImapAppendedUidSet appendedUid;

      var result = Session.EndAppend(appendAsyncResult, out appendedUid);

      if (result.Failed)
        return new ImapWebResponse(result);

      appendMessageBodyStream.Dispose();

      var response = new ImapWebResponse(result);

      if (appendedUid != null)
        response.SetResponseUri(new ImapUri(string.Format("{0}/;UID={1}", requestUri, appendedUid.ToArray()[0])));

      return response;
    }

    private IAsyncResult appendAsyncResult = null;
    private ImapAppendMessageBodyStream appendMessageBodyStream = null;
#endregion

    private ImapWebResponse GetCreateResponse(ImapUri requestUri)
    {
      ImapMailbox createdMailbox;

      var result = Session.Create(requestUri.Mailbox, out createdMailbox);

      if (result.Failed)
        return new ImapWebResponse(result);

      var response = new ImapWebResponse(result);

      response.Mailboxes = new[] {createdMailbox};

      Session.Subscribe(createdMailbox);

      return response;
    }

    private ImapWebResponse GetDeleteResponse(ImapUri requestUri)
    {
      Session.Unsubscribe(requestUri.Mailbox);

      var result = Session.Delete(requestUri.Mailbox);

      if (result.Failed)
        return new ImapWebResponse(result);

      var response = new ImapWebResponse(result);

      return response;
    }

    private ImapWebResponse GetExpungeResponse(ImapUri requestUri)
    {
      var response = SelectMailbox(requestUri);

      if (response != null)
        return response;

      return new ImapWebResponse(Session.Expunge());
    }

    private ImapWebResponse GetExamineResponse(ImapUri requestUri)
    {
      return new ImapWebResponse(Session.Examine(requestUri.Mailbox));
    }

    private ImapWebResponse GetRenameResponse(ImapUri requestUri)
    {
      ImapMailbox renamedMailbox;

      if (DestinationUri.Scheme != requestUri.Scheme || DestinationUri.Server != requestUri.Server)
        throw new InvalidOperationException("destination must be same server");

      var result = Session.Rename(requestUri.Mailbox, GetDestinationUri(ImapUriForm.ListMessages).Mailbox, out renamedMailbox);

      if (result.Failed)
        return new ImapWebResponse(result);

      Session.Unsubscribe(requestUri.Mailbox);
      Session.Subscribe(renamedMailbox);

      var response = new ImapWebResponse(result);

      response.Mailboxes = new[] {renamedMailbox};

      return response;
    }

    private ImapWebResponse GetSubscribeResponse(ImapUri requestUri)
    {
      return new ImapWebResponse(Session.Subscribe(requestUri.Mailbox));
    }

    private ImapWebResponse GetUnsubscribeResponse(ImapUri requestUri)
    {
      return new ImapWebResponse(Session.Unsubscribe(requestUri.Mailbox));
    }
  }
}
