// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Text;

using Smdn.Protocols.Imap4.Client.Session;
using Smdn.Protocols.Imap4.Client.Transaction;

namespace Smdn.Protocols.Imap4.WebClient {
  internal class ImapSearchMessageWebRequest : ImapWebRequest {
    public ImapSearchMessageWebRequest(ImapUri requestUri, ImapSessionManager sessionManager)
      : base(requestUri, sessionManager)
    {
    }

    private delegate ImapWebResponse GetResponseDelegate(ImapUri requestUri, ImapMatchedSequenceSet matched);

    protected override ImapWebResponse InternalGetResponse(ImapUri requestUri)
    {
      var response = SelectMailbox(requestUri);

      if (response != null)
        return response;

      GetResponseDelegate getResponse = null;

      if (string.IsNullOrEmpty(Method) ||
          string.Equals(Method, ImapWebRequestMethods.Search, StringComparison.InvariantCultureIgnoreCase))
        getResponse = GetSearchResponse;
      else if (string.Equals(Method, ImapWebRequestMethods.Sort, StringComparison.InvariantCultureIgnoreCase))
        return GetSortResponse(requestUri);
      else if (string.Equals(Method, ImapWebRequestMethods.Thread, StringComparison.InvariantCultureIgnoreCase))
        return GetThreadResponse(requestUri);
      else if (string.Equals(Method, ImapWebRequestMethods.Copy, StringComparison.InvariantCultureIgnoreCase))
        getResponse = GetCopyResponse;
      else if (string.Equals(Method, ImapWebRequestMethods.Expunge, StringComparison.InvariantCultureIgnoreCase))
        getResponse = GetExpungeResponse;
      else if (string.Equals(Method, ImapWebRequestMethods.Store, StringComparison.InvariantCultureIgnoreCase))
        getResponse = GetStoreResponse;
      else
        throw new NotSupportedException(string.Format("unsupported request method: {0}", Method));

      ImapTransactionResult result;
      ImapMatchedSequenceSet matched;

      if (Session.ServerCapabilities.Has(ImapCapability.Searchres))
        // use SEARCHRES extension
        result = Session.UidESearch(requestUri.GetSearchCriteria(), Encoding.UTF8, ImapSearchResultOptions.Save, out matched);
      else
        result = Session.UidSearch(requestUri.GetSearchCriteria(), Encoding.UTF8, out matched);

      if (result.Failed)
        return new ImapWebResponse(result);
      else
        return getResponse(requestUri, matched);
    }

    private ImapWebResponse GetSearchResponse(ImapUri requestUri, ImapMatchedSequenceSet matched)
    {
      ImapMessage[] messages;

      var result = Session.Fetch(matched, ImapFetchDataItem.All, out messages);

      if (result.Failed)
        return new ImapWebResponse(result);

      var response = new ImapWebResponse(result);

      response.Messages = messages;

      return response;
    }

    private ImapWebResponse GetSortResponse(ImapUri requestUri)
    {
      ImapMatchedSequenceSet matched;

      var result = Session.UidSort(SortCriteria, requestUri.GetSearchCriteria(), Encoding.UTF8, out matched);

      ImapMessage[] messages;

      result = Session.Fetch(matched, ImapFetchDataItem.All, out messages);

      if (result.Failed)
        return new ImapWebResponse(result);

      var sortedUids = matched.ToArray();
      var sortedMessages = new ImapMessage[sortedUids.Length];

      for (var i = 0; i < sortedUids.Length; i++) {
        sortedMessages[i] = null;

        for (var j = 0; j < messages.Length; j++) {
          if (sortedUids[i] == messages[j].Uid) {
            sortedMessages[i] = messages[j];
            break;
          }
        }
      }

      var response = new ImapWebResponse(result);

      response.Messages = sortedMessages;

      return response;
    }

    private ImapWebResponse GetThreadResponse(ImapUri requestUri)
    {
      ImapThreadList threadList;

      var result = Session.UidThread(ThreadingAlgorithm, requestUri.GetSearchCriteria(), out threadList);

      if (result.Failed)
        return new ImapWebResponse(result);

      ImapMessage[] messages;

      result = Session.Fetch(threadList.ToSequenceSet(), ImapFetchDataItem.All, out messages);

      if (result.Failed)
        return new ImapWebResponse(result);

      var response = new ImapWebResponse(result);

      var messageDictionary = new Dictionary<long, ImapMessage>();

      foreach (var message in messages) {
        messageDictionary.Add(message.Uid, message);
      }

      response.ThreadTree = new ImapThreadTree(true, null, CreateMessageTree(threadList.Children, messageDictionary));

      return response;
    }

    private ImapThreadTree[] CreateMessageTree(ImapThreadList[] children, Dictionary<long, ImapMessage> messages)
    {
      return Array.ConvertAll(children, delegate(ImapThreadList list) {
        return new ImapThreadTree(false, messages[list.Number], CreateMessageTree(list.Children, messages));
      });
    }

    private ImapWebResponse GetCopyResponse(ImapUri requestUri, ImapMatchedSequenceSet matched)
    {
      ImapCopiedUidSet copiedUids;

      if (DestinationUri.Scheme != requestUri.Scheme || DestinationUri.Server != requestUri.Server)
        throw new InvalidOperationException("destination must be same server");

      var result = Session.Copy(matched, GetDestinationUri(ImapUriForm.ListMessages).Mailbox, out copiedUids);

      if (result.Failed)
        return new ImapWebResponse(result);

      var response = new ImapWebResponse(result);

      // XXX: copiedUids
      //response.Messages = messages;

      return response;
    }

    private ImapWebResponse GetExpungeResponse(ImapUri requestUri, ImapMatchedSequenceSet matched)
    {
      var result = Session.Store(matched, ImapStoreDataItem.AddFlagsSilent(ImapMessageFlag.Deleted));

      if (result.Failed)
        return new ImapWebResponse(result);

      if (Session.ServerCapabilities.Has(ImapCapability.UidPlus))
        result = Session.UidExpunge(matched);
      else
        result = Session.Expunge();

      return new ImapWebResponse(result);
    }

    private ImapWebResponse GetStoreResponse(ImapUri requestUri, ImapMatchedSequenceSet matched)
    {
      return new ImapWebResponse(Session.Store(matched, StoreDataItem));
    }
  }
}
