// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;

using Smdn.Protocols.Imap4.Client.Protocol;
using Smdn.Protocols.Imap4.Client.Session;
using Smdn.Protocols.Imap4.Client.Transaction;

namespace Smdn.Protocols.Imap4.WebClient {
  public class ImapSessionManager : IWebRequestCreate {
    private static readonly ImapSessionManager instance = new ImapSessionManager();
    private static readonly NetworkCredential AnonymousCredential = new NetworkCredential("anonymous", string.Empty);

    public static X509Certificate2Collection CertCollection {
      get { return ImapConnection.CertCollection; }
    }

    public static RemoteCertificateValidationCallback CertificateValidationCallback {
      get { return ImapConnection.CertificateValidationCallback; }
      set { ImapConnection.CertificateValidationCallback = value; }
    }

    public static LocalCertificateSelectionCallback CertificateSelectionCallback {
      get { return ImapConnection.CertificateSelectionCallback; }
      set { ImapConnection.CertificateSelectionCallback = value; }
    }

    private static ImapUpgradeConnectionStreamCallback createSslStreamCallback = ImapConnection.CreateSslStream;

    public static ImapUpgradeConnectionStreamCallback CreateSslStreamCallback {
      get { return createSslStreamCallback; }
      set
      {
        if (createSslStreamCallback == null)
          throw new ArgumentNullException("CreateSslStreamCallback");
        createSslStreamCallback = value;
      }
    }

    private static Dictionary<string, string> clientID;

    public static Dictionary<string, string> ClientID {
      get { return clientID; }
      set { clientID = value; }
    }

    static ImapSessionManager()
    {
      /*
       * 3.3. Defined Field Values
       *    Any string may be sent as a field, but the following are defined to
       *    describe certain values that might be sent.
       * 
       *      name            Name of the program
       *      version         Version number of the program
       *      os              Name of the operating system
       *      os-version      Version of the operating system
       *      vendor          Vendor of the client/server
       *      support-url     URL to contact for support
       *      address         Postal address of contact/vendor
       *      date            Date program was released, specified as a date-time
       *                        in IMAP4rev1
       *      command         Command used to start the program
       *      arguments       Arguments supplied on the command line, if any
       *                        if any
       *      environment     Description of environment, i.e., UNIX environment
       *                        variables or Windows registry settings
       */
      var assembly = System.Reflection.Assembly.GetExecutingAssembly();

      clientID = new Dictionary<string, string>();
      clientID.Add("name", assembly.GetName().Name);
      clientID.Add("version", assembly.GetName().Version.ToString());
      clientID.Add("environment", string.Format("{0} {1}",
                                                EnvironmentExtension.IsRunningOnMono ? "Mono" : ".NET Framework",
                                                Environment.Version));
    }

    public static bool RegisterWebRequestPrefix()
    {
      return WebRequest.RegisterPrefix(ImapUri.UriSchemeImap, instance) | WebRequest.RegisterPrefix(ImapUri.UriSchemeImaps, instance);
    }

    public static void DisconnectFrom(ImapUri serverUri)
    {
      if (serverUri == null)
        throw new ArgumentNullException("serverUri");

      if (serverUri.Port == -1) {
        if (serverUri.Scheme == ImapUri.UriSchemeImap)
          DisconnectFrom(serverUri.Host, ImapConnection.ImapDefaultPort);
        else
          DisconnectFrom(serverUri.Host, ImapConnection.ImapsDefaultPort);
      }
      else {
        DisconnectFrom(serverUri.Host, serverUri.Port);
      }
    }

    public static void DisconnectFrom(string host,  int port)
    {
      instance.InternalDisconnectFrom(host, port);
    }

    private ImapSessionManager()
    {
    }

    WebRequest IWebRequestCreate.Create(Uri uri)
    {
      var requestUri = new ImapUri(uri);
      ImapWebRequest request;

      switch (requestUri.Form) {
        case ImapUriForm.Server:
          request = new ImapServerWebRequest(requestUri, this);
          break;

        case ImapUriForm.ListMessages:
          request = new ImapMailboxWebRequest(requestUri, this);
          break;

        case ImapUriForm.FetchMessage:
          request = new ImapFetchMessageWebRequest(requestUri, this);
          break;

        case ImapUriForm.SearchMessages:
          request = new ImapSearchMessageWebRequest(requestUri, this);
          break;

        default: // case ImapUriForm.Relative:
          throw new ArgumentException("relative URI is invalid", "uri");
      }

      request.Method = ImapWebRequestMethods.GetDefaultMethod(requestUri.Form);

      return request;
    }

    internal ImapSession CreateSession(IImapSessionProfile profile)
    {
      ImapSession session = null;

      var serverUri = profile.ServerUri;
      var username = serverUri.UserName;
      var authMechanism = serverUri.Authentication;

      for (var referral = 0;; referral++) {
        try {
          if (serverUri.Scheme == ImapUri.UriSchemeImap)
            session = new ImapSession(serverUri.Host, serverUri.Port, true, profile.Timeout, null);
          else
            session = new ImapSession(serverUri.Host, serverUri.Port, true, profile.Timeout, createSslStreamCallback);
        }
        catch (ImapConnectionException ex) {
          throw new WebException("connection error", ex, WebExceptionStatus.ConnectFailure, null);
        }

        session.HandlesIncapableAsException = false;
        session.HandlesReferralAsException = true;
        session.Credentials = profile.Credentials;
        session.TransactionTimeout = profile.Timeout;
        session.SendTimeout = profile.Timeout;
        session.ReceiveTimeout = profile.Timeout;

        if (profile.UseTlsIfAvailable && session.ServerCapabilities.Has(ImapCapability.StartTls)) {
          var result = session.StartTls(createSslStreamCallback);

          if (result.Failed)
            throw new WebException(result.Description, null, WebExceptionStatus.ConnectFailure, null);
        }

        if (profile.UseDeflateIfAvailable && session.ServerCapabilities.Has(ImapCapability.CompressDeflate)) {
          var result = session.Compress(ImapCompressionMechanism.Deflate);

          if (result.Failed && result.GetResponseCode(ImapResponseCode.CompressionActive) == null)
            throw new WebException(result.Description, null, WebExceptionStatus.ProtocolError, null);
        }

        try {
          Authenticate(session, username, authMechanism);
        }
        catch (ImapLoginReferralException ex) {
          if (profile.AllowLoginRedirect && referral == 0) {
            serverUri = ex.ReferToUri;
            continue;
          }
          else {
            throw new WebException("authentication failed", ex, WebExceptionStatus.ProtocolError, null);
          }
        }

        if (session.State == ImapSessionState.Authenticated)
          break;

        try {
          session.Disconnect(false);
        }
        catch (ImapConnectionException) {
          // ignore
        }

        throw new WebException("authentication failed", WebExceptionStatus.ProtocolError);
      }

      // update server info
      if (session.Capability().Succeeded) {
        if (!session.ServerCapabilities.Has(ImapCapability.Imap4Rev1))
          throw new WebException("IMAP4rev1 not supported", WebExceptionStatus.ProtocolError);
      }
      else {
        throw new WebException("capability failed", WebExceptionStatus.ProtocolError);
      }

      if (session.ServerCapabilities.Has(ImapCapability.Namespace))
        session.Namespace();

      if (session.ServerCapabilities.Has(ImapCapability.ID))
        session.ID(clientID);

      return session;
    }

    private void Authenticate(ImapSession session, string username, ImapAuthenticationMechanism authMechanism)
    {
      // update server capability
      if (!session.ServerCapabilities.Has(ImapCapability.Imap4Rev1)) {
        var result = session.Capability();

        if (result.Succeeded) {
          if (!session.ServerCapabilities.Has(ImapCapability.Imap4Rev1))
            throw new WebException("IMAP4rev1 not supported", WebExceptionStatus.ProtocolError);
        }
        else {
          throw new WebException(result.Description, WebExceptionStatus.ProtocolError);
        }
      }

      /*
       * http://tools.ietf.org/html/rfc5092
       * 3.2. IMAP User Name and Authentication Mechanism
       * 
       *    If no user name and no authentication mechanism are supplied, the
       *    client MUST authenticate as anonymous to the server.  If the server
       *    advertises AUTH=ANONYMOUS IMAP capability, the client MUST use the
       *    AUTHENTICATE command with ANONYMOUS [ANONYMOUS] SASL mechanism.  If
       *    SASL ANONYMOUS is not available, the (case-insensitive) user name
       *    "anonymous" is used with the "LOGIN" command and the Internet email
       *    address of the end user accessing the resource is supplied as the
       *    password.  The latter option is given in order to provide for
       *    interoperability with deployed servers.
       * 
       *    Note that, as described in RFC 3501, the "LOGIN" command MUST NOT be
       *    used when the IMAP server advertises the LOGINDISABLED capability.
       */

      /*
       *    If a user name is included with no authentication mechanism, then
       *    ";AUTH=*" is assumed.
       */
      if (authMechanism == null)
        authMechanism = ImapAuthenticationMechanism.SelectAppropriate;

      // TODO: URLAUTH
      /*
      if (!session.ServerCapabilities.Has(ImapCapability.UrlAuth)
        throw new ImapWebException("URLAUTH incapable", WebExceptionStatus.ProtocolError);
      */

      var loginDisabled = session.ServerCapabilities.Has(ImapCapability.LoginDisabled);

      if (authMechanism == ImapAuthenticationMechanism.SelectAppropriate) {
        // AUTH=*
        ImapTransactionResult result = null;

        foreach (var mechanism in new[] {
          ImapAuthenticationMechanism.DigestMD5,
          ImapAuthenticationMechanism.CRAMMD5,
          ImapAuthenticationMechanism.NTLM,
        }) {
          if (session.ServerCapabilities.Has((mechanism as IImapExtension).RequiredCapability)) {
            result = session.Authenticate(username, mechanism);

            if (result.Succeeded)
              break;
          }
        }

        if (result == null)
          result = session.Login(username);

        if (result == null)
          throw new WebException("no appropriate authentication mechanism supported", WebExceptionStatus.ProtocolError);
        else if (result.Failed)
          throw new WebException(result.Description, WebExceptionStatus.ProtocolError);
      }
      else if (authMechanism == ImapAuthenticationMechanism.Anonymous || string.IsNullOrEmpty(username)) {
        if (session.ServerCapabilities.Has(ImapCapability.AuthAnonymous)) {
          // try AUTH=ANONYUMOUS
          try {
            session.Credentials = AnonymousCredential;

            var result = session.Authenticate(ImapAuthenticationMechanism.Anonymous);

            if (result.Failed)
              throw new WebException(result.Description, WebExceptionStatus.ProtocolError);
          }
          finally {
            session.Credentials = null;
          }
        }
        else if (!loginDisabled) {
          // try anonymous LOGIN
          var creds = session.Credentials;

          try {
            session.Credentials = AnonymousCredential;

            var result = session.Login();

            if (result.Failed)
              throw new WebException(result.Description, WebExceptionStatus.ProtocolError);
          }
          finally {
            session.Credentials = creds;
          }
        }
      }
      else {
        var result = session.Authenticate(username, authMechanism);

        if (result.Failed)
          throw new WebException(result.Description, WebExceptionStatus.ProtocolError);
      }
    }

    internal ImapSession GetExistSession(ImapUri uri)
    {
      var serverUri = uri.ServerUri;

      if (sessions.ContainsKey(serverUri))
        return sessions[serverUri];
      else
        return null;
    }

    internal void RegisterSession(ImapSession session)
    {
      var serverUri = session.ServerUrl;

      ImapSession existSession = null;

      lock ((sessions as System.Collections.ICollection).SyncRoot) {
        if (sessions.ContainsKey(serverUri)) {
          existSession = sessions[serverUri];
          sessions.Remove(serverUri);
        }

        sessions.Add(session.ServerUrl, session);
      }

      if (existSession != null)
        existSession.Disconnect(true);
    }

    internal void UnregisterSession(ImapUri uri)
    {
      var serverUri = uri.ServerUri;

      lock ((sessions as System.Collections.ICollection).SyncRoot) {
        if (!sessions.ContainsKey(serverUri))
          return;

        sessions.Remove(serverUri);
      }
    }

    private void InternalDisconnectFrom(string host, int port)
    {
      if (host == null)
        throw new ArgumentNullException("host", host);
      if (port < IPEndPoint.MinPort || IPEndPoint.MaxPort < port)
        throw new ArgumentOutOfRangeException("port");

      var keys = new List<ImapUri>();

      foreach (var pair in sessions) {
        var session = pair.Value;

        if (string.Equals(session.Connection.Host, host, StringComparison.InvariantCultureIgnoreCase) &&
            session.Connection.Port == port) {
          keys.Add(pair.Key);

          session.Disconnect(true);
        }
      }

      foreach (var key in keys) {
        UnregisterSession(key);
      }
    }

    private Dictionary<ImapUri, ImapSession> sessions = new Dictionary<ImapUri, ImapSession>();
  }
}