// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;
using System.Threading;

using Smdn.Protocols.Imap4.Client.Session;
using Smdn.Protocols.Imap4.Client.Transaction;
using Smdn.Protocols.Imap4.Client.Protocol;

using Threading = System.Threading;

namespace Smdn.Protocols.Imap4.WebClient {
  public abstract class ImapWebRequest : WebRequest, IImapSessionProfile {
    private ImapSessionManager sessionManager;

    private ImapSession session = null;

    protected ImapSession Session {
      get { return session; }
    }

    private ICredentialsByHost credentials = null;

    /// <value>value must implement ICredentialsByHost</value>
    public override ICredentials Credentials {
      get { return credentials as ICredentials; }
      set
      {
        if (value != null && !(value is ICredentialsByHost))
          throw new ArgumentException("value must be ICredentialsByHost");
        credentials = value as ICredentialsByHost;
      }
    }

    ICredentialsByHost IImapSessionProfile.Credentials {
      get { return credentials; }
    }

    private ImapUri requestUri;

    public override Uri RequestUri {
      get { return requestUri; }
    }

    ImapUri IImapSessionProfile.ServerUri {
      get { return requestUri.ServerUri; }
    }

    private int timeout = Threading.Timeout.Infinite;

    public override int Timeout {
      get { return timeout; }
      set
      {
        if (value < -1)
          throw new ArgumentOutOfRangeException("must be greater than or equals to -1");
        timeout = value;
      }
    }

    public override string Method {
      get; set;
    }

    /// <value>be used when method is COPY or RENAME</value>
    public ImapUri DestinationUri {
      get; set;
    }

    /// <value>be used when method is STORE</value>
    public ImapStoreDataItem StoreDataItem {
      get; set;
    }

    /// <value>be used when method is SORT</value>
    public ImapSortCriteria SortCriteria {
      get; set;
    }

    /// <value>be used when method is THREAD</value>
    public ImapThreadingAlgorithm ThreadingAlgorithm {
      get; set;
    }

    /// <value>be used when method is APPEND</value>
    public override long ContentLength {
      get; set;
    }

    public bool AllowLoginRedirect {
      get; set;
    }

    public bool UseTlsIfAvailable {
      get; set;
    }

    public bool UseDeflateIfAvailable {
      get; set;
    }

    public bool KeepAlive {
      get; set;
    }

    protected ImapWebRequest(ImapUri requestUri, ImapSessionManager sessionManager)
      : base()
    {
      this.requestUri = requestUri;
      this.sessionManager = sessionManager;
      this.Method = null;
      this.DestinationUri = null;
      this.ContentLength = 0;
      this.StoreDataItem = null;
      this.SortCriteria = null;
      this.ThreadingAlgorithm = null;
      this.AllowLoginRedirect = true;
      this.UseTlsIfAvailable = true;
      this.UseDeflateIfAvailable = false;
      this.KeepAlive = true;
    }

    public override WebResponse GetResponse()
    {
      return EndGetResponse(BeginGetResponse(null, null));
    }

    private delegate ImapWebResponse GetResponseDelegate();

    public override IAsyncResult BeginGetResponse(AsyncCallback callback, object state)
    {
      return (new GetResponseDelegate(InternalGetResponse)).BeginInvoke(null, null);
    }

    public override WebResponse EndGetResponse(IAsyncResult asyncResult)
    {
      var ar = asyncResult as System.Runtime.Remoting.Messaging.AsyncResult;

      if (ar == null)
        throw new ArgumentException("invalid IAsyncResult", "asyncResult");

      return (ar.AsyncDelegate as GetResponseDelegate).EndInvoke(ar);
    }

    protected abstract ImapWebResponse InternalGetResponse(ImapUri requestUri);

    private ImapWebResponse InternalGetResponse()
    {
      try {
        GetSession();

        var response = InternalGetResponse(requestUri);

        if (response.ResponseUri == null)
          response.SetResponseUri(requestUri);

        response.Session = session;
        response.SessionManager = sessionManager;
        response.KeepAlive = KeepAlive;

        if (response.Result.Succeeded)
          return response;
        else
          throw new WebException(response.ResponseDescription, null, WebExceptionStatus.ProtocolError, response);
      }
      finally {
        session = null;
      }
    }

    protected ImapUri GetDestinationUri(ImapUriForm requiredForm)
    {
      var destination =  DestinationUri;

      if (destination.Form == ImapUriForm.Relative)
        destination = new ImapUri(requestUri, destination);

      if (destination.Form == requiredForm)
        return destination;
      else
        throw new InvalidOperationException(string.Format("invalid form of destination URI, required form is {0} but was {1}", requiredForm, destination.Form));
    }

    protected ImapSession GetSession()
    {
      var s = sessionManager.GetExistSession(requestUri);

      if (s != null && !s.IsTransactionProceeding && s.NoOp().Failed && s.State == ImapSessionState.NotConnected) {
        // disconnected
        sessionManager.UnregisterSession(requestUri);
        s = null;
      }

      if (s == null) {
        var session = sessionManager.CreateSession(this);

        if (KeepAlive)
          sessionManager.RegisterSession(session);

        if (requestUri.ServerUri != session.ServerUrl) {
          var pathAndQuery = requestUri.ToString().Substring(requestUri.GetLeftPart(UriPartial.Authority).Length);

          requestUri = new ImapUri(session.ServerUrl, pathAndQuery);
        }

        s = session;
      }

      this.session = s;

      return s;
    }

    protected ImapWebResponse SelectMailbox(ImapUri mailboxUri)
    {
      var mailbox = mailboxUri.Mailbox;

      if (Session.State == ImapSessionState.Selected && !string.Equals(Session.SelectedMailbox.Name, mailbox)) {
        var resp = UnselectMailbox();

        if (resp != null)
          return resp;
      }

      if (Session.State != ImapSessionState.Selected) {
        var result = Session.Select(mailbox);

        if (result.Failed)
          return new ImapWebResponse(result);
      }

      return null;
    }

    protected ImapWebResponse UnselectMailbox()
    {
      if (Session.State != ImapSessionState.Selected)
        return null;

      if (Session.ServerCapabilities.Has(ImapCapability.Unselect)) {
        var result = Session.Unselect();

        if (result.Failed)
          return new ImapWebResponse(result);
      }
      else {
        var result = Session.Close();

        if (result.Failed)
          return new ImapWebResponse(result);
      }

      return null;
    }
  }
}
