// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

namespace Smdn.Protocols.Imap4 {
  // enum types:
  //   ImapStringEnum : ImapString
  //     => handles string constants
  //   * ImapAuthenticationMechanism
  //       => handles authentication mechanisms
  //     ImapCapability
  //       => handles capability constants
  //     ImapCompressionMechanism
  //       => handles COMPRESS extension compression mechanism
  //     ImapMailboxFlag
  //       => handles mailbox flags
  //     ImapMessageFlag
  //       => handles system flags and keywords
  //     ImapThreadingAlgorithm
  //       => handles THREAD extension threading algorithm
  //     Protocol.ImapResponseCode
  //       => handles response codes
  //     Protocol.ImapDataResponseType
  //       => handles server response types

  public sealed class ImapAuthenticationMechanism : ImapStringEnum, IImapExtension {
    public static readonly ImapStringEnumList<ImapAuthenticationMechanism> AllMechanisms;

    public static readonly ImapAuthenticationMechanism KerberosV4
      = new ImapAuthenticationMechanism(ImapCapability.AuthKerberosV4);
    public static readonly ImapAuthenticationMechanism CRAMMD5
      = new ImapAuthenticationMechanism(ImapCapability.AuthCRAMMD5);
    public static readonly ImapAuthenticationMechanism OTP
      = new ImapAuthenticationMechanism(ImapCapability.AuthOTP);
    public static readonly ImapAuthenticationMechanism SKey
      = new ImapAuthenticationMechanism(ImapCapability.AuthSKey);
    public static readonly ImapAuthenticationMechanism Plain
      = new ImapAuthenticationMechanism(ImapCapability.AuthPlain);
    public static readonly ImapAuthenticationMechanism SecureID
      = new ImapAuthenticationMechanism(ImapCapability.AuthSecureID);
    public static readonly ImapAuthenticationMechanism DigestMD5
      = new ImapAuthenticationMechanism(ImapCapability.AuthDigestMD5);
    public static readonly ImapAuthenticationMechanism IsoIec9798_U_RSA_SHA1_ENC
      = new ImapAuthenticationMechanism(ImapCapability.AuthIsoIec9798_U_RSA_SHA1_ENC);
    public static readonly ImapAuthenticationMechanism IsoIec9798_M_RSA_SHA1_ENC
      = new ImapAuthenticationMechanism(ImapCapability.AuthIsoIec9798_M_RSA_SHA1_ENC);
    public static readonly ImapAuthenticationMechanism IsoIec9798_U_DSA_SHA1
      = new ImapAuthenticationMechanism(ImapCapability.AuthIsoIec9798_U_DSA_SHA1);
    public static readonly ImapAuthenticationMechanism IsoIec9798_M_DSA_SHA1
      = new ImapAuthenticationMechanism(ImapCapability.AuthIsoIec9798_M_DSA_SHA1);
    public static readonly ImapAuthenticationMechanism IsoIec9798_U_ECDSA_SHA1
      = new ImapAuthenticationMechanism(ImapCapability.AuthIsoIec9798_U_ECDSA_SHA1);
    public static readonly ImapAuthenticationMechanism IsoIec9798_M_ECDSA_SHA1
      = new ImapAuthenticationMechanism(ImapCapability.AuthIsoIec9798_M_ECDSA_SHA1);
    public static readonly ImapAuthenticationMechanism External
      = new ImapAuthenticationMechanism(ImapCapability.AuthExternal);
    public static readonly ImapAuthenticationMechanism Anonymous
      = new ImapAuthenticationMechanism(ImapCapability.AuthAnonymous);
    public static readonly ImapAuthenticationMechanism Gssapi
      = new ImapAuthenticationMechanism(ImapCapability.AuthGssapi);
    public static readonly ImapAuthenticationMechanism KerberosV5
      = new ImapAuthenticationMechanism(ImapCapability.AuthKerberosV5);
    public static readonly ImapAuthenticationMechanism NTLM
      = new ImapAuthenticationMechanism(ImapCapability.AuthNTLM);

    public static readonly ImapAuthenticationMechanism Login
      = new ImapAuthenticationMechanism(ImapCapability.AuthLogin);

    /*
     * RFC 5092 IMAP URL Scheme
     * http://tools.ietf.org/html/rfc5092
     */
    public static readonly ImapAuthenticationMechanism SelectAppropriate = new ImapAuthenticationMechanism("*");

    static ImapAuthenticationMechanism()
    {
      AllMechanisms = new ImapStringEnumList<ImapAuthenticationMechanism>(true, GetDefinedConstants<ImapAuthenticationMechanism>());
    }

    internal static ImapAuthenticationMechanism GetKnownOrCreate(string mechanism)
    {
      if (AllMechanisms.Has(mechanism))
        return AllMechanisms[mechanism];

      Trace.Verbose("unknown authentication mechanism: {0}", mechanism);

      return new ImapAuthenticationMechanism(mechanism);
    }

    ImapCapability IImapExtension.RequiredCapability {
      get { return requiredCapability; }
    }

    public ImapAuthenticationMechanism(string name)
      : this(name, null)
    {
    }

    public ImapAuthenticationMechanism(string name, ImapCapability requiredCapability)
      : base(name)
    {
      if (name == null)
        throw new ArgumentNullException("name");

      this.requiredCapability = requiredCapability;
    }

    private ImapAuthenticationMechanism(ImapAuthenticationCapability requiredCapability)
      : base(requiredCapability.Name)
    {
      this.requiredCapability = requiredCapability;
    }

    private /*readonly*/ ImapCapability requiredCapability;
  }
}