// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Text;

namespace Smdn.Protocols.Imap4 {
  public static class ImapEncoding {
#region "TransferEncoding class"
    private class TransferEncoding : Encoding {
      public TransferEncoding(int bits)
      {
        if (bits < 0 || 8 < bits)
          throw new ArgumentOutOfRangeException("bits");

        maxValue = (char)(1 << bits);
      }

      public override int GetMaxCharCount(int byteCount)
      {
        if (byteCount < 0)
          throw new ArgumentOutOfRangeException("byteCount");

        return byteCount;
      }

      public override int GetMaxByteCount(int charCount)
      {
        if (charCount < 0)
          throw new ArgumentOutOfRangeException("charCount");

        return charCount;
      }

      public override int GetByteCount(char[] chars, int index, int count)
      {
        if (index < 0)
          throw new ArgumentOutOfRangeException("index");
        if (count < 0)
          throw new ArgumentOutOfRangeException("count");

        if (EncoderFallback == null)
          return count - index;

        EncoderFallbackBuffer buffer = null;
        var byteCount = 0;

        for (;index < count; index++) {
          if (chars[index] < maxValue) {
            byteCount++;
          }
          else {
            if (buffer == null)
              buffer = EncoderFallback.CreateFallbackBuffer();

            buffer.Fallback(chars[index], index);

            var fallbackChars = new char[buffer.Remaining];

            for (var r = 0; r < fallbackChars.Length; r++) {
              fallbackChars[r] = buffer.GetNextChar();
            }

            byteCount += GetByteCount(fallbackChars, 0, fallbackChars.Length);
          }
        }

        return byteCount;
      }

      public override int GetCharCount(byte[] bytes, int index, int count)
      {
        if (index < 0)
          throw new ArgumentOutOfRangeException("index");
        if (count < 0)
          throw new ArgumentOutOfRangeException("count");

        return count;
      }

      public override int GetBytes(char[] chars, int charIndex, int charCount, byte[] bytes, int byteIndex)
      {
        if (chars == null)
          throw new ArgumentNullException("chars");
        if (bytes == null)
          throw new ArgumentNullException("chars");
        if (charIndex < 0)
          throw new ArgumentOutOfRangeException("charIndex");
        if (chars.Length < charIndex + charCount)
          throw new ArgumentOutOfRangeException("charCount");
        if (byteIndex < 0)
          throw new ArgumentOutOfRangeException("byteIndex");

        var byteCount = 0;

        for (var i = 0; i < charCount; i++, charIndex++, byteIndex++, byteCount++) {
          if (chars[charIndex] < maxValue) {
            bytes[byteIndex] = (byte)chars[charIndex];
            byteCount++;
          }
          else {
            if (EncoderFallback == null) {
              bytes[byteIndex] = (byte)'?';
              byteCount++;
            }
            else {
              var buffer = EncoderFallback.CreateFallbackBuffer();

              if (buffer.Fallback(chars[charIndex], charIndex)) {
                var fallbackChars = new char[buffer.Remaining];

                for (var r = 0; r < fallbackChars.Length; r++) {
                  fallbackChars[r] = buffer.GetNextChar();
                }

                var c = GetBytes(fallbackChars, 0, fallbackChars.Length, bytes, byteIndex);

                byteIndex += c - 1;
                byteCount += c;
              }
              else {
                bytes[byteIndex] = (byte)'?';
                byteCount++;
              }
            }
          }
        }

        return byteCount;
      }

      public override int GetChars(byte[] bytes, int byteIndex, int byteCount, char[] chars, int charIndex)
      {
        if (bytes == null)
          throw new ArgumentNullException("bytes");
        if (chars == null)
          throw new ArgumentNullException("chars");
        if (byteIndex < 0)
          throw new ArgumentOutOfRangeException("byteIndex");
        if (bytes.Length < byteIndex + byteCount)
          throw new ArgumentOutOfRangeException("byteCount");
        if (charIndex < 0)
          throw new ArgumentOutOfRangeException("charIndex");

        var charCount = 0;

        for (var i = 0; i < byteCount; i++, byteIndex++, charIndex++, charCount++) {
          chars[charIndex] = (char)bytes[byteIndex];
        }

        return charCount;
      }

      private readonly char maxValue;
    }
#endregion

    // for string which contains 7-bit characters
    public static readonly Encoding Transfer7Bit;

    // for string which contains 8-bit characters
    public static readonly Encoding Transfer8Bit;

    // US-ASCII
    private static readonly Encoding USAscii;

    static ImapEncoding()
    {
      Transfer7Bit = (new TransferEncoding(7)).Clone() as Encoding;
      Transfer7Bit.DecoderFallback = new DecoderExceptionFallback();
      Transfer7Bit.EncoderFallback = new EncoderExceptionFallback();

      Transfer8Bit = (new TransferEncoding(8)).Clone() as Encoding;
      Transfer8Bit.DecoderFallback = new DecoderExceptionFallback();
      Transfer8Bit.EncoderFallback = new EncoderExceptionFallback();

      USAscii = Encoding.ASCII.Clone() as Encoding;
      USAscii.DecoderFallback = new DecoderExceptionFallback();
      USAscii.EncoderFallback = new EncoderExceptionFallback();
    }

    public static string ToBase64String(string str)
    {
      return Convert.ToBase64String(USAscii.GetBytes(str));
    }

    public static string ToBase64String(string str, Encoding encoding)
    {
      return Convert.ToBase64String(encoding.GetBytes(str));
    }

    public static string FromBase64String(string str)
    {
      return USAscii.GetString(Convert.FromBase64String(str));
    }

    public static string FromBase64String(string str, Encoding encoding)
    {
      return encoding.GetString(Convert.FromBase64String(str));
    }

    internal static string FromModifiedUTF7StringNoException(string str)
    {
      try {
        return FromModifiedUTF7String(str);
      }
      catch (EncoderFallbackException) {
        return str;
      }
      catch (FormatException) {
        return str;
      }
    }

    public static string FromModifiedUTF7String(string str)
    {
      // RFC 3501 INTERNET MESSAGE ACCESS PROTOCOL - VERSION 4rev1
      // 5.1.3. Mailbox International Naming Convention
      // http://tools.ietf.org/html/rfc3501#section-5.1.3
      if (!str.Contains("&"))
        return str;

      var bytes = USAscii.GetBytes(str);
      var decoded = new StringBuilder();

      for (var index = 0; index < bytes.Length; index++) {
        // In modified UTF-7, printable US-ASCII characters, except for "&",
        // represent themselves
        // "&" is used to shift to modified BASE64
        if (bytes[index] != 0x26) { // '&'
          decoded.Append((char)bytes[index]);
          continue;
        }

        if (bytes.Length <= ++index)
          // incorrect form
          throw new FormatException("incorrect form");

        if (bytes[index] == 0x2d) { // '-'
          // The character "&" (0x26) is represented by the two-octet sequence "&-".
          decoded.Append('&');
          continue;
        }

        var nonprintable = new StringBuilder();

        for (; index < bytes.Length; index++) {
          if (bytes[index] == 0x2d) // '-'
            // "-" is used to shift back to US-ASCII
            break;

          nonprintable.Append((char)bytes[index]);
        }

        // modified UTF7 -> string
        decoded.Append(FromModifiedUTF7(nonprintable.ToString()));
      }

      return decoded.ToString();
    }

    private static string FromModifiedUTF7(string str)
    {
      return Encoding.BigEndianUnicode.GetString(FromModifiedBase64(str));
    }

    private static byte[] FromModifiedBase64(string str)
    {
      // "," is used instead of "/"
      str = str.Replace(',', '/');

      var padding = 4 - str.Length & 3;

      if (padding == 4)
        return Convert.FromBase64String(str);
      else if (padding == 3)
        throw new FormatException("incorrect form");
      else
        return Convert.FromBase64String(str + (new string('=', padding)));
    }

    public static string ToModifiedUTF7String(string str)
    {
      // RFC 3501 INTERNET MESSAGE ACCESS PROTOCOL - VERSION 4rev1
      // 5.1.3. Mailbox International Naming Convention
      // http://tools.ietf.org/html/rfc3501#section-5.1.3

      var encoded = new StringBuilder();
      var shiftFrom = -1;

      for (var index = 0; index < str.Length; index++) {
        var c = str[index];

        if ((0x20 <= c && c <= 0x7e)) {
          if (0 <= shiftFrom) {
            // string -> modified UTF7
            encoded.Append('&');
            encoded.Append(ToModifiedUTF7(str.Substring(shiftFrom, index - shiftFrom)));
            encoded.Append('-');

            shiftFrom = -1;
          }

          // printable US-ASCII characters
          if (c == 0x26)
            // except for "&"
            encoded.Append("&-");
          else
            encoded.Append(c);
        }
        else {
          if (shiftFrom == -1)
            shiftFrom = index;
        }
      }

      if (0 <= shiftFrom) {
        // string -> modified UTF7
        encoded.Append('&');
        encoded.Append(ToModifiedUTF7(str.Substring(shiftFrom)));
        encoded.Append('-');
      }

      return encoded.ToString();
    }

    private static string ToModifiedUTF7(string str)
    {
      return ToModifiedBase64(Encoding.BigEndianUnicode.GetBytes(str));
    }

    private static string ToModifiedBase64(byte[] bytes)
    {
      var base64 = Convert.ToBase64String(bytes).Replace('/', ',');
      var padding = base64.IndexOf('=');

      if (0 <= padding)
        return base64.Substring(0, padding);
      else
        return base64;
    }
  }
}
