// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Text;

using Smdn.Protocols.Imap4.Client.Session;

namespace Smdn.Protocols.Imap4 {
  public class ImapMailbox {
    internal ImapSession Session {
      get; set;
    }

    public ImapUri Url {
      get
      {
        if (Session == null)
          return new ImapUri(ImapUri.EscapeUriString(ImapEncoding.ToModifiedUTF7String(name)));
        else
          return new ImapUri(Session.ServerUrl, ImapUri.EscapeUriString(ImapEncoding.ToModifiedUTF7String(name)));
      }
    }

    /// <value>decoded mailbox name</value>
    public string Name {
      get { return name; }
      internal set { name = value; }
    }

    public string HierarchyDelimiter {
      get { return hierarchyDelimiter; }
      internal set { hierarchyDelimiter = value; }
    }

    public ImapMailboxFlagList Flags {
      get { return flags; }
      internal set { flags = value; }
    }

    public bool ReadOnly {
      get { return readOnly; }
      internal set { readOnly = value; }
    }

    public ImapMessageFlagList ApplicableFlags {
      get { return applicableFlags; }
      internal set { applicableFlags = value; }
    }

    public ImapMessageFlagList PermanentFlags {
      get { return permanentFlags; }
      internal set { permanentFlags = value; }
    }

    public long ExistsMessage {
      get { return existsMessage; }
      internal set { existsMessage = value; }
    }

    public long RecentMessage {
      get { return recentMessage; }
      internal set { recentMessage = value; }
    }

    public long UnseenMessage {
      get { return unseenMessage; }
      internal set { unseenMessage = value; }
    }

    public long FirstUnseen {
      get { return firstUnseen; }
      internal set { firstUnseen = value; }
    }

    public long UidNext {
      get { return uidNext; }
      internal set { uidNext = value; }
    }

    public long UidValidity {
      get { return uidValidity; }
      internal set { uidValidity = value; }
    }

    /*
     * RFC 4315 Internet Message Access Protocol (IMAP) - UIDPLUS extension
     * http://tools.ietf.org/html/rfc4315
     */
    public bool UidPersistent {
      get { return uidPersistent; }
      internal set { uidPersistent = value; }
    }

    /*
     * RFC 4551 - IMAP Extension for Conditional STORE Operation or Quick Flag Changes Resynchronization
     * http://tools.ietf.org/html/rfc4551
     */
    public bool ModificationSequences {
      get { return modificationSequences; }
      internal set { modificationSequences = value; }
    }

    public ulong HighestModSeq {
      get { return highestModSeq; }
      internal set { highestModSeq = value; }
    }

    internal static ImapMailbox CreateFrom(ImapMailboxList list)
    {
      if (list == null)
        throw new ArgumentNullException("list");

      var mailbox = new ImapMailbox(list.Name);

      mailbox.hierarchyDelimiter = list.HierarchyDelimiter;
      mailbox.flags = list.NameAttributes;

      return mailbox;
    }

    internal ImapMailbox()
    {
      this.name = null;
    }

    /// <remarks>name must be decoded</remarks>
    internal ImapMailbox(string name)
    {
      if (name == null)
        throw new ArgumentNullException("name");

      this.name = name;
    }

    internal void UpdateStatus(ImapStatusAttributeList status)
    {
      if (status.Messages.HasValue)      existsMessage  = (long)status.Messages;
      if (status.Recent.HasValue)        recentMessage  = (long)status.Recent;
      if (status.Unseen.HasValue)        unseenMessage  = (long)status.Unseen;
      if (status.UidNext.HasValue)       uidNext        = (long)status.UidNext;
      if (status.UidValidity.HasValue)   uidValidity    = (long)status.UidValidity;
      if (status.HighestModSeq.HasValue) highestModSeq  = (ulong)status.HighestModSeq;
    }

    public override string ToString()
    {
      return string.Format("{{Name={0}, HierarchyDelimiter={1}, Flags={2}, ReadOnly={3}, ApplicableFlags={4}, PermanentFlags={5}, ExistsMessage={6}, RecentMessage={7}, UnseenMessage={8}, FirstUnseen={9}, UidNext={10}, UidValidity={11}, UidPersistent={12}, ModificationSequences={13}, HighestModSeq={14}}}",
                           name,
                           hierarchyDelimiter,
                           flags,
                           readOnly,
                           applicableFlags,
                           permanentFlags,
                           existsMessage,
                           recentMessage,
                           unseenMessage,
                           firstUnseen,
                           uidNext,
                           uidValidity,
                           uidPersistent,
                           modificationSequences,
                           highestModSeq);
    }

    private string name;
    private string hierarchyDelimiter = null;
    private bool readOnly = false;
    private ImapMailboxFlagList flags = null;
    private ImapMessageFlagList applicableFlags = null;
    private ImapMessageFlagList permanentFlags = null;
    private long existsMessage = 0;
    private long recentMessage = 0;
    private long unseenMessage = 0;
    private long firstUnseen = 0;
    private long uidNext = 0;
    private long uidValidity = 0;
    private bool uidPersistent = true;
    private bool modificationSequences = false;
    private ulong highestModSeq = 0;
  }
}