// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Text;

namespace Smdn.Protocols.Imap4 {
  // combinable data item types:
  //   ImapCombinableDataItem
  //     => handles 'data item names' etc.
  //     ImapFetchDataItem
  //       => handles 'message data item names or macro'
  //   * ImapSearchCriteria
  //       => handles 'searching criteria'
  //     ImapSortCriteria
  //       => handles 'sort criteria'
  //     ImapStatusDataItem
  //       => handles 'status data item names'
  //     ImapStoreDataItem
  //       => handles 'message data item name' and 'value for message data item'

  // 6.4.4. SEARCH Command
  //   searching criteria
  public class ImapSearchCriteria : ImapCombinableDataItem, IImapMultipleExtension {
    //  <sequence set>
    //     Messages with message sequence numbers corresponding to the
    //     specified message sequence number set.
    public static ImapSearchCriteria SequenceSet(ImapSequenceSet sequenceSet)
    {
      return new ImapSearchCriteria(sequenceSet.ToString());
    }

    // ALL
    //     All messages in the mailbox; the default initial key for
    //     ANDing.
    public static readonly ImapSearchCriteria All
      = new ImapSearchCriteria("ALL");

    // ANSWERED
    //     Messages with the \Answered flag set.
    public static readonly ImapSearchCriteria Answered
      = new ImapSearchCriteria("ANSWERED");

    // BCC <string>
    //     Messages that contain the specified string in the envelope
    //     structure's BCC field.
    public static ImapSearchCriteria Bcc(string str)
    {
      return new ImapSearchCriteria("BCC", new ImapQuotedString(str));
    }

    // BEFORE <date>
    //     Messages whose internal date (disregarding time and timezone)
    //     is earlier than the specified date.
    public static ImapSearchCriteria Before(DateTime date)
    {
      return new ImapSearchCriteria("BEFORE", ImapDateTimeFormat.ToDateString(date));
    }

    // BODY <string>
    //     Messages that contain the specified string in the body of the
    //     message.
    public static ImapSearchCriteria Body(string str)
    {
      return new ImapSearchCriteria("BODY", new ImapLiteralString(str));
    }

    // CC <string>
    //     Messages that contain the specified string in the envelope
    //     structure's CC field.
    public static ImapSearchCriteria Cc(string str)
    {
      return new ImapSearchCriteria("CC", new ImapQuotedString(str));
    }

    // DELETED
    //     Messages with the \Deleted flag set.
    public static readonly ImapSearchCriteria Deleted
      = new ImapSearchCriteria("DELETED");

    // DRAFT
    //     Messages with the \Draft flag set.
    public static readonly ImapSearchCriteria Draft
      = new ImapSearchCriteria("DRAFT");

    // FLAGGED
    //     Messages with the \Flagged flag set.
    public static readonly ImapSearchCriteria Flagged
      = new ImapSearchCriteria("FLAGGED");

    // FROM <string>
    //     Messages that contain the specified string in the envelope
    //     structure's FROM field.
    public static ImapSearchCriteria From(string str)
    {
      return new ImapSearchCriteria("FROM", new ImapQuotedString(str));
    }

    // HEADER <field-name> <string>
    //     Messages that have a header with the specified field-name (as
    //     defined in [RFC-2822]) and that contains the specified string
    //     in the text of the header (what comes after the colon).  If the
    //     string to search is zero-length, this matches all messages that
    //     have a header line with the specified field-name regardless of
    //     the contents.
    public static ImapSearchCriteria Header(string fieldName, string str)
    {
      return new ImapSearchCriteria("HEADER", fieldName, new ImapQuotedString(str));
    }

    // KEYWORD <flag>
    //     Messages with the specified keyword flag set.
    public static ImapSearchCriteria Keyword(string keyword)
    {
      return new ImapSearchCriteria("KEYWORD", keyword);
    }

    // LARGER <n>
    //     Messages with an [RFC-2822] size larger than the specified
    //     number of octets.
    public static ImapSearchCriteria Larger(long size)
    {
      if (size < 0)
        throw new ArgumentOutOfRangeException("size", "must be zero or positive number");

      return new ImapSearchCriteria("LARGER", size.ToString());
    }

    // NEW
    //     Messages that have the \Recent flag set but not the \Seen flag.
    //     This is functionally equivalent to "(RECENT UNSEEN)".
    public static readonly ImapSearchCriteria New
      = new ImapSearchCriteria("NEW");

    // OLD
    //     Messages that do not have the \Recent flag set.  This is
    //     functionally equivalent to "NOT RECENT" (as opposed to "NOT
    //     NEW").
    public static readonly ImapSearchCriteria Old
      = new ImapSearchCriteria("OLD");

    // ON <date>
    //     Messages whose internal date (disregarding time and timezone)
    //     is within the specified date.
    public static ImapSearchCriteria On(DateTime date)
    {
      return new ImapSearchCriteria("ON", ImapDateTimeFormat.ToDateString(date));
    }

    // RECENT
    //     Messages that have the \Recent flag set.
    public static readonly ImapSearchCriteria Recent
      = new ImapSearchCriteria("RECENT");

    // SEEN
    //     Messages that have the \Seen flag set.
    public static readonly ImapSearchCriteria Seen
      = new ImapSearchCriteria("SEEN");

    // SENTBEFORE <date>
    //     Messages whose [RFC-2822] Date: header (disregarding time and
    //     timezone) is earlier than the specified date.
    public static ImapSearchCriteria SentBefore(DateTime date)
    {
      return new ImapSearchCriteria("SENTBEFORE", ImapDateTimeFormat.ToDateString(date));
    }

    // SENTON <date>
    //     Messages whose [RFC-2822] Date: header (disregarding time and
    //     timezone) is within the specified date.
    public static ImapSearchCriteria SentOn(DateTime date)
    {
      return new ImapSearchCriteria("SENTON", ImapDateTimeFormat.ToDateString(date));
    }

    // SENTSINCE <date>
    //     Messages whose [RFC-2822] Date: header (disregarding time and
    //     timezone) is within or later than the specified date.
    public static ImapSearchCriteria SentSince(DateTime date)
    {
      return new ImapSearchCriteria("SENTSINCE", ImapDateTimeFormat.ToDateString(date));
    }

    // SINCE <date>
    //     Messages whose internal date (disregarding time and timezone)
    //     is within or later than the specified date.
    public static ImapSearchCriteria Since(DateTime date)
    {
      return new ImapSearchCriteria("SINCE", ImapDateTimeFormat.ToDateString(date));
    }

    // SMALLER <n>
    //     Messages with an [RFC-2822] size smaller than the specified
    //     number of octets.
    public static ImapSearchCriteria Smaller(long size)
    {
      if (size < 0)
        throw new ArgumentOutOfRangeException("size", "must be zero or positive number");

      return new ImapSearchCriteria("SMALLER", size.ToString());
    }

    // SUBJECT <string>
    //     Messages that contain the specified string in the envelope
    //     structure's SUBJECT field.
    public static ImapSearchCriteria Subject(string str)
    {
      return new ImapSearchCriteria("SUBJECT", new ImapLiteralString(str));
    }

    // TEXT <string>
    //     Messages that contain the specified string in the header or
    //     body of the message.
    public static ImapSearchCriteria Text(string str)
    {
      return new ImapSearchCriteria("TEXT", new ImapLiteralString(str));
    }

    // TO <string>
    //     Messages that contain the specified string in the envelope
    //     structure's TO field.
    public static ImapSearchCriteria To(string str)
    {
      return new ImapSearchCriteria("TO", new ImapQuotedString(str));
    }

    // UID <sequence set>
    //     Messages with unique identifiers corresponding to the specified
    //     unique identifier set.  Sequence set ranges are permitted.
    public static ImapSearchCriteria Uid(ImapSequenceSet sequenceSet)
    {
      return new ImapSearchCriteria("UID", sequenceSet.ToString());
    }

    // UNANSWERED
    //     Messages that do not have the \Answered flag set.
    public static readonly ImapSearchCriteria Unanswered
      = new ImapSearchCriteria("UNANSWERED");

    // UNDELETED
    //     Messages that do not have the \Deleted flag set.
    public static readonly ImapSearchCriteria Undeleted
      = new ImapSearchCriteria("UNDELETED");

    // UNDRAFT
    //     Messages that do not have the \Draft flag set.
    public static readonly ImapSearchCriteria Undraft
      = new ImapSearchCriteria("UNDRAFT");

    // UNFLAGGED
    //     Messages that do not have the \Flagged flag set.
    public static readonly ImapSearchCriteria Unflagged
      = new ImapSearchCriteria("UNFLAGGED");

    // UNKEYWORD <flag>
    //     Messages that do not have the specified keyword flag set.
    public static ImapSearchCriteria Unkeyword(string keyword)
    {
      return new ImapSearchCriteria("UNKEYWORD", keyword);
    }

    // UNSEEN
    //     Messages that do not have the \Seen flag set.
    public static readonly ImapSearchCriteria Unseen
      = new ImapSearchCriteria("UNSEEN");

    // NOT <search-key>
    //     Messages that do not match the specified search key.
    public ImapSearchCriteria Not()
    {
      return !this;
    }

    public static ImapSearchCriteria Not(ImapSearchCriteria c)
    {
      return !c;
    }

    public static ImapSearchCriteria operator! (ImapSearchCriteria c)
    {
      return new ImapSearchCriteria(c.requiredCapabilities,
                                    "NOT", new ImapParenthesizedString(c.Items));
    }

    // OR <search-key1> <search-key2>
    //     Messages that match either search key.
    public ImapSearchCriteria Or(ImapSearchCriteria searchingCriteria)
    {
      return this | searchingCriteria;
    }

    public static ImapSearchCriteria Or(ImapSearchCriteria x, ImapSearchCriteria y)
    {
      return x | y;
    }

    public static ImapSearchCriteria operator| (ImapSearchCriteria x, ImapSearchCriteria y)
    {
      return new ImapSearchCriteria(MergeRequiredCapabilities(x, y),
                                    "OR", new ImapParenthesizedString(x.Items), new ImapParenthesizedString(y.Items));
    }

    // AND
    public ImapSearchCriteria And(ImapSearchCriteria searchingCriteria)
    {
      return this & searchingCriteria;
    }

    public static ImapSearchCriteria And(ImapSearchCriteria x, ImapSearchCriteria y)
    {
      return x & y;
    }

    public static ImapSearchCriteria operator& (ImapSearchCriteria x, ImapSearchCriteria y)
    {
      return new ImapSearchCriteria(MergeRequiredCapabilities(x, y),
                                    GetCombinedItems(x, y));
    }

    /*
     * RFC 4551 - IMAP Extension for Conditional STORE Operation or Quick Flag Changes Resynchronization
     * http://tools.ietf.org/html/rfc4551
     */

    // 3.4. MODSEQ Search Criterion in SEARCH
    //    The MODSEQ criterion for the SEARCH command allows a client to search
    //    for the metadata items that were modified since a specified moment.
    // 
    //    Syntax:  MODSEQ [<entry-name> <entry-type-req>] <mod-sequence-valzer>
    public static ImapSearchCriteria ModSeq(ulong modificationSequence)
    {
      return ModSeq(modificationSequence, null, null);
    }

    public static ImapSearchCriteria ModSeqSharedEntry(ulong modificationSequence, ImapMessageFlag flag)
    {
      return ModSeq(modificationSequence, flag, "shared");
    }

    public static ImapSearchCriteria ModSeqPrivateEntry(ulong modificationSequence, ImapMessageFlag flag)
    {
      return ModSeq(modificationSequence, flag, "priv");
    }

    public static ImapSearchCriteria ModSeqAllEntry(ulong modificationSequence, ImapMessageFlag flag)
    {
      return ModSeq(modificationSequence, flag, "all");
    }

    private static ImapSearchCriteria ModSeq(ulong modSequenceValzer, ImapMessageFlag entryFlagName, ImapString entryType)
    {
      if (entryType == null)
        return new ImapSearchCriteria(new[] {ImapCapability.CondStore},
                                      "MODSEQ",
                                      modSequenceValzer.ToString());
      else
        return new ImapSearchCriteria(new[] {ImapCapability.CondStore},
                                      "MODSEQ",
                                      new ImapQuotedString(string.Format("/flags/{0}", entryFlagName.ToString())),
                                      entryType,
                                      modSequenceValzer.ToString());
    }

    /*
     * RFC 5032 - WITHIN Search Extension to the IMAP Protocol
     * http://tools.ietf.org/html/rfc5032
     */

    // 3. Formal Syntax
    //    This document extends RFC 3501 [RFC3501] with two new search keys:
    //    OLDER <interval> and YOUNGER <interval>.
    // 
    //    search-key =/ ( "OLDER" / "YOUNGER" ) SP nz-number
    //                   ; search-key defined in RFC 3501
    public static ImapSearchCriteria Older(long intervalSeconds)
    {
      if (intervalSeconds <= 0)
        throw new ArgumentOutOfRangeException("intervalSeconds", "must be non-zero positive number");

      return new ImapSearchCriteria(new[] {ImapCapability.Within}, "OLDER", intervalSeconds.ToString());
    }

    public static ImapSearchCriteria Younger(long intervalSeconds)
    {
      if (intervalSeconds <= 0)
        throw new ArgumentOutOfRangeException("intervalSeconds", "must be non-zero positive number");

      return new ImapSearchCriteria(new[] {ImapCapability.Within}, "YOUNGER", intervalSeconds.ToString());
    }

    /*
     * draft-ietf-morg-inthread-00 - The IMAP SEARCH=INTHREAD and THREAD=REFS Extensions
     * http://tools.ietf.org/html/draft-ietf-morg-inthread-00
     */

    // 3.1. The INTHREAD Search Key
    //     INTHREAD takes one argument, which is another search key.
    // 
    //     The INTHREAD search-key matches a message if its subsidiary search-
    //     key matches at least one message in the same thread as the message.
    // 
    //     This command finds all messages in an entire thread concerning the
    //     meetings where fizzle was discussed:
    // 
    //          C: a UID SEARCH INTHREAD (SUBJECT meeting BODY fizzle)
    // 
    //     This command threads all threads containing at least one message
    //     from fred@example.com:
    // 
    //          C: a UID THREAD REFS utf-8 INTHREAD FROM <fred@example.com>
    public static readonly ImapSearchCriteria InThread
      = new ImapSearchCriteria(new[] {ImapCapability.SearchInThread, ImapCapability.ThreadRefs}, "INTHREAD");

    // 3.2. The THREADROOT Search Key
    //     The THREADROOT search key matches a message if that message does not
    //     have any extant parent according to the active threading algorithm
    //     (see section 3.5).
    // 
    //     This command finds the roots of all threads containing unread
    //     messages:
    // 
    //          C: a UID SEARCH THREADROOT INTHREAD UNSEEN
    public static readonly ImapSearchCriteria ThreadRoot
      = new ImapSearchCriteria(new[] {ImapCapability.SearchInThread, ImapCapability.ThreadRefs}, "THREADROOT");

    // 3.3. The THREADLEAF Search Key
    //     The THREADLEAF search key matches a message if that message has no
    //     extant children in the same mailbox, according to the active
    //     threading algorithm.
    // 
    //     Note that THEADLEAF interacts badly with THREAD=ORDEREDSUBJECT.
    //     THREAD=ORDEREDSUBJECT is defined such that every message is either a
    //     root or a leaf, there are no intermediate nodes.
    // 
    //     This command finds all messages that were (also) sent to me, and to
    //     which noone has answered:
    // 
    //          C: a UID SEARCH THREADLEAF OR TO <me@example.com> CC
    //             <me@example.com>
    public static readonly ImapSearchCriteria ThreadLeaf
      = new ImapSearchCriteria(new[] {ImapCapability.SearchInThread, ImapCapability.ThreadRefs}, "THREADLEAF");

    // 3.4. The MESSAGEID Search Key
    //     The MESSAGEID search key takes a sigle argument, and matches a
    //     message if that message's normalized nessage-id is the same as the
    //     argument.
    // 
    //     This command finds all in the same thread as
    //     <4321.1234321@example.com>:
    // 
    //          C: a UID SEARCH INTHREAD MESSAGEID <4321.1234321@example.com>
    public static ImapSearchCriteria MessageId(string messageId)
    {
      return new ImapSearchCriteria(new[] {ImapCapability.SearchInThread, ImapCapability.ThreadRefs},
                                    "MESSAGEID",
                                    string.Format("<{0}>", messageId));
    }

    internal static ImapSearchCriteria FromUri(ImapUri uri)
    {
      if (uri.Query.Length <= 1)
        return new ImapSearchCriteria(string.Empty);
      else
        return new ImapSearchCriteria(Uri.UnescapeDataString(uri.Query.Substring(1)));
    }

    private static IEnumerable<ImapCapability> MergeRequiredCapabilities(ImapSearchCriteria x, ImapSearchCriteria y)
    {
      var requiredCapabilities = new List<ImapCapability>(x.requiredCapabilities);

      foreach (var cap in y.requiredCapabilities) {
        if (!requiredCapabilities.Contains(cap))
          requiredCapabilities.Add(cap);
      }

      return requiredCapabilities;
    }

    ImapCapability[] IImapMultipleExtension.RequiredCapabilities {
      get { return requiredCapabilities.ToArray(); }
    }

    internal List<ImapCapability> RequiredCapabilities {
      get { return requiredCapabilities; }
    }

    private ImapSearchCriteria(params ImapString[] items)
      : this(new ImapCapability[] {}, items)
    {
    }

    private ImapSearchCriteria(IEnumerable<ImapCapability> requiredCapabilities, params ImapString[] items)
      : base(items)
    {
      this.requiredCapabilities.AddRange(requiredCapabilities);
    }

    internal override ImapStringList GetCombined()
    {
      return ToStringList();
    }

    private /*readonly*/ List<ImapCapability> requiredCapabilities = new List<ImapCapability>();
  }
}