// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections;
using System.Collections.Generic;

namespace Smdn.Protocols.Imap4 {
  public class ImapStringEnumList<TStringEnum> : IEnumerable<TStringEnum>
    where TStringEnum : ImapStringEnum
  {
    public TStringEnum this[string @value] {
      get { return values[@value]; }
    }

    public bool ReadOnly {
      get { return readOnly; }
    }

    public int Count {
      get { return values.Count; }
    }

    public ImapStringEnumList()
      : this(false)
    {
    }

    protected internal ImapStringEnumList(bool readOnly)
      : this(readOnly, new TStringEnum[] {})
    {
    }

    public ImapStringEnumList(IEnumerable<TStringEnum> values)
      : this(false, values)
    {
    }

    protected internal ImapStringEnumList(bool readOnly, IEnumerable<TStringEnum> values)
    {
      if (values == null)
        throw new ArgumentNullException("values");

      this.readOnly = readOnly;

      InternalAddRange(values);
    }

#region "list operations"
    public bool Has(string @value)
    {
      return values.ContainsKey(@value);
    }

    public bool Has(TStringEnum @value)
    {
      return values.ContainsKey(@value.Value);
    }

    public TStringEnum Find(string @value)
    {
      if (values.ContainsKey(@value))
        return values[@value];
      else
        return null;
    }

    public void Add(TStringEnum @value)
    {
      if (readOnly)
        throw new InvalidOperationException("list is readonly");

      InternalAdd(@value);
    }

    private void InternalAdd(TStringEnum @value)
    {
      if (Has(@value))
        throw new InvalidOperationException("already contains value");

      values.Add(@value.Value, @value);
    }

    public void AddRange(IEnumerable<TStringEnum> vals)
    {
      if (readOnly)
        throw new InvalidOperationException("list is readonly");

      InternalAddRange(vals);
    }

    private void InternalAddRange(IEnumerable<TStringEnum> vals)
    {
      foreach (var val in vals) {
        InternalAdd(val);
      }
    }

    public bool Remove(TStringEnum @value)
    {
      if (readOnly)
        throw new InvalidOperationException("list is readonly");

      if (Has(@value))
        return values.Remove(@value.Value);
      else
        return false;
    }
#endregion

#region "IEnumerable"
    public IEnumerator GetEnumerator()
    {
      return values.Values.GetEnumerator();
    }

    IEnumerator<TStringEnum> IEnumerable<TStringEnum>.GetEnumerator()
    {
      return values.Values.GetEnumerator();
    }
#endregion

    public TStringEnum[] ToArray()
    {
      var vals = new TStringEnum[values.Count];

      values.Values.CopyTo(vals, 0);

      return vals;
    }

    public ImapParenthesizedString ToParenthesizedString()
    {
      return new ImapParenthesizedString(ToArray());
    }

    public ImapStringList ToStringList()
    {
      return new ImapStringList(ToArray());
    }

    public string[] ToStringArray()
    {
      return Array.ConvertAll(ToArray(), delegate(TStringEnum val) {
        return val.Value;
      });
    }

    public override string ToString()
    {
      return string.Join(", ", ToStringArray());
    }

    private readonly bool readOnly;
    private /*readonly*/ Dictionary<string, TStringEnum> values
      = new Dictionary<string, TStringEnum>(StringComparer.InvariantCultureIgnoreCase);
  }
}
