// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

namespace Smdn.Protocols.Imap4 {
  // 7.2. Server Responses - Server and Mailbox Status
  // 7.2.6. FLAGS Response
  // 7.3. Server Responses - Mailbox Size
  // 7.3.1. EXISTS Response
  // 7.3.2. RECENT Response
  // 7.4. Server Responses - Message Status
  // 7.4.1. EXPUNGE Response
  // 7.4.2. FETCH Response
  public class ImapUpdatedStatus {
    public ImapUpdatedStatusKind UpdatedKind {
      get { return updatedKind; }
    }

    public ImapMessageFlagList Flags {
      get { return flags; }
    }

    public ImapMessage[] Messages {
      get { return messages; }
    }

    public long[] Expunged {
      get { return expunged; }
    }

    public long? Exists {
      get { return exists; }
    }

    public long? Recent {
      get { return recent; }
    }

    public static ImapUpdatedStatus CreateFlagsUpdated(ImapMessageFlagList flags)
    {
      if (flags == null)
        throw new ArgumentNullException("flags");

      return new ImapUpdatedStatus(flags, new ImapMessage[] {}, new long[] {}, null, null);
    }

    public static ImapUpdatedStatus CreateMessageUpdated(ImapMessage message)
    {
      if (message == null)
        throw new ArgumentNullException("message");

      return CreateMessageUpdated(new[] {message});
    }

    public static ImapUpdatedStatus CreateMessageUpdated(ImapMessage[] messages)
    {
      if (messages == null)
        throw new ArgumentNullException("messages");

      return new ImapUpdatedStatus(null, messages, new long[] {}, null, null);
    }

    public static ImapUpdatedStatus CreateExpungedUpdated(long expunged)
    {
      return CreateExpungedUpdated(new[] {expunged});
    }

    public static ImapUpdatedStatus CreateExpungedUpdated(long[] expunged)
    {
      if (expunged == null)
        throw new ArgumentNullException("expunged");

      return new ImapUpdatedStatus(null, new ImapMessage[] {}, expunged, null, null);
    }

    public static ImapUpdatedStatus CreateExistsUpdated(long exists)
    {
      return new ImapUpdatedStatus(null, new ImapMessage[] {}, new long[] {}, exists, null);
    }

    public static ImapUpdatedStatus CreateRecentUpdated(long recent)
    {
      return new ImapUpdatedStatus(null, new ImapMessage[] {}, new long[] {}, null, recent);
    }

    public ImapUpdatedStatus()
      : this(null, new ImapMessage[] {}, new long[] {}, null, null)
    {
    }

    public ImapUpdatedStatus(ImapMessageFlagList flags, ImapMessage message, long? expunged, long? exists, long? recent)
      : this(flags,
             (message == null) ? new ImapMessage[] {} : new[] {message},
             (expunged == null) ? new long[] {} : new[] {expunged.Value},
             exists,
             recent)
    {
    }

    public ImapUpdatedStatus(ImapMessageFlagList flags, ImapMessage[] messages, long[] expunged, long? exists, long? recent)
    {
      if (messages == null)
        throw new ArgumentNullException("messages");
      if (expunged == null)
        throw new ArgumentNullException("expunged");

      this.flags = flags;
      this.messages = messages;
      this.expunged = expunged;
      this.exists = exists;
      this.recent = recent;
      this.updatedKind =
        ((flags == null)        ? ImapUpdatedStatusKind.None : ImapUpdatedStatusKind.Flags) |
        ((messages.Length == 0) ? ImapUpdatedStatusKind.None : ImapUpdatedStatusKind.Messages) |
        ((expunged.Length == 0) ? ImapUpdatedStatusKind.None : ImapUpdatedStatusKind.Expunged) |
        ((exists == null)       ? ImapUpdatedStatusKind.None : ImapUpdatedStatusKind.Exists) |
        ((recent == null)       ? ImapUpdatedStatusKind.None : ImapUpdatedStatusKind.Recent);
    }

    public ImapUpdatedStatus MergeWith(ImapUpdatedStatus other)
    {
      return new ImapUpdatedStatus(other.flags ?? this.flags,
                               ArrayExtension.Concat(this.messages, other.messages),
                               ArrayExtension.Concat(this.expunged, other.expunged),
                               other.exists ?? this.exists,
                               other.recent ?? this.recent);
    }

    public override string ToString()
    {
      return string.Format("{{UpdatedKind={0}, Messages={1}, Expunge={2}, Exists={3}, Recent={4}, Flags={5}}}",
                           updatedKind,
                           string.Join(",", Array.ConvertAll(messages, delegate(ImapMessage m) { return m.ToString(); })),
                           string.Join(",", Array.ConvertAll(expunged, delegate(long e) { return e.ToString(); })),
                           exists,
                           recent,
                           flags);
    }

    private /*readonly*/ ImapUpdatedStatusKind updatedKind;
    private /*readonly*/ ImapMessageFlagList flags;
    private /*readonly*/ ImapMessage[] messages;
    private /*readonly*/ long[] expunged;
    private /*readonly*/ long? exists;
    private /*readonly*/ long? recent;
  }
}
