// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Text;

using Smdn.Protocols.Imap4.Client.Protocol;

namespace Smdn.Protocols.Imap4 {
  /*
   * RFC 5092 - IMAP URL Scheme
   * http://tools.ietf.org/html/rfc5092
   */
  public class ImapUri : Uri {
    public static readonly string UriSchemeImap = "imap";
    public static readonly string UriSchemeImaps = "imaps";

    public ImapUri(Uri uri)
      : this(uri.ToString())
    {
    }

    public ImapUri(string uriString)
      : base(uriString)
    {
      if (!(string.Equals(Scheme, UriSchemeImap, StringComparison.InvariantCultureIgnoreCase) ||
            string.Equals(Scheme, UriSchemeImaps, StringComparison.InvariantCultureIgnoreCase)))
        throw new ArgumentException(string.Format("scheme must be {0} or {1}", UriSchemeImap, UriSchemeImaps), "uriString");
    }

    public ImapUri(ImapUri baseUri, string relativeUri)
      : base(baseUri, relativeUri)
    {
    }

    public ImapUri(ImapUri baseUri, Uri relativeUri)
      : base(baseUri, relativeUri)
    {
    }

    /*
     *    iserver          = [iuserinfo "@"] host [ ":" port ]
     *                            ; This is the same as "authority" defined
     *                            ; in [URI-GEN].  See [URI-GEN] for "host"
     *                            ; and "port" definitions.
     */

    /// <summary>returns [iuserinfo "@"] host [ ":" port ]</summary>
    public string Server {
      get
      {
        if (string.Empty.Equals(UserInfo)) {
          if (Port == -1)
            return Host;
          else
            return string.Format("{0}:{1}", Host, Port);
        }
        else {
          if (Port == -1)
            return string.Format("{0}@{1}", UserInfo, Host);
          else
            return string.Format("{0}@{1}:{2}", UserInfo, Host, Port);
        }
      }
    }

    public ImapUri ServerUri {
      get { return new ImapUri(string.Format("{0}://{1}", Scheme, Server)); }
    }

    /*
     *    iuserinfo        = enc-user [iauth] / [enc-user] iauth
     *                                 ; conforms to the generic syntax of
     *                                 ; "userinfo" as defined in [URI-GEN].
     */

    /// <summary>returns username</summary>
    public string UserName {
      get
      {
        var index = UserInfo.ToUpperInvariant().IndexOf(";AUTH=");

        if (index < 0)
          return UserInfo;
        else
          return UserInfo.Substring(0, index);
      }
    }

    /// <summary>returns authentication type</summary>
    public ImapAuthenticationMechanism Authentication {
      get
      {
        var index = UserInfo.ToUpperInvariant().IndexOf(";AUTH=");

        if (index == -1)
          return null;
        else
          return ImapAuthenticationMechanism.GetKnownOrCreate(UserInfo.Substring(index + 6));
      }
    }

    /*
     *    imailbox-ref     = enc-mailbox [uidvalidity]
     *    enc-mailbox      = 1*bchar
     *                   ; %-encoded version of [IMAP4] "mailbox"
     *    bchar            = achar / ":" / "@" / "/"
     *    achar            = uchar / "&" / "="
     *                       ;; Same as [URI-GEN] 'unreserved / sub-delims /
     *                       ;; pct-encoded', but ";" is disallowed.
     *    uchar            = unreserved / sub-delims-sh / pct-encoded
     *    sub-delims-sh = "!" / "$" / "'" / "(" / ")" /
     *                    "*" / "+" / ","
     *                       ;; Same as [URI-GEN] sub-delims,
     *                       ;; but without ";", "&" and "=".
     */

    /// <summary>returns mailbox</summary>
    public string Mailbox {
      get
      {
        var segments = Segments;

        if (segments.Length <= 1)
          return null;

        var mailbox = string.Empty;

        for (var i = 1; i < segments.Length; i++) {
          if (segments[i].StartsWith(";"))
            break;

          var delim = segments[i].ToUpperInvariant().IndexOf(";UIDVALIDITY");

          if (0 <= delim) {
            mailbox += segments[i].Substring(0, delim);
            break;
          }

          mailbox += segments[i];
        }

        if (mailbox.EndsWith("/"))
          return ImapEncoding.FromModifiedUTF7String(UnescapeDataString(mailbox.Substring(0, mailbox.Length - 1)));
        else
          return ImapEncoding.FromModifiedUTF7String(UnescapeDataString(mailbox));
      }
    }

    /*
     *    uidvalidity      = ";UIDVALIDITY=" nz-number
     *                        ; See [IMAP4] for "nz-number" definition
     */

    /// <summary>returns UIDVALIDITY value</summary>
    public long UidValidity {
      get
      {
        var uidValidity = CutAbsolutePath(";UIDVALIDITY=", "/");

        if (uidValidity == null)
          return 0;
        else
          return ImapDataConverter.ToNonZeroNumber(uidValidity);
      }
    }

    /*
     *    iuid             = "/" iuid-only
     * 
     *    iuid-only        = ";UID=" nz-number
     *                   ; See [IMAP4] for "nz-number" definition
     */

    /// <summary>returns UID value</summary>
    public long Uid {
      get
      {
        var uid = CutAbsolutePath("/;UID=", "/");

        if (uid == null)
          return 0;
        else
          return ImapDataConverter.ToNonZeroNumber(uid);
      }
    }

    /*
     *    ipartial         = "/" ipartial-only
     * 
     *    ipartial-only    = ";PARTIAL=" partial-range
     * 
     *    partial-range    = number ["." nz-number]
     *                   ; partial FETCH.  The first number is
     *                            ; the offset of the first byte,
     *                            ; the second number is the length of
     *                            ; the fragment.
     */

    /// <summary>returns PARTIAL value</summary>
    public string Partial {
      get { return CutAbsolutePath("/;PARTIAL=", "/"); }
    }

    /*
     *    isection         = "/" isection-only
     * 
     *    isection-only    = ";SECTION=" enc-section
     * 
     *    enc-section      = 1*bchar
     *                   ; %-encoded version of [IMAP4] "section-spec"
     */

    /// <summary>returns SECTION value</summary>
    public string Section {
      get { return CutAbsolutePath("/;SECTION=", "/"); }
    }

    /*
     *    iurlauth        = iurlauth-rump iua-verifier
     * 
     *    iua-verifier    = ":" uauth-mechanism ":" enc-urlauth
     * 
     *    iurlauth-rump   = [expire] ";URLAUTH=" access
     * 
     *    access          = ("submit+" enc-user) / ("user+" enc-user) /
     *                        "authuser" / "anonymous"
     * 
     *    expire          = ";EXPIRE=" date-time
     *                          ; date-time is defined in [DATETIME]
     * 
     *    uauth-mechanism = "INTERNAL" / 1*(ALPHA / DIGIT / "-" / ".")
     *                         ; Case-insensitive.
     *                         ; New mechanisms MUST be registered with IANA.
     */
    // TODO: URLAUTH

    /// <summary>returns URI form of current instance</summary>
    public ImapUriForm Form {
      get
      {
        if (Uid != 0L)
          return ImapUriForm.FetchMessage;
        else if (!string.IsNullOrEmpty(Query))
          return ImapUriForm.SearchMessages;
        else if (!string.IsNullOrEmpty(Mailbox))
          return ImapUriForm.ListMessages;
        else if (!string.IsNullOrEmpty(Host))
          return ImapUriForm.Server;
        else
          return ImapUriForm.Relative;
      }
    }

    private string CutAbsolutePath(string delim, string endsWith)
    {
      var ret = AbsolutePath;

      var start = ret.ToUpperInvariant().IndexOf(delim);

      if (start < 0)
        return null;

      ret = ret.Substring(start + delim.Length);

      var end = ret.IndexOf(endsWith);

      if (end < 0)
        return ret;
      else
        return ret.Substring(0, end);
    }

    public ImapSearchCriteria GetSearchCriteria()
    {
      return ImapSearchCriteria.FromUri(this);
    }

    public ImapFetchDataItem GetFetchDataItem()
    {
      return ImapFetchDataItem.FromUri(this);
    }

    public ImapSequenceSet GetSequenceSet()
    {
      return ImapSequenceSet.FromUri(this);
    }
  }
}
