using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.IO;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Windows.Forms;

using Smdn.Formats.Ini;
using Smdn.Formats.Mime;
using Smdn.Protocols.Imap4;
using Smdn.Protocols.Imap4.WebClient;

namespace Smdn.Applications.TundereBird {
  public class MainWindow : Form {
    public MainWindow()
    {
      var assemblyName = System.Reflection.Assembly.GetExecutingAssembly().GetName();

      ImapSessionManager.CertificateValidationCallback += ValidateServerCertificate;
      ImapSessionManager.ClientID["name"] = assemblyName.Name;
      ImapSessionManager.ClientID["version"] = assemblyName.Version.ToString();

      InitializeComponents();
    }

    protected override void OnLoad (EventArgs e)
    {
      base.OnLoad (e);
    }

    protected override void OnShown (EventArgs e)
    {
      base.OnShown (e);

      Application.DoEvents();

      LoadConfig();

      InitializeMailboxTree();
    }

    protected override void OnClosing (CancelEventArgs e)
    {
      base.OnClosing (e);

      if (!e.Cancel) {
        LogoutAll();
      }
    }

    private bool ValidateServerCertificate(object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
    {
      if (sslPolicyErrors == SslPolicyErrors.None)
        return true;

      var message = new StringBuilder();

      message.AppendLine("サーバ証明書に問題があります。　続行しますか?");
      message.AppendLine();
      message.AppendFormat("SslPolicyErrors: {0}", sslPolicyErrors);
      message.AppendLine();
      message.AppendFormat("X509Certificate: {0}{1}", Environment.NewLine, certificate.ToString(false));
      message.AppendLine();

      if (chain != null) {
        foreach (var element in chain.ChainElements) {
          message.AppendFormat("発行先: {0}", element.Certificate.Subject);
          message.AppendLine();
          message.AppendFormat("発行者: {0}", element.Certificate.Issuer);
          message.AppendLine();
          message.AppendFormat("有効期間: {0} から {1}", element.Certificate.GetEffectiveDateString(), element.Certificate.GetExpirationDateString());
          message.AppendLine();

          foreach (var status in element.ChainElementStatus) {
            message.AppendFormat("Status: {0}", status.Status);
            message.AppendLine();
            message.AppendFormat("StatusInformation: {0}", status.StatusInformation);
            message.AppendLine();
            message.AppendLine();
          }
        }
      }

      var result = MessageBox.Show(message.ToString(),
                                   "ValidateServerCertificate",
                                   MessageBoxButtons.YesNo);

      return (result == DialogResult.Yes);
    }

    private void LoadConfig()
    {
      var config = Document.Load("tunderebird.ini");
      var imapAccounts = new List<ImapAccount>();

      foreach (var section in config) {
        if (string.Empty.Equals(section.Name)) {
          mimeTypesFile = section["MimeTypesFile"] ?? "/etc/mime.types";
        }
        else if (section.Name == "CertFiles") {
          // 証明書ファイルのセクション
          foreach(var entry in section) {
            try {
              ImapSessionManager.CertCollection.Add(new X509Certificate2(entry.Value));
            }
            catch (Exception ex) {
              AppendStatus("証明書{0}を読み込めませんでした", entry.Value);
              AppendStatus("Exception: {0}", ex);
            }
          }
        }
        else {
          // アカウントのセクション
          try {
            var account = new ImapAccount(section.Name, new ImapUri(section["ServerURL"]), section["Password"], (section["TryTLS"] != "no"));

            imapAccounts.Add(account);
          }
          catch (Exception ex) {
            AppendStatus("{0}のアカウント設定を読み込めませんでした", section.Name);
            AppendStatus("Exception: {0}", ex);
          }
        }
      }

      this.imapAccounts = imapAccounts.ToArray();

      if (this.imapAccounts.Length == 0)
        AppendStatus("表示できるアカウント設定がありません。");
    }

    private void InitializeMailboxTree()
    {
      InitializeMailboxTree(imapAccounts);
    }

    private void InitializeMailboxTree(ImapAccount[] accounts)
    {
      foreach (var account in accounts) {
        ImapMailbox[] mailboxes;

        try {
          AppendStatus("{0}のメールボックスを取得しています", account.Name);

          using (var response = account.CreateRequest().GetResponse() as ImapWebResponse) {
            mailboxes = response.Mailboxes;

            account.ServerID = response.ServerID;
            account.ServerNamespaces = response.Namespaces;
            account.ServerCapabilities = response.ServerCapabilities;
          }
        }
        catch (Exception ex) {
          AppendStatus("{0}のメールボックス一覧を取得できませんでした", account.Name);
          AppendStatus("Exception: {0}", ex);
          continue;
        }

        foreach (var mailbox in mailboxes) {
          if (mailbox.Flags.Has(ImapMailboxFlag.NoSelect))
            continue;

          try {
            AppendStatus("メールボックス'{0}'のステータスを取得しています", mailbox.Name);

            var request = account.CreateRequest(mailbox.Url, ImapWebRequestMethods.Examine) as ImapWebRequest;

            using (var response = request.GetResponse() as ImapWebResponse) {
            }
          }
          catch (Exception ex) {
            AppendStatus("メールボックス'{0}'のステータスを取得できませんでした", mailbox.Name);
            AppendStatus("Exception: {0}", ex);
          }
        }

        // アカウントとメールボックスのノードを作成する
        try {
          mailboxTree.BeginUpdate();

          var rootNode = new ImapAccountNode(account);
          var sortedMailboxes = new List<ImapMailbox>(mailboxes);

          sortedMailboxes.Sort(delegate(ImapMailbox x, ImapMailbox y) {
            return string.Compare(x.Name, y.Name);
          });

          foreach (var mailbox in mailboxes) {
            var node = new ImapMailboxNode(account, mailbox);

            if (mailbox.Name.Contains(mailbox.HierarchyDelimiter)) {
              // find parent
              var hierarchy = mailbox.Name.Split(new [] {mailbox.HierarchyDelimiter}, StringSplitOptions.None);
              TreeNode parent = rootNode;
              var found = false;

              for (var i = 0; i < hierarchy.Length - 1; i++) {
                found = false;

                foreach (TreeNode n in parent.Nodes) {
                  if (n is ImapMailboxNode) {
                    if ((n as ImapMailboxNode).Mailbox.Name == string.Join(mailbox.HierarchyDelimiter, hierarchy, 0, i + 1)) {
                      found = true;
                      parent = n;
                      break;
                    }
                  }
                }

                if (!found)
                  break;
              }

              if (found)
                parent.Nodes.Add(node);
              else
                rootNode.Nodes.Add(node);
            }
            else {
              rootNode.Nodes.Add(node);
            }

            imapMailboxNodeMap.Add(mailbox, node);
          }

          rootNode.ExpandAll();

          mailboxTree.Nodes.Add(rootNode);
        }
        finally {
          mailboxTree.EndUpdate();
        }

        mailboxTree.Refresh();
      }
    }

    private void AccountOrMailboxSelected(object sender, TreeNodeMouseClickEventArgs e)
    {
      if (e.Node is ImapMailboxNode) {
        var mailboxNode = e.Node as ImapMailboxNode;

        ViewServerInfo(mailboxNode.Account);

        if (mailboxNode.Mailbox.Flags.Has(ImapMailboxFlag.NoSelect)) {
          AppendStatus("{0}は選択できません", mailboxNode.Mailbox.Name);
          return;
        }

        // メッセージの一覧を取得
        AppendStatus("{0}のメッセージの一覧を取得しています", mailboxNode.Mailbox.Name);

        try {
          using (var response = mailboxNode.Account.CreateRequest(mailboxNode.Mailbox.Url).GetResponse() as ImapWebResponse) {
            CreateMessageList(mailboxNode.Account, response.Messages);

            mailboxNode.UnseenMessage = 0;

            foreach (var message in response.Messages) {
              if (!message.Flags.Has(ImapMessageFlag.Seen))
                mailboxNode.UnseenMessage++;
            }

            mailboxNode.UpdateText();

            AppendStatus("全{0}件のメッセージがあります", response.Messages.Length);
          }
        }
        catch (Exception ex) {
          AppendStatus("{0}のメッセージ一覧を取得できませんでした", mailboxNode.Mailbox.Name);
          AppendStatus("Exception: {0}", ex);
        }
      }
      else if (e.Node is ImapAccountNode) {
        ViewServerInfo((e.Node as ImapAccountNode).Account);
      }
    }

    private void CreateMessageList(ImapAccount account, ImapMessage[] messages)
    {
      try {
        messageListView.BeginUpdate();
        messageListView.Items.Clear();

        foreach (var message in messages) {
          messageListView.Items.Add(new ImapMessageItem(account, message));
        }
      }
      finally {
        messageListView.EndUpdate();
      }
    }

    private void MessageSelected(object sender, EventArgs e)
    {
      if (messageListView.SelectedItems.Count <= 0)
        return;

      MimeMessage message = null;

      if (messageListView.SelectedItems[0] is ImapMessageItem) {
        var messageItem = messageListView.SelectedItems[0] as ImapMessageItem;

        AppendStatus("{0}の本文を取得しています", messageItem.Message.Url);

        try {
          using (var response = messageItem.Account.CreateRequest(messageItem.Message.Url).GetResponse() as ImapWebResponse) {
            try {
              message = MimeMessage.Create(response.GetResponseStream());
            }
            catch (Exception ex) {
              AppendStatus("本文の解析に失敗しました");
              AppendStatus("Exception: {0}", ex);
              return;
            }
          }
        }
        catch (Exception ex) {
          AppendStatus("{0}の本文を取得できませんでした", messageItem.Message.Url);
          AppendStatus("Exception: {0}", ex);
          return;
        }
      }

      ViewMessage(message);

      AppendStatus("完了");
    }

    private void ViewMessage(MimeMessage message)
    {
      if (message.Body == null)
        messageView.Text = "(非テキストメッセージです)";
      else
        messageView.Text = message.Body;

      if (0 < message.SubParts.Count) {
        messageView.Text += "-- multipart message --";
        messageView.Text += Environment.NewLine;

        foreach (var subPart in message.SubParts) {
          if (subPart.Body == null)
            messageView.Text = "(非テキストメッセージです)";
          else
            messageView.Text = subPart.Body;

          messageView.Text += Environment.NewLine;
          messageView.Text += "-- end of part --";
        }
      }
    }

    private void ViewServerInfo(ImapAccount account)
    {
      var sb = new StringBuilder();

      sb.AppendLine("[server capabilities]");
      sb.AppendLine(string.Join(", ", account.ServerCapabilities.ToStringArray()));
      sb.AppendLine();

      sb.AppendLine("[server ID]");

      foreach (var pair in account.ServerID) {
        sb.AppendFormat("{0}: {1}", pair.Key, pair.Value);
        sb.AppendLine();
      }
      sb.AppendLine();

      foreach (var ns in new[] {
        new {Namespaces = account.ServerNamespaces.PersonalNamespaces, Title = "personal"},
        new {Namespaces = account.ServerNamespaces.SharedNamespaces, Title = "shared"},
        new {Namespaces = account.ServerNamespaces.OtherUsersNamespaces, Title = "other users"},
      }) {
        sb.AppendFormat("[namespaces ({0})]", ns.Title);
        sb.AppendLine();
        foreach (var desc in ns.Namespaces) {
          sb.AppendFormat("  prefix: \"{0}\" delimiter: \"{1}\"", desc.Prefix, desc.HierarchyDelimiter);
          sb.AppendLine();
        }
        sb.AppendLine();
      }

      serverView.Text = sb.ToString();
    }

    private readonly string fileDialogMessageFilter = string.Format("{0} ({1})|{1}|{2} ({3})|{3}", "emlファイル", "*.eml", "すべてのファイル", "*.*");

    private void UploadMessage()
    {
      if (mailboxTree.SelectedNode == null || mailboxTree.SelectedNode is ImapAccountNode)
        return;

      var selected = mailboxTree.SelectedNode as ImapMailboxNode;
      var mailboxUrl = selected.Mailbox.Url;
      var account = selected.Account;

      using (var openFileDialog = new OpenFileDialog()) {
        openFileDialog.Filter = fileDialogMessageFilter;
        openFileDialog.Title = string.Format("{0}へファイルをアップロード", selected.Mailbox.Name);
        openFileDialog.Multiselect = false;

        if (openFileDialog.ShowDialog(this) != DialogResult.OK)
          return;

        var request = account.CreateRequest(mailboxUrl, ImapWebRequestMethods.Append);

        using (var requestStream = request.GetRequestStream()) {
          if (Path.GetExtension(openFileDialog.FileName) == ".eml") {
            using (var stream = openFileDialog.OpenFile()) {
              var writer = new BinaryWriter(requestStream);
              var reader = new BinaryReader(stream);

              for (;;) {
                var read = reader.ReadBytes(0x4000);

                writer.Write(read);

                if (read.Length < 0x4000)
                  break;
              }
            }
          }
          else {
            WriteMimeCapsulatedMessageTo(openFileDialog.FileName, requestStream);
          }
        }

        request.GetResponse().Close();

        MessageBox.Show(string.Format("{0}を{1}へアップロードしました", openFileDialog.FileName, selected.Mailbox.Name));
      }
    }

    private void WriteMimeCapsulatedMessageTo(string path, Stream stream)
    {
      var writer = new StreamWriter(stream, Encoding.ASCII);

      writer.NewLine = "\r\n";

      var boundary = string.Format("----------------{0:D16}", DateTime.Now.Ticks);
      var assemblyName = System.Reflection.Assembly.GetExecutingAssembly().GetName();
      var mimeType = MimeType.GetMimeTypeByExtension(mimeTypesFile, Path.GetExtension(path)) ?? "application/octet-stream";

      writer.WriteLine("Date: {0}", DateTime.Now.ToString("r"));
      writer.WriteLine("Subject: uploaded file by {0}", assemblyName.Name);
      writer.WriteLine("User-Agent: {0} {1}", assemblyName.Name, assemblyName.Version);
      writer.WriteLine("Content-Type: multipart/mixed; boundary=\"{0}\"", boundary);
      writer.WriteLine();
      writer.WriteLine("This is a multi-part message in MIME format.");
      writer.WriteLine("--{0}", boundary);
      writer.WriteLine("Content-Type: text/plain; chartset=utf-8; format=flowed");
      writer.WriteLine("Content-Transfer-Encoding: 8bit");
      writer.WriteLine();
      writer.WriteLine("uploaded file by {0}", assemblyName.Name);
      writer.WriteLine("--{0}", boundary);
      writer.WriteLine("Content-Type: {0}; name=\"{1}\"", mimeType, Path.GetFileName(path));
      writer.WriteLine("Content-Transfer-Encoding: base64");
      writer.WriteLine("Content-Disposition: attatchment; filename=\"{0}\"", Path.GetFileName(path));
      writer.WriteLine();
      writer.Write(Convert.ToBase64String(File.ReadAllBytes(path), Base64FormattingOptions.InsertLineBreaks));
      writer.WriteLine("--{0}--", boundary);
    }

    private void SaveAsMessage()
    {
      if (messageListView.SelectedItems.Count <= 0)
        return;

      var selected = messageListView.SelectedItems[0] as ImapMessageItem;
      var messageUrl = selected.Message.Url;
      var account = (imapMailboxNodeMap[selected.Message.Mailbox] as ImapMailboxNode).Account;

      using (var saveFileDialog = new SaveFileDialog()) {
        saveFileDialog.Filter = fileDialogMessageFilter;
        saveFileDialog.Title = string.Format("{0}をファイルに保存", messageUrl.LocalPath);

        if (saveFileDialog.ShowDialog(this) != DialogResult.OK)
          return;

        using (var response = account.CreateRequest(messageUrl).GetResponse()) {
          using (var stream = saveFileDialog.OpenFile()) {
            var reader = new BinaryReader(response.GetResponseStream());
            var writer = new BinaryWriter(stream);

            for (;;) {
              var read = reader.ReadBytes(0x4000);

              writer.Write(read);

              if (read.Length < 0x4000)
                break;
            }
          }
        }

        MessageBox.Show(string.Format("{0}を{1}に保存しました", messageUrl.LocalPath, saveFileDialog.FileName));
      }
    }

    private void NoOp()
    {
      foreach (var account in imapAccounts) {
        using (var response = account.CreateRequest(null, ImapWebRequestMethods.NoOp).GetResponse() as ImapWebResponse) {
        }
      }
    }

    private void LogoutAll()
    {
      foreach (var account in imapAccounts) {
        try {
          ImapSessionManager.DisconnectFrom(account.ServerUri);
        }
        catch (Exception ex) {
          AppendStatus("{0}からログアウト出来ませんでした", account.ServerUri);
          AppendStatus("Exception: {0}", ex);
        }
      }
    }

    private void AppendStatus(string format, params object[] args)
    {
      statusView.Text += string.Format(format, args);
      statusView.Text += Environment.NewLine;
      statusView.Select(statusView.TextLength - 1, 0);
      statusView.ScrollToCaret();
      statusView.Update();

      Console.Error.WriteLine(format, args);
    }

    private void InitializeComponents()
    {
      this.Text = "TundereBird";
      this.ClientSize = new Size(800, 600);
      this.StartPosition = FormStartPosition.CenterScreen;

      menuStrip = new MenuStrip();
      menuStrip.Dock = DockStyle.Top;

      var menuItemFile = new ToolStripMenuItem("ファイル (&F)");

      menuItemFileSaveAs = new ToolStripMenuItem("メッセージを別名で保存 (&A)");
      menuItemFileSaveAs.Click += delegate {
        SaveAsMessage();
      };

      menuItemFile.DropDownItems.Add(menuItemFileSaveAs);

      menuItemFileUpload = new ToolStripMenuItem("メッセージをアップロード (&U)");
      menuItemFileUpload.Click += delegate {
        UploadMessage();
      };

      menuItemFile.DropDownItems.Add(menuItemFileUpload);

      menuItemFile.DropDownItems.Add(new ToolStripSeparator());

      var menuItemFileQuit = new ToolStripMenuItem("終了 (&Q)");

      menuItemFileQuit.Click += delegate {
        this.Close();
      };

      menuItemFile.DropDownItems.Add(menuItemFileQuit);

      menuStrip.Items.Add(menuItemFile);

      this.Controls.Add(menuStrip);

      mailboxTree = new TreeView();
      mailboxTree.Location = new Point(0, menuStrip.Height);
      mailboxTree.Size = new Size(280, 600 - menuStrip.Height);
      mailboxTree.Anchor = AnchorStyles.Left | AnchorStyles.Top | AnchorStyles.Bottom;
      mailboxTree.ShowLines = true;
      mailboxTree.ShowPlusMinus = true;
      mailboxTree.NodeMouseClick += AccountOrMailboxSelected;

      this.Controls.Add(mailboxTree);

      messageListView = new ListView();
      messageListView.Location = new Point(280, menuStrip.Height);
      messageListView.Size = new Size(520, 180 - menuStrip.Height);
      messageListView.Anchor = AnchorStyles.Left | AnchorStyles.Right | AnchorStyles.Top;
      messageListView.View = View.Details;
      messageListView.FullRowSelect = true;
      messageListView.Columns.Add("タイトル", 160);
      messageListView.Columns.Add("サイズ", 60);
      messageListView.Columns.Add("受信日", 100);
      messageListView.Columns.Add("送信者", 100);
      messageListView.Columns.Add("フラグ", 200);
      messageListView.SelectedIndexChanged += MessageSelected;

      this.Controls.Add(messageListView);

      messageView = new TextBox();
      messageView.Location = new Point(280, 180);
      messageView.Size = new Size(520, 320);
      messageView.Anchor = AnchorStyles.Left | AnchorStyles.Right | AnchorStyles.Top | AnchorStyles.Bottom;
      messageView.Multiline = true;
      messageView.ScrollBars = ScrollBars.Both;

      this.Controls.Add(messageView);

      statusTab = new TabControl();
      statusTab.Location = new Point(280, 500);
      statusTab.Size = new Size(520, 100);
      statusTab.Anchor = AnchorStyles.Left | AnchorStyles.Right | AnchorStyles.Bottom;

      var tabPage1 = new TabPage("ステータス");

      statusView = new TextBox();
      statusView.Location = new Point(0, 0);
      statusView.Dock = DockStyle.Fill;
      statusView.ReadOnly = true;
      statusView.Multiline = true;
      statusView.ScrollBars = ScrollBars.Both;

      tabPage1.Controls.Add(statusView);

      var tabPage2 = new TabPage("サーバ情報");

      serverView = new TextBox();
      serverView.Location = new Point(0, 0);
      serverView.Dock = DockStyle.Fill;
      serverView.ReadOnly = true;
      serverView.Multiline = true;
      serverView.ScrollBars = ScrollBars.Both;

      tabPage2.Controls.Add(serverView);

      statusTab.TabPages.Add(tabPage1);
      statusTab.TabPages.Add(tabPage2);

      statusTab.SelectedIndex = 0;

      this.Controls.Add(statusTab);

      timerUpdate = new Timer();
      timerUpdate.Enabled = true;
      timerUpdate.Interval = 5 * 60 * 1000; // 5 min
      timerUpdate.Tick += delegate {
        NoOp();
      };
    }

    private class ImapAccount {
      public string Name {
        get; private set;
      }

      public ImapUri ServerUri {
        get; private set;
      }

      public bool UseTlsIfAvailable {
        get; private set;
      }

      public Dictionary<string, string> ServerID {
        get; set;
      }

      public ImapNamespace ServerNamespaces {
        get; set;
      }

      public ImapCapabilityList ServerCapabilities {
        get; set;
      }

      public ImapAccount(string name, ImapUri serverUri, string password, bool useTlsIfAvailable)
      {
        this.Name = name;
        this.ServerUri = serverUri;
        this.UseTlsIfAvailable = useTlsIfAvailable;

        var port = serverUri.Port;

        if (port == -1) {
          if (serverUri.Scheme == ImapUri.UriSchemeImap)
            port = Smdn.Protocols.Imap4.Client.Protocol.ImapConnection.ImapDefaultPort;
          else
            port = Smdn.Protocols.Imap4.Client.Protocol.ImapConnection.ImapsDefaultPort;
        }

        this.credential = new NetworkCredential(serverUri.UserName, password);
      }

      public ImapWebRequest CreateRequest()
      {
        return CreateRequest(null, null);
      }

      public ImapWebRequest CreateRequest(ImapUri requestUri)
      {
        return CreateRequest(requestUri, null);
      }

      public ImapWebRequest CreateRequest(ImapUri requestUri, string method)
      {
        if (requestUri == null)
          requestUri = ServerUri;
        else
          requestUri = new ImapUri(ServerUri, ServerUri.MakeRelativeUri(requestUri));

        var request = WebRequest.Create(requestUri) as ImapWebRequest;

        request.UseTlsIfAvailable = UseTlsIfAvailable;
        request.Credentials = credential;

        if (method != null)
          request.Method = method;

        return request;
      }

      private NetworkCredential credential;
    }

    private class ImapAccountNode : TreeNode {
      public ImapAccount Account{
        get; private set;
      }

      public ImapAccountNode(ImapAccount account)
      {
        this.Account = account;

        Text = string.Format("{0} - {1}", account.Name, account.ServerUri.Server);
      }
    }

    private class ImapMailboxNode : TreeNode {
      public ImapAccount Account {
        get; private set;
      }

      public ImapMailbox Mailbox {
        get; set;
      }

      public long? UnseenMessage {
        get; set;
      }

      public ImapMailboxNode(ImapAccount account, ImapMailbox mailbox)
      {
        this.Account = account;
        this.Mailbox = mailbox;
        this.UnseenMessage = null;

        if (mailbox.Flags.Has(ImapMailboxFlag.NoSelect))
          this.ForeColor = SystemColors.GrayText;

        UpdateText();
      }

      public void UpdateText()
      {
        var nameHierarchy = Mailbox.Name.Split(new [] {Mailbox.HierarchyDelimiter}, StringSplitOptions.None);
        var name = nameHierarchy[nameHierarchy.Length - 1];

        if (UnseenMessage == null) {
          Text = string.Format("{0} 新着{1} 全{2}",
                               name,
                               Mailbox.RecentMessage,
                               Mailbox.ExistsMessage);
        }
        else {
          Text = string.Format("{0} 新着{1} 未読{2} 全{3}",
                               name,
                               Mailbox.RecentMessage,
                               UnseenMessage,
                               Mailbox.ExistsMessage);
        }
      }
    }

    private class ImapMessageItem : ListViewItem {
      public ImapAccount Account {
        get; private set;
      }

      public ImapMessage Message {
        get; private set;
      }

      public ImapMessageItem(ImapAccount account, ImapMessage message)
      {
        this.Account = account;
        this.Message = message;

        if (message.Envelope == null) {
          this.Text = string.Empty;
          this.SubItems.Add(message.Rfc822Size.ToString());
          this.SubItems.Add(string.Empty);
          this.SubItems.Add(string.Empty);
          this.SubItems.Add(message.Flags.ToString());
        }
        else {
          this.Text = Smdn.Formats.Mime.Decoder.DecodeMimeEncodedString(message.Envelope.Subject ?? string.Empty);
          this.SubItems.Add(message.Rfc822Size.ToString());
          this.SubItems.Add(message.Envelope.Date);
          this.SubItems.Add(FromAddress(message.Envelope.From));
          this.SubItems.Add(message.Flags.ToString());
        }
      }

      private string FromAddress(ImapAddress[] addresses)
      {
        if (addresses == null || addresses.Length == 0)
          return string.Empty;

        return string.Join(", ", Array.ConvertAll(addresses, delegate(ImapAddress address) {
          return string.Format("{0} <{1}@{2}>",
                               Smdn.Formats.Mime.Decoder.DecodeMimeEncodedString(address.Name ?? string.Empty),
                               address.Mailbox,
                               address.Host);
        }));
      }
    }

    private string mimeTypesFile;

    private ImapAccount[] imapAccounts;
    private readonly Dictionary<ImapMailbox, ImapMailboxNode> imapMailboxNodeMap = new Dictionary<ImapMailbox, ImapMailboxNode>();

    private MenuStrip menuStrip;
    private ToolStripMenuItem menuItemFileSaveAs;
    private ToolStripMenuItem menuItemFileUpload;
    private TreeView mailboxTree;
    private ListView messageListView;
    private TextBox messageView;
    private TabControl statusTab;
    private TextBox statusView;
    private TextBox serverView;
    private Timer timerUpdate;
  }
}
